(* CPPLc *)
Require Export CPPL.
Open Scope list_scope .

(* Cont Strands *)
  Definition cont_strands c :
   { ss:list (strand * t) |
   forall (s:strand) (v:t),
   List.In (s,v) ss <->
   eval_cont c s v}.
  Proof.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ].

   exists (((faildesc::nil),empty) :: nil). simpl.
   intros s v. split.
   intros [EQ | F]; [idtac | contradiction ].
   inversion_clear EQ. intuition.
   intros EC. inversion_clear EC.
   intuition.

   exists (((((Outgoing, (Join (MConst C.RETURN) 
              (Join (MVar ai) (vars_msg evs)))), f) :: nil),
            empty) :: ((faildesc::nil),empty) :: nil). 
   simpl. intros s v. intuition.
   inversion_clear H0. intuition.
   inversion_clear H. intuition.
   inversion_clear H. intuition.
   left. reflexivity.

   destruct (Hc) as [ss IH].
   destruct lv as [| | | | cv | m ].

   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (MConst C.NEW_NONCE), 
                 (Formula C.NEW_NONCE ((LogicV v) :: nil))::nil)
               :: s,
             (add (varid v) nv))) ss)).
   intros s nv. simpl. 
   split. intros [EQ | REC].
   inversion_clear EQ. intuition.
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion_clear EQ.
   rewrite IH in InSS. intuition.
   intros EC. inversion_clear EC. 
   left. reflexivity.
   right. rewrite in_map_iff.
   exists (s0,v0). subst s'.
   rewrite IH. intuition.
   
   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
             let (s,nv) := snv in
            ((Outgoing, (MConst C.NEW_SYMKEY), 
                 (Formula C.NEW_SYMKEY ((LogicV v) :: nil))::nil)
              :: s,
             (add (varid v) nv))) ss)).
   intros s nv. simpl. 
   split. intros [EQ | REC].
   inversion_clear EQ. intuition.
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion_clear EQ.
   rewrite IH in InSS. intuition.
   intros EC. inversion_clear EC. 
   left. reflexivity.
   right. rewrite in_map_iff.
   exists (s0,v0). subst s'.
   rewrite IH. intuition.

   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (MConst C.NEW_PUBKEY), 
                 (Formula C.NEW_PUBKEY ((LogicV v) :: nil))::nil)
                 :: s,
             (add (varid v) nv))) ss)).
   intros s nv. simpl. 
   split. intros [EQ | REC].
   inversion_clear EQ. intuition.
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion_clear EQ.
   rewrite IH in InSS. intuition.
   intros EC. inversion_clear EC. 
   left. reflexivity.
   right. rewrite in_map_iff.
   exists (s0,v0). subst s'.
   rewrite IH. intuition.

   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (MConst C.ACCEPT), 
                 (Formula C.ACCEPT ((LogicV v) :: nil))::nil)
                 :: s,
              nv)) ss)).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. intuition.
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion EQ. subst s nv. clear EQ.
   rewrite IH in InSS. intuition.
   intros EC. inversion_clear EC.
   left. intuition.
   rewrite in_map_iff. rewrite <- IH in H.
   right. exists (s0,nv). subst s'.
   intuition.

   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (MConst C.CONNECT), 
                 (Formula C.CONNECT ((DataV cv) :: (LogicV v)
                 :: nil))::nil) :: s,
              nv)) ss)).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion EQ. subst s nv. clear EQ. apply LET_CONNECT.
   rewrite <- IH. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite in_map_iff.
   exists (s0,nv). subst s'.
   split. reflexivity. rewrite IH.
   assumption.

   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (MConst C.BIND), 
                 (Formula C.BIND ((LogicV v) :: nil))::nil) :: 
                 (strand_replace v m s),
              nv)) ss)).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion EQ. subst s nv. clear EQ. apply LET_LVMSG.
   rewrite <- IH. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite in_map_iff.
   exists (s0,nv). subst s'.
   split. reflexivity. rewrite IH.
   assumption.

   destruct Hsk as [ss IH].
   destruct Hfk as [sfk IHfk].
   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (MConst C.ASSERT), f) :: s,
             nv)) ss) ++ sfk).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite <- in_app_iff in REC.
   destruct REC as [REC | InSFK].
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSS]].
   inversion EQ. subst s nv. clear EQ. apply ASSERT.
   rewrite <- IH. assumption.
   apply ASSERT_ALTERNATIVE.
   rewrite <- IHfk. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite <- in_app_iff.
   rewrite in_map_iff.
   left. exists (s0,nv). subst s'.
   split. reflexivity. rewrite IH.
   assumption.
   right. rewrite <- in_app_iff.
   right. rewrite IHfk. assumption.   

   destruct Hsk as [ssk IHsk].
   destruct Hfk as [sfk IHfk].
   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (Join (MConst C.MSG) 
              (Join (MVar v) m)), f) :: s,
             nv)) ssk) ++ sfk).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite <- in_app_iff in REC.
   destruct REC as [REC | InSFK].
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSSK]].
   inversion EQ. subst s nv. clear EQ. apply SEND_WITH_GUARANTEE.
   rewrite <- IHsk. assumption.
   apply SEND_ALTERNATIVE. rewrite <- IHfk. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite <- in_app_iff. left.
   rewrite in_map_iff. 
   exists (s0,nv). subst s'.
   split. reflexivity. rewrite IHsk.
   assumption.
   right. rewrite <- in_app_iff. right.
   rewrite IHfk. assumption.

   destruct Hsk as [ssk IHsk].
   destruct Hfk as [sfk IHfk].
   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Incoming, (Join (MConst C.MSG) 
              (Join (MVar v) m)), f) :: s,
             nv)) ssk) ++ sfk).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite <- in_app_iff in REC.
   destruct REC as [REC | InSFK].
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSSK]].
   inversion EQ. subst s nv. clear EQ. apply RECEIVE_AND_RELY.
   rewrite <- IHsk. assumption.
   apply RECEIVE_ALTERNATIVE. rewrite <- IHfk. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite <- in_app_iff. left.
   rewrite in_map_iff. 
   exists (s0,nv). subst s'.
   split. reflexivity. rewrite IHsk.
   assumption.
   right. rewrite <- in_app_iff. right.
   rewrite IHfk. assumption.

   destruct Hsk as [ssk IHsk].
   destruct Hfk as [sfk IHfk].
   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            (((Outgoing, (call_msg p svs)), bf) ::
		  ((Incoming,
		    (Join (MConst C.RETURN)
		      (Join (MVar ai)
			(vars_msg rvs)))), ef) ::
		  s,
             (add (varid ai) nv))) ssk) ++
            (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (call_msg p svs), bf) ::
	     (Incoming, (Join (MConst C.FAIL) (MVar ai)),
               nil) :: s, (add (varid ai) nv))) sfk)
            ++ sfk).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite <- in_app_iff in REC.
   destruct REC as [REC | REC].
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSSK]].
   inversion EQ. subst s nv. clear EQ.
   apply CALL_AND_RELY. rewrite <- IHsk. assumption.
   rewrite <- in_app_iff in REC.
   destruct REC as [REC | InSFK].
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSSK]].
   inversion EQ. subst s nv. clear EQ.
   apply CALLEE_FAILS. rewrite <- IHfk. assumption.
   apply CALL_ALTERNATIVE. rewrite <- IHfk. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite <- in_app_iff.
   left. rewrite in_map_iff. 
   exists (s0,v). subst s' v'.
   split. reflexivity. rewrite IHsk.
   assumption.
   right. rewrite <- in_app_iff.
   right. rewrite <- in_app_iff.
   left. rewrite in_map_iff.
   exists (s0,v). subst s' v'.
   split. reflexivity. rewrite IHfk. assumption.
   right. rewrite <- in_app_iff.
   right. rewrite <- in_app_iff.
   right. rewrite IHfk. assumption.

   destruct Hsk as [ssk IHsk].
   destruct Hfk as [sfk IHfk].
   exists (((faildesc::nil),empty) ::
           (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Outgoing, (Join (MConst C.MATCH) (MVar v)), nil)
              :: (Incoming, (Join (MConst C.MATCH) m), f) :: s,
             nv)) ssk) ++ sfk).
   intros s nv. simpl.
   split. intros [EQ | REC].
   inversion_clear EQ. apply FAIL.
   rewrite <- in_app_iff in REC.
   destruct REC as [REC | InSFK].
   rewrite in_map_iff in REC.
   destruct REC as [[s' nv'] [EQ InSSK]].
   inversion EQ. subst s nv. clear EQ. apply MATCH_AND_RELY.
   rewrite <- IHsk. assumption.
   apply MATCH_ALTERNATIVE. rewrite <- IHfk. assumption.
   intros EC. inversion_clear EC.
   left. reflexivity.
   right. rewrite <- in_app_iff. left.
   rewrite in_map_iff. 
   exists (s0,nv). subst s'.
   split. reflexivity. rewrite IHsk.
   assumption.
   right. rewrite <- in_app_iff. right.
   rewrite IHfk. assumption.
  Qed.

  Definition proc_strands p:
   { ss:list (strand * t) |
   forall (s:strand) (v:t),
   List.In (s,v) ss <->
   eval_proc p s v }.
  Proof.
   intros p.
   destruct p as [n xs psi c].
   destruct (cont_strands c) as [ss Hc].
   exists (map (fun (snv:strand * t) =>
            let (s,nv) := snv in
            ((Incoming, (call_msg n xs), psi) :: s, nv)) ss).
   intros s v. rewrite in_map_iff.
   split. intros [[s' nv'] [EQ InSS]].
   inversion EQ. subst s v. clear EQ. apply PROC.
   rewrite <- Hc. assumption.
   intros EP. inversion EP. 
   exists (s0,v). split.
   subst s'. subst msg. subst n0 xs0.
   reflexivity.
   rewrite Hc.
   assumption.
  Defined.

(* Matchm *)
  Fixpoint matchm cmp_var p m :=
  match (p,m) with
  | (MNil, MNil) => True
  | (MVar (Var pi Msg), _) => True
  | (MVar pv, MVar mv) =>
    cmp_var pv mv
  | (MConst pc, MConst mc) =>
    mc = pc
  | (Join p1 p2, Join m1 m2) =>
    (matchm cmp_var p1 m1) /\ (matchm cmp_var p2 m2)
  | (Hash p1, Hash m1) =>
    matchm cmp_var p1 m1
  | (SymEnc pv p1, SymEnc mv m1) =>
    (cmp_var pv mv) /\ (matchm cmp_var p1 m1)
  | (SymSign pv p1, SymSign mv m1) =>
    (cmp_var pv mv) /\ (matchm cmp_var p1 m1)
  | (PubEnc pv p1, PubEnc mv m1) =>
    (cmp_var pv mv) /\ (matchm cmp_var p1 m1)
  | (PubSign pv p1, PubSign mv m1) =>
    (cmp_var pv mv) /\ (matchm cmp_var p1 m1)
  | (Common.Bind pv p1, Common.Bind mv m1) =>
    (cmp_var pv mv) /\ (matchm cmp_var p1 m1)
  | (Common.Bind pv p1, m) =>
    matchm cmp_var p1 m
  | (_,_) => False
  end.

  Lemma matchm_dec:
   forall (cmpv:var -> var -> Prop),
   (forall (u v:var), {cmpv u v} + {~ cmpv u v}) ->
   forall (p m:msg),
   {matchm cmpv p m} + {~ matchm cmpv p m}.
  Proof.
   intros cmpv cmpv_dec.
   induction p; destruct m; simpl;
   try (intuition; fail);
   try (destruct v as [vi vt]; destruct vt; intuition;
        fail).

   apply const_eq_dec.
   destruct (IHp1 m1); destruct (IHp2 m2); intuition.
   destruct (IHp m); destruct (cmpv_dec v v0); intuition.
   destruct (IHp m); destruct (cmpv_dec v v0); intuition.
   destruct (IHp m); destruct (cmpv_dec v v0); intuition.
   destruct (IHp m); destruct (cmpv_dec v v0); intuition.
   destruct (IHp m); destruct (cmpv_dec v v0); intuition.
  Qed.

  Definition matchm_either cmpvL cmpvR m p :=
  (matchm cmpvL m p) \/ (matchm cmpvR p m).

Hint Unfold matchm_either.

  Lemma matchm_either_dec:
   forall (cmpvL cmpvR:var -> var -> Prop),
   (forall (u v:var), {cmpvL u v} + {~ cmpvL u v}) ->
   (forall (u v:var), {cmpvR u v} + {~ cmpvR u v}) ->
   forall (m p:msg),
   {matchm_either cmpvL cmpvR m p} 
   + {~ matchm_either cmpvL cmpvR m p}.
  Proof.
   intros cmpvL cmpvR cmpvL_dec cmpvR_dec m p.
   unfold matchm_either.
   destruct (matchm_dec cmpvL cmpvL_dec m p).
   left. left. assumption.
   destruct (matchm_dec cmpvR cmpvR_dec p m).
   left. right. assumption.
   right. intros [F | F]; contradiction.
  Qed.

  Definition listP (A:Type) (P:A -> Prop) (l:list A) :=
   exists e:A, P e /\ List.In e l.

Hint Unfold listP.

  Lemma listP_dec:
   forall (A:Type) (P:A -> Prop),
    (forall (e:A), {P e} + {~ P e}) ->
    forall (l:list A),
    {listP A P l} + {~ listP A P l}.
  Proof.
   intros A P P_dec l. unfold listP.
   set (f := (fun e => if P_dec e then true else false)).
   case_eq (existsb f l); intros EQe.
   left. rewrite existsb_exists in EQe.
   destruct EQe as [e [Ine Fe]].
   exists e. split. revert Fe.
   unfold f. case (P_dec e). trivial.
   intros. discriminate.
   assumption.
   right.
   intros [e [Pe InE]].

   destruct (In_split e l InE) as [l1 [l2 EQl]].
   set (n := length l1).
   cut (n < length l). intros LEn.

   generalize (existsb_nth f l e LEn EQe).
   cut (n >= length l1). intros GEn.
   rewrite EQl.
   rewrite (app_nth2 l1 (e::l2) e GEn).
   unfold n. 
   rewrite <- Minus.minus_n_n.
   simpl. unfold f.
   case (P_dec e). intros. discriminate.
   contradiction.
   unfold n. unfold ge. apply le_refl.

   unfold n. rewrite EQl.
   rewrite app_length. simpl.
   rewrite <- Plus.plus_Snm_nSm.
   apply Plus.lt_plus_trans.
   apply Lt.lt_n_Sn.
  Qed.

  Definition list2P (A:Type) (P:A -> A -> Prop) (l1 l2:list A) :=
   exists e1:A, exists e2:A,
    P e1 e2 /\ List.In e1 l1 /\ List.In e2 l2.

Hint Unfold list2P.

  Lemma list2P_dec:
   forall (A:Type) (P:A -> A -> Prop),
    (forall (e1 e2:A), {P e1 e2} + {~ P e1 e2}) ->
    forall (l1 l2:list A),
    {list2P A P l1 l2} + {~ list2P A P l1 l2}.
  Proof.
   intros A P P_dec l1 l2.
   set (P' := (fun e => P (fst e) (snd e))).
   set (l := list_prod l1 l2).
   cut ((forall e, {P' e} + {~ P' e})).
   intros P'_dec.
   case (listP_dec (A*A) P' P'_dec l); intros LP.
   left. unfold list2P, listP in *.
   destruct LP as [[e1 e2] [Pe InE]].
   exists e1. exists e2.
   split. unfold P' in *. simpl in Pe. assumption.
   rewrite <- in_prod_iff. assumption.
   right. unfold list2P, listP in *.
   intros [e1 [e2 [Pe [InE1 InE2]]]].
   apply LP. exists (e1,e2).
   split. unfold P'. simpl. assumption.
   unfold l. rewrite in_prod_iff. intuition.
   intros [e1 e2]. unfold P'. simpl.
   apply P_dec.
  Qed.

  Definition list_share_recvs cmpvL cmpvR l1 l2 :=
   exists m1:msg, exists m2:msg,
    matchm_either cmpvL cmpvR m1 m2
    /\ List.In m1 l1 /\ List.In m2 l2.

Hint Unfold list_share_recvs.

  Lemma list_share_recvs_dec:
   forall (cmpvL cmpvR:var -> var -> Prop),
   (forall (u v:var), {cmpvL u v} + {~ cmpvL u v}) ->
   (forall (u v:var), {cmpvR u v} + {~ cmpvR u v}) ->
   forall (l1 l2:list msg),
   {list_share_recvs cmpvL cmpvR l1 l2}
   + {~ list_share_recvs cmpvL cmpvR l1 l2}.
  Proof.
   intros cmpvL cmpvR cmpvL_dec cmpvR_dec.
   intros l1 l2. unfold list_share_recvs.

   set (P := matchm_either cmpvL cmpvR).
   cut (forall m1 m2, {P m1 m2} + {~ P m1 m2}). intros P_dec.
   case (list2P_dec msg P P_dec l1 l2).
   left. intuition.
   right. intuition.
   unfold P. apply matchm_either_dec; assumption.
  Qed.

  Fixpoint strand_recv_msgs (s:strand) :=
  match s with
  | nil => nil
  | sd :: s =>
    match sd with
    | (Outgoing, m, f) =>		  
      strand_recv_msgs s
    | (Incoming, m, f) =>
      m :: (strand_recv_msgs s)
    end
  end.

  Definition strand_share_recvs cmpvL cmpvR s1 s2 :=
   list_share_recvs cmpvL cmpvR
   (strand_recv_msgs s1) (strand_recv_msgs s2).

Hint Unfold strand_share_recvs.

  Lemma strand_share_recvs_dec:
   forall (cmpvL cmpvR:var -> var -> Prop),
   (forall (u v:var), {cmpvL u v} + {~ cmpvL u v}) ->
   (forall (u v:var), {cmpvR u v} + {~ cmpvR u v}) ->
   forall (s1 s2:strand),
   {strand_share_recvs cmpvL cmpvR s1 s2}
   + {~ strand_share_recvs cmpvL cmpvR s1 s2}.
  Proof.
   intros cmpvL cmpvR cmpvL_dec cmpvR_dec.
   unfold strand_share_recvs.
   intros s1 s2.
   apply list_share_recvs_dec;
   assumption.
  Qed.

  Definition strand_share_recvs_any mk_cmpv f1 f2 ss1 ss2 :=
   exists s1:strand, exists s2:strand,
   exists v1:t, exists v2:t,
   strand_share_recvs (mk_cmpv v1 v2) (mk_cmpv v2 v1)
   (f1 s1) (f2 s2)
   /\ List.In (s1,v1) ss1 /\ List.In (s2,v2) ss2.
   
Hint Unfold strand_share_recvs_any.

  Lemma strand_share_recvs_any_dec:
   forall (f1 f2:strand -> strand)
          (mk_cmpv:t -> t -> var -> var -> Prop),
   (forall (ut vt:t) (u v:var), 
    {mk_cmpv ut vt u v} + {~ mk_cmpv ut vt u v}) ->
   forall (ss1 ss2:list (strand*t)),
   {strand_share_recvs_any mk_cmpv f1 f2 ss1 ss2} 
   + {~ strand_share_recvs_any mk_cmpv f1 f2 ss1 ss2}.
  Proof.
   intros f1 f2 mk_cmpv mk_cmpv_dec ss1 ss2.
   unfold strand_share_recvs_any.

   set (P := fun s1 s2 => 
    strand_share_recvs
     (mk_cmpv (snd s1) (snd s2)) (mk_cmpv (snd s2) (snd s1))
     (f1 (fst s1)) (f2 (fst s2))).
   cut (forall s1 s2, {P s1 s2} + {~ P s1 s2}). intros P_dec.
   case (list2P_dec (strand*t) P P_dec ss1 ss2);
   unfold list2P, P; intros LP.
   left. destruct LP as [[s1 v1] [[s2 v2] P']].
   exists s1. exists s2. exists v1. exists v2.
   intuition.
   right. intros [s1 [s2 [v1 [v2 P']]]].
   apply LP. exists (s1,v1). exists (s2,v2).
   intuition.
   unfold P. intros s1 s2.
   apply strand_share_recvs_dec;
   apply mk_cmpv_dec.
  Qed.

  Theorem strand_share_recvs_proc_dec:
   forall (f1 f2:strand -> strand)
     (mk_cmpv:t -> t -> var -> var -> Prop),
   (forall (ut vt:t) (u v:var), 
    {mk_cmpv ut vt u v} + {~ mk_cmpv ut vt u v}) ->
   forall (p1 p2:proc),
   {exists s1:strand, exists s2:strand,
    exists v1:t, exists v2:t,
    eval_proc p1 s1 v1 
    /\ eval_proc p2 s2 v2 
    /\ strand_share_recvs (mk_cmpv v1 v2)
       (mk_cmpv v2 v1) (f1 s1) (f2 s2)}
   +
   {forall (s1 s2:strand) (v1 v2:IdSet.t),
    eval_proc p1 s1 v1 ->
    eval_proc p2 s2 v2 ->
    ~ strand_share_recvs (mk_cmpv v1 v2)
     (mk_cmpv v2 v1) (f1 s1) (f2 s2)}.
  Proof.
   intros f1 f2 mk_cmpv mk_cmpv_dec p1 p2.
   destruct (proc_strands p1) as [ss1 IH1].
   destruct (proc_strands p2) as [ss2 IH2].
   destruct (strand_share_recvs_any_dec 
             f1 f2 mk_cmpv mk_cmpv_dec ss1 ss2).
 
   left. unfold strand_share_recvs_any in s.
   destruct s as [s1 [s2 [v1 [v2 [SS12 [In1 In2]]]]]].
   exists s1. exists s2. exists v1. exists v2.
   rewrite IH1 in In1. rewrite IH2 in In2.
   split. assumption. split; assumption.

   right. intros s1 s2 v1 v2.
   rewrite <- IH1. rewrite <- IH2. clear IH1 IH2.
   intros In1 In2. intros SS12.
   apply n. clear n.
   unfold strand_share_recvs_any.
   exists s1. exists s2. exists v1. exists v2.
   split. assumption. split; assumption.
  Qed.

  Definition strand_identity (s:strand) := s.
  Definition lift_cmpv (cmpv:var -> var -> Prop) (u:t) (v:t) 
   := cmpv.

Hint Unfold strand_identity.
Hint Unfold lift_cmpv.

  Lemma lift_cmpv_dec:
   forall (cmpv:var -> var -> Prop),
   (forall (u v:var), {cmpv u v} + {~ cmpv u v}) ->
   forall (ut vt:t) (u v:var),
   {lift_cmpv cmpv ut vt u v} + {~ lift_cmpv cmpv ut vt u v}.
  Proof.
   intros cmpv cmpv_dec ut vt u v.
   unfold lift_cmpv. apply cmpv_dec.
  Qed.

  Definition cmpv_type u v :=
  match u with
  | Var _ ut =>
  match v with
  | Var _ vt =>
    ut = vt
  end
  end.

  Lemma cmpv_type_dec:
   forall (u v:var), {cmpv_type u v} + {~ cmpv_type u v}.
  Proof.
   intros u v. destruct u as [ui ut].
   destruct v as [vi vt]. simpl.
   apply type_eq_dec.
  Qed.

  Definition strand_share_recvs_proc_type_dec p1 p2 :=
   strand_share_recvs_proc_dec 
    strand_identity strand_identity
    (lift_cmpv cmpv_type)
    (lift_cmpv_dec cmpv_type cmpv_type_dec)
    p1 p2.
    
  Fixpoint strand_ext_only (s:strand) : strand :=
  match s with
  | nil => nil
  | sd :: s =>
    match sd with
    | (Incoming, m, f) =>
      match m with
      | Join (MConst k) (Join (MVar x) m) =>
        if const_eq_dec k C.MSG then
         sd :: (strand_ext_only s)
        else
         strand_ext_only  s
      | _ =>
        strand_ext_only s
      end
    | (Outgoing, m, f) =>
      match m with
      | Join (MConst k) (Join (MVar x) m) =>
        if const_eq_dec k C.MSG then
         sd :: (strand_ext_only s)
        else
         strand_ext_only  s
      | _ =>
        strand_ext_only s
      end
    end
  end.

  Inductive substrand : strand -> strand -> Prop :=
  | SUBSTRAND_NIL: substrand nil nil
  | SUBSTRAND_EQ: 
    forall (sd:strand_desc) (s1 s2:strand),
    substrand s1 s2 ->
    substrand (sd :: s1) (sd :: s2)
  | SUBSTRAND_SKIP:
    forall (sd:strand_desc) (s1 s2:strand),
    substrand s1 s2 ->
    substrand s1 (sd :: s2).

Hint Constructors substrand.

  Lemma substrand_refl:
   forall (s:strand), substrand s s.
  Proof.
   induction s as [|sd s].
   apply SUBSTRAND_NIL.
   apply SUBSTRAND_EQ.
   assumption.
  Qed.

Hint Resolve substrand_refl.

  Lemma strand_ext_only_sub:
   forall (s:strand),
   substrand (strand_ext_only s) s.
  Proof.
   induction s as [|sd s].
   simpl. apply SUBSTRAND_NIL.
   destruct sd as [[d m] f].
   destruct d; destruct m; simpl;
   try (apply SUBSTRAND_SKIP; assumption);
   destruct m1; simpl;
   try (apply SUBSTRAND_SKIP; assumption);
   destruct m2; simpl;
   try (apply SUBSTRAND_SKIP; assumption);
   destruct m2_1; simpl;
   try (apply SUBSTRAND_SKIP; assumption);
   (case (const_eq_dec c C.MSG); intro CMP;
    [ apply SUBSTRAND_EQ; assumption
    | apply SUBSTRAND_SKIP; assumption]).
  Qed.

Hint Resolve strand_ext_only_sub.

  Definition strand_share_recvs_proc_type_ext_dec p1 p2 :=
   strand_share_recvs_proc_dec 
    strand_ext_only strand_ext_only
    (lift_cmpv cmpv_type)
    (lift_cmpv_dec cmpv_type cmpv_type_dec)
    p1 p2.

  Definition strand_chop_fst_recv (s:strand) : strand :=
  match s with
  | nil => nil
  | sd :: s =>
    match sd with
    | (Incoming, _, _) => s
    | (Outgoing, _, _) => sd :: s
    end
  end.

  Lemma strand_chop_fst_recv_sub:
   forall (s:strand),
   substrand (strand_chop_fst_recv s) s.
  Proof.
   destruct s as [|sd s].
   simpl. apply SUBSTRAND_NIL.
   destruct sd as [[d m] f].
   destruct d; simpl.
   apply substrand_refl.
   apply SUBSTRAND_SKIP.
   apply substrand_refl.
  Qed.

Hint Resolve strand_chop_fst_recv_sub.

  (* Only take it from one side, so it is still compared *)
  Definition strand_share_recvs_proc_type_self_dec p1 :=
   strand_share_recvs_proc_dec 
    (fun s => strand_chop_fst_recv (strand_ext_only s))
    strand_ext_only
    (lift_cmpv cmpv_type)
    (lift_cmpv_dec cmpv_type cmpv_type_dec)	
    p1 p1.

(* DValues *)
  Definition cmpv_dist uds vds :=
  fun u v =>
  match u with
  | Var ui ut =>
  match v with
  | Var vi vt =>
    ~ In ui uds 
    /\ ~ In vi vds 
    /\ ut = vt
  end
  end.

  Lemma cmpv_dist_dec:
   forall (uds vds:t) (u v:var), 
   {cmpv_dist uds vds u v} + {~ cmpv_dist uds vds u v}.
  Proof.
   intros uds vds u v. destruct u as [ui ut].
   destruct v as [vi vt]. simpl.
   destruct (In_dec ui uds).
   right. intros [F H]. contradiction.
   destruct (In_dec vi vds).
   right. intros [F1 [F H]]. contradiction.
   destruct (type_eq_dec ut vt).
   left. split. assumption. split; assumption.
   right. intros [F1 [F2 F3]]. contradiction.
  Qed.

  Definition strand_share_recvs_proc_dist_self_dec p1 :=
   strand_share_recvs_proc_dec 
    (fun s => strand_chop_fst_recv (strand_ext_only s))
    strand_ext_only
    cmpv_dist cmpv_dist_dec
    p1 p1.

  Definition strand_share_recvs_proc_dist_ext_dec p1 p2 :=
   strand_share_recvs_proc_dec 
    strand_ext_only strand_ext_only
    cmpv_dist cmpv_dist_dec
    p1 p2.

  Definition strand_share_recvs_proc_dist_dec p1 p2 :=
   strand_share_recvs_proc_dec 
    strand_identity strand_identity
    cmpv_dist cmpv_dist_dec
    p1 p2.

  Definition cmpv_jdist uds vds :=
  fun u v =>
  match u with
  | Var ui ut =>
  match v with
  | Var vi vt =>
    ~ In ui uds 
    /\ ~ In vi vds 
  end
  end.

  Lemma cmpv_jdist_dec:
   forall (uds vds:t) (u v:var), 
   {cmpv_jdist uds vds u v} + {~ cmpv_jdist uds vds u v}.
  Proof.
   intros uds vds u v. destruct u as [ui ut].
   destruct v as [vi vt]. simpl.
   destruct (In_dec ui uds).
   right. intros [F H]. contradiction.
   destruct (In_dec vi vds).
   right. intros [F1 F]. contradiction.
   left. split; assumption.
  Qed.

  Definition strand_share_recvs_proc_jdist_self_dec p1 :=
   strand_share_recvs_proc_dec 
    (fun s => strand_chop_fst_recv (strand_ext_only s))
    strand_ext_only
    cmpv_jdist cmpv_jdist_dec
    p1 p1.

  Definition strand_share_recvs_proc_jdist_ext_dec p1 p2 :=
   strand_share_recvs_proc_dec 
    strand_ext_only strand_ext_only
    cmpv_jdist cmpv_jdist_dec
    p1 p2.

  Definition strand_share_recvs_proc_jdist_dec p1 p2 :=
   strand_share_recvs_proc_dec 
    strand_identity strand_identity
    cmpv_jdist cmpv_jdist_dec
    p1 p2.

(* Foldm *)
  Fixpoint subst_msg x xm m {struct m} :=
  match m with
  | MNil => MNil
  | MVar v =>
    if var_eq_dec x v then
     Bind x xm 
    else 
     MVar v
  | MConst c => MConst c
  | Join m0 m1 => 
    Join (subst_msg x xm m0) (subst_msg x xm m1)
  | Hash m => Hash m
  | SymEnc v m => SymEnc v (subst_msg x xm m)
  | SymSign v m => SymSign v (subst_msg x xm m)
  | PubEnc v m => PubEnc v (subst_msg x xm m)
  | PubSign v m => PubSign v (subst_msg x xm m)
  | Bind v m =>	  
    if var_eq_dec x v then	    
     Bind v m	    
    else
     Bind v (subst_msg x xm m)	      
  end.

  Lemma subst_msg_type_recv:
   forall (m:msg) (x:var) (xm:msg) (vs:t),
    type_varT x Msg ->
    type_msg_recv vs xm ->
    type_msg_recv vs m ->
    type_msg_recv vs (subst_msg x xm m).
  Proof.
   induction m; simpl; try (intuition; fail);
   intros x xm vs;
   case (var_eq_dec x v); simpl;
   try (intuition; fail).
  Qed.

Hint Resolve subst_msg_type_recv.

  Lemma subst_msg_vars:
   forall (m:msg) (x:var) (xm:msg),
   Subset (message_vars_bound m)
          (message_vars_bound (subst_msg x xm m)).
  Proof.
   induction m; simpl; try (intuition; fail);
   intros x xm.
   case (var_eq_dec x v); simpl.
   intro EQ. rewrite EQ. clear EQ.
   unfold Subset. intro a. set_iff. intuition.
   intro NEQ.
   unfold Subset. intro a. set_iff. intuition.
   unfold Subset. intro a. set_iff. intuition.
   left. apply IHm1. assumption.
   right. apply IHm2. assumption.
   case (var_eq_dec x v); simpl.
   intro EQ. apply subset_refl.
   intro NEQ.
   unfold Subset. intro a. set_iff. intuition.
   right. apply IHm. assumption.
  Qed.

Hint Resolve subst_msg_vars.

(* XXX What about variables that go from matching to binding? *)

  Fixpoint subst_cont tvs sm c :=
  match c with
  | Fail => sm
  | Return f xs => sm
  | Let v lv c => subst_cont tvs sm c
  | Derive f sk fk =>
    let sm' := subst_cont tvs sm sk in
    subst_cont tvs sm' fk
  | Send f v m sk fk =>
    let sm' := subst_cont tvs sm sk in
    subst_cont tvs sm' fk
  | Recv v m f sk fk =>
    let sm' := subst_cont tvs sm sk in
    subst_cont tvs sm' fk
  | Call bf p bs es ef sk fk =>
    let sm' := subst_cont tvs sm sk in
    subst_cont tvs sm' fk
  | Match v m f sk fk => 
    if type_msg_recv_dec tvs m then
     let sm' := subst_msg (Var (varid v) Msg) m sm in
     let sm'' := subst_cont tvs sm' sk in
     subst_cont tvs sm'' fk
    else
     let sm' := subst_cont tvs sm sk in
     subst_cont tvs sm' fk
  end.

  Lemma subst_cont_type:
   forall (c:cont) (sm:msg) (tvs:t),
    type_msg_recv tvs sm ->
    type_msg_recv tvs (subst_cont tvs sm c).
  Proof.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ];
   intros sm vs' Tmr; try (assumption);
   try (apply Hc; assumption);
   try (apply Hsk; assumption);
   try (apply Hfk; assumption);
   try (apply Hfk; apply Hsk; assumption).

   case (type_msg_recv_dec vs' m); intros TMR.
   apply Hfk. apply Hsk.
   apply subst_msg_type_recv.
   simpl. case (type_eq_dec Msg Msg); intuition.
   assumption. assumption.
   apply Hfk; apply Hsk; assumption.
  Qed.

Hint Resolve subst_cont_type.

  Lemma subst_cont_vars:
   forall (c:cont) (m:msg) (vs:t),
   Subset (message_vars_bound m)
          (message_vars_bound (subst_cont vs m c)).
  Proof.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ];
   intros sm vs;
   try (assumption);
   try (apply subset_refl);
   try (apply Hc; assumption);
   try (intros a InSM; 
   apply Hfk; apply Hsk; assumption).

   case (type_msg_recv_dec vs m); intros TMR.
   intros a InSM. apply Hfk. apply Hsk. 
   apply subst_msg_vars. assumption.

   intros a InSM; apply Hfk; apply Hsk; assumption.
  Qed.

Hint Resolve subst_cont_vars.

  Fixpoint foldm_cont vs c {struct c} :=
  match c with
  | Fail => Fail
  | Return f xs => Return f xs
  | Let v lv c =>
    Let v lv (foldm_cont (add (varid v) vs) c)
  | Derive f sk fk => 
    let vs' := union vs (formulas_vars_bound f) in
    Derive f (foldm_cont vs' sk) (foldm_cont vs fk)
  | Send f v m sk fk => 
    let vs' := union vs (formulas_vars_bound f) in
    Send f v m (foldm_cont vs' sk) (foldm_cont vs fk)
  | Recv v m f sk fk =>
    let m' := subst_cont vs m sk in
    let vs' := union (union vs (message_vars_bound m'))
                      (formulas_vars_bound f) in
    Recv v m' f (foldm_cont vs' sk) (foldm_cont vs fk)
  | Call bf n bs es ef sk fk =>
    let vs' := union (union vs (formulas_vars_bound bf))
                      (union (varlist_set es)
                      (formulas_vars_bound ef)) in
    Call bf n bs es ef (foldm_cont vs' sk) (foldm_cont vs fk)
  | Match v m f sk fk => 
    let vs' := union (union vs (message_vars_bound m))
                      (formulas_vars_bound f) in
    Match v m f (foldm_cont vs' sk) (foldm_cont vs fk)
  end.

  Lemma foldm_cont_type:
   forall (c:cont) (vs:t),
   type_cont vs c ->
   type_cont vs (foldm_cont vs c).
  Proof.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ];
   intros vs Tc;
   try (intuition; fail).

   destruct Tc as [EVfvp [EVv [TVc [TMR [TCsk TCfk]]]]].
   unfold exist_vars.
   split. apply subset_trans
   with (union vs (message_vars_bound m)).
   assumption. intro. set_iff. intuition.
   right. apply subst_cont_vars. assumption.
   split. assumption.
   split. assumption.
   split. apply subst_cont_type. assumption.
   split. apply Hsk.
   apply type_cont_subset with
     (union (union vs (message_vars_bound m))
      (formulas_vars_bound f)).
   assumption.
   unfold Subset. intros a. set_iff. intuition.
   left. right. apply subst_cont_vars.
   assumption.
   apply Hfk. assumption.
  Qed.

Hint Resolve foldm_cont_type.

  Lemma foldm_cont_returnpath :
   forall (c:cont) (vs:t),
    return_path c -> return_path (foldm_cont vs c).
  Proof.
   intros c. unfold return_path.
   induction c; simpl; intuition.

   revert IHc1 IHc2 H.
   set (cr1 := cont_returns c1).
   set (cr2 := cont_returns c2).
   rewrite app_length.
   rewrite app_length.
   intros IHc1 IHc2 H.
   destruct (plus_gt_or (length cr1) (length cr2)).
   assumption. apply plus_or_gt.
   left. apply IHc1. assumption.
   apply plus_or_gt. right. apply IHc2. assumption.

   revert IHc1 IHc2 H.
   set (cr1 := cont_returns c1).
   set (cr2 := cont_returns c2).
   rewrite app_length.
   rewrite app_length.
   intros IHc1 IHc2 H.
   destruct (plus_gt_or (length cr1) (length cr2)).
   assumption. apply plus_or_gt.
   left. apply IHc1. assumption.
   apply plus_or_gt. right. apply IHc2. assumption.

   revert IHc1 IHc2 H.
   set (cr1 := cont_returns c1).
   set (cr2 := cont_returns c2).
   rewrite app_length.
   rewrite app_length.
   intros IHc1 IHc2 H.
   destruct (plus_gt_or (length cr1) (length cr2)).
   assumption. apply plus_or_gt.
   left. apply IHc1. assumption.
   apply plus_or_gt. right. apply IHc2. assumption.

   rename c1 into c0.
   rename c2 into c1.
   rename c3 into c2.
   revert IHc1 IHc2 H.
   set (cr1 := cont_returns c1).
   set (cr2 := cont_returns c2).
   rewrite app_length.
   rewrite app_length.
   intros IHc1 IHc2 H.
   destruct (plus_gt_or (length cr1) (length cr2)).
   assumption. apply plus_or_gt.
   left. apply IHc1. assumption.
   apply plus_or_gt. right. apply IHc2. assumption.

   revert IHc1 IHc2 H.
   set (cr1 := cont_returns c1).
   set (cr2 := cont_returns c2).
   rewrite app_length.
   rewrite app_length.
   intros IHc1 IHc2 H.
   destruct (plus_gt_or (length cr1) (length cr2)).
   assumption. apply plus_or_gt.
   left. apply IHc1. assumption.
   apply plus_or_gt. right. apply IHc2. assumption.
  Qed.

Hint Resolve foldm_cont_returnpath.

  Lemma foldm_cont_returns:
   forall (c:cont) (vs:t),
   cont_returns c = cont_returns (foldm_cont vs c).
  Proof.
   intros c. induction c; simpl; intuition.

   rewrite (IHc1 (union vs (formulas_vars_bound l))).
   rewrite (IHc2 vs). reflexivity.

   rewrite (IHc1 (union vs (formulas_vars_bound l))).
   rewrite (IHc2 vs). reflexivity.

   rewrite (IHc1 (union (union vs (message_vars_bound 
                  (subst_cont vs m c1)))
                 (formulas_vars_bound l))).
   rewrite (IHc2 vs). reflexivity.

   rewrite (IHc1 (union (union vs (formulas_vars_bound l))
                   (union (varlist_set l1)
                    (formulas_vars_bound l2)))).
   rewrite (IHc2 vs). reflexivity.

   rewrite (IHc1 (union (union vs (message_vars_bound m))
                   (formulas_vars_bound l))).
   rewrite (IHc2 vs). reflexivity.
  Qed.

Hint Rewrite <- foldm_cont_returns : ppl.

  Lemma foldm_cont_returnssame:
   forall (c:cont) (vs:t),
    returns_same c -> returns_same (foldm_cont vs c).
  Proof.
   intros c. unfold returns_same. intros vs.
   rewrite (foldm_cont_returns c vs).
   trivial.
  Qed.

Hint Resolve foldm_cont_returnssame.

  Definition foldm_proc p :=
  match p with
  | Proc n bs bf c =>
    Proc n bs bf 
    (foldm_cont (union (varlist_set bs)
      (formulas_vars_bound bf)) c)
  end.

  Lemma foldm_proc_type:
   forall (p:proc), type_proc p -> type_proc (foldm_proc p).
  Proof.
   intros p. destruct p as [n bs bf c].
   simpl. intros [Ev [Tc [RPc RSc]]].
   split. assumption.
   split. apply foldm_cont_type. assumption.
   split. apply foldm_cont_returnpath. assumption.
   apply foldm_cont_returnssame. assumption.
  Qed.

Hint Resolve foldm_proc_type.

(* More Specific *)

  Definition cmpv_eq (u:var) v := u = v.

Hint Unfold cmpv_eq.

  Definition cmpv_eq_dec := var_eq_dec.
  Lemma cmpv_eq_eq:
   forall (v:var), cmpv_eq v v.
  Proof.
   unfold cmpv_eq. intros v. reflexivity.
  Qed.

  Definition MsgMoreSpecific m1 m2 :=
   forall (M:msg),
   matchm cmpv_eq m1 M -> matchm cmpv_eq m2 M.

Hint Unfold MsgMoreSpecific.

  Lemma MsgMoreSpecific_refl:
   forall (m:msg), MsgMoreSpecific m m.
  Proof.
   unfold MsgMoreSpecific. trivial.
  Qed.

Hint Resolve MsgMoreSpecific_refl.

  Lemma MsgMoreSpecific_trans:
   forall (m1 m2 m3:msg),
   MsgMoreSpecific m1 m2 ->
   MsgMoreSpecific m2 m3 ->
   MsgMoreSpecific m1 m3.
  Proof.
   unfold MsgMoreSpecific. intuition.
  Qed.

Hint Resolve MsgMoreSpecific_trans.

  Lemma subst_msg_more_specific:
   forall (m:msg) (xi:id) (xm:msg),
   MsgMoreSpecific (subst_msg (Var xi Msg) xm m) m.
  Proof.
   induction m; simpl; intuition;
   try (apply MsgMoreSpecific_refl);
   unfold MsgMoreSpecific in *;
   intuition.

   revert H.
   case (var_eq_dec (Var xi Msg) v).
   intros EQ. subst v.
   simpl. trivial.
   trivial.

   revert H. destruct M; simpl; try (contradiction).
   intuition. apply IHm1 with xi xm. assumption.
   apply IHm2 with xi xm; assumption.

   revert H. destruct M; simpl; try (contradiction).
   intuition. apply IHm with xi xm. assumption.

   revert H. destruct M; simpl; try (contradiction).
   intuition. apply IHm with xi xm. assumption.

   revert H. destruct M; simpl; try (contradiction).
   intuition. apply IHm with xi xm. assumption.

   revert H. destruct M; simpl; try (contradiction).
   intuition. apply IHm with xi xm. assumption.

   revert H.
   case (var_eq_dec (Var xi Msg) v).
   intros EQ. subst v.
   simpl. trivial.
   intros NEQ.
   destruct M; simpl;
   intuition; apply IHm with xi xm;
   assumption.
  Qed.

Hint Resolve subst_msg_more_specific.

  Lemma subst_cont_more_specific:
   forall (c:cont) (m m':msg) (vs:t),
   MsgMoreSpecific m m' ->
   MsgMoreSpecific (subst_cont vs m c) m'.
  Proof.
   induction c; intros sm sm' vs; simpl; intuition.
   case (type_msg_recv_dec vs m); intro TMR; intuition.
   apply IHc2. apply IHc1.
   apply MsgMoreSpecific_trans with sm.
   apply subst_msg_more_specific.
   assumption.
  Qed.

Hint Resolve subst_cont_more_specific.

  Fixpoint cont_more_specific c1 c2 :=
  match (c1,c2) with
  | (Fail, Fail) => True
  | (Return f1 xs1, Return f2 xs2) => f1 = f2 /\ xs1 = xs2
  | (Let v1 lv1 c1, Let v2 lv2 c2) =>
    v1 = v2 /\ lv1 = lv2
    /\ (cont_more_specific c1 c2)
  | (Derive f1 sk1 fk1, Derive f2 sk2 fk2) =>
    f1 = f2
    /\ cont_more_specific sk1 sk2
    /\ cont_more_specific fk1 fk2
  | (Send f1 v1 m1 sk1 fk1, Send f2 v2 m2 sk2 fk2) => 
    f1 = f2 /\ v1 = v2 /\ m1 = m2
    /\ cont_more_specific sk1 sk2
    /\ cont_more_specific fk1 fk2
  | (Recv v1 m1 f1 sk1 fk1, Recv v2 m2 f2 sk2 fk2) =>
    f1 = f2 /\ v1 = v2
    /\ MsgMoreSpecific m1 m2
    /\ cont_more_specific sk1 sk2
    /\ cont_more_specific fk1 fk2
  | (Call bf1 n1 bs1 es1 ef1 sk1 fk1,
     Call bf2 n2 bs2 es2 ef2 sk2 fk2) =>
    bf1 = bf2 /\ n1 = n2 /\ bs1 = bs2 /\ es1 = es2 /\ ef1 = ef2
    /\ cont_more_specific sk1 sk2
    /\ cont_more_specific fk1 fk2
  | (Match v1 m1 f1 sk1 fk1, Match v2 m2 f2 sk2 fk2) => 
    f1 = f2 /\ v1 = v2 /\ m1 = m2
    /\ cont_more_specific sk1 sk2
    /\ cont_more_specific fk1 fk2
  | _ => False
  end.
  
  Lemma foldm_cont_more_specific:
   forall (c:cont) (vs:t),
    cont_more_specific (foldm_cont vs c) c.
  Proof.
   induction c; intros vs; simpl; intuition.
  Qed.

Hint Resolve foldm_cont_more_specific.

  Definition proc_more_specific p1 p2 :=
  match p1 with
  | Proc _ _ _ c1 =>
    match p2 with
    | Proc _ _ _ c2 =>
    cont_more_specific c1 c2
    end
  end.

  Lemma foldm_proc_more_specific:
   forall (p:proc),
    proc_more_specific (foldm_proc p) p.
  Proof.
   intros p. destruct p as [n bs es c].
   simpl. apply foldm_cont_more_specific.
  Qed.

Hint Resolve foldm_proc_more_specific.

 (* XXX Unsound version of foldm that does not preserve 
        well-formed on recv?. This would fold back matches
        that are not well-formed strictly, but would be if
        pattern-matching were recursive *)
