(* CPPLlb *)
Require Export CPPLc.
Open Scope list_scope .

(* Msg Redact *)
  Fixpoint msg_redact_recv (vs:t) (m:msg) :=
  match m with
  | MNil => MNil
  | MVar v => MVar v
  | MConst c => MConst c
  | Join m0 m1 => 
    Join (msg_redact_recv vs m0) (msg_redact_recv vs m1)
  | Hash m =>
    if type_msg_recv_dec vs (Hash m) then
     Hash m
    else
     let n := new_id tt in
     MVar (Var n Msg)
  | SymEnc v m =>
    if mem (varid v) vs then
     SymEnc v (msg_redact_recv vs m)
    else
     let n := new_id tt in
     MVar (Var n Msg)
  | SymSign v m =>
    if mem (varid v) vs then
     SymSign v (msg_redact_recv vs m)
    else
     let n := new_id tt in
     MVar (Var n Msg)
  | PubEnc v m =>
    if mem (varid v) vs then
     PubEnc v (msg_redact_recv vs m)
    else
     let n := new_id tt in
     MVar (Var n Msg)
  | PubSign v m =>
    if mem (varid v) vs then
     PubSign v (msg_redact_recv vs m)
    else
     let n := new_id tt in
     MVar (Var n Msg)
  | Bind v m =>	Bind v (msg_redact_recv vs m)
  end.

  Lemma msg_redact_recv_type_recv:
   forall (vs:t) (m:msg),
    type_msgT m ->
    type_msg_recv vs (msg_redact_recv vs m).
  Proof.
   intros vs m. revert vs.
   induction m; simpl; intuition;
   (* Hash *)
   try (case (type_msg_recv_dec vs (Hash m)); simpl; trivial; fail);
   (* Keys *)
   case_eq (mem (varid v) vs); simpl; trivial;
   rewrite <- mem_iff; unfold exist_var;
   intuition.
  Qed.

Hint Resolve msg_redact_recv_type_recv.

  Lemma MsgMoreSpecific_var:
   forall (i:id) (m:msg),
    MsgMoreSpecific m (MVar (Var i Msg)).
  Proof.
   unfold MsgMoreSpecific.
   intros i m M.
   simpl. trivial.
  Qed.

Hint Resolve MsgMoreSpecific_var.

  Lemma msg_redact_recv_specific:
   forall (vs:t) (m:msg),
    MsgMoreSpecific m (msg_redact_recv vs m).
  Proof.
   intros vs m. revert vs.
   induction m; simpl; intros vs;
   try (apply MsgMoreSpecific_refl).
   (* Join *)
   generalize (IHm1 vs). clear IHm1.
   generalize (IHm2 vs). clear IHm2.
   unfold MsgMoreSpecific. simpl.
   destruct M; intuition.
   (* Hash *)
   case (type_msg_recv_dec vs (Hash m)); simpl; intros TMR.
   apply MsgMoreSpecific_refl.
   apply MsgMoreSpecific_var.
   (* Keys *)
   case (mem (varid v) vs); simpl; [ idtac | apply MsgMoreSpecific_var ].
   unfold MsgMoreSpecific. intros M.
   simpl. destruct M; intuition.
   apply IHm. assumption.
   case (mem (varid v) vs); simpl; [ idtac | apply MsgMoreSpecific_var ].
   unfold MsgMoreSpecific. intros M.
   simpl. destruct M; intuition.
   apply IHm. assumption.
   case (mem (varid v) vs); simpl; [ idtac | apply MsgMoreSpecific_var ].
   unfold MsgMoreSpecific. intros M.
   simpl. destruct M; intuition.
   apply IHm. assumption.
   case (mem (varid v) vs); simpl; [ idtac | apply MsgMoreSpecific_var ].
   unfold MsgMoreSpecific. intros M.
   simpl. destruct M; intuition.
   apply IHm. assumption.
   (* Bind *)
   unfold MsgMoreSpecific. intros M.
   simpl. destruct M; intuition;
   apply IHm; assumption.
  Qed.

Hint Resolve msg_redact_recv_specific.

  Definition MsgRedact vs m rm :=
   MsgMoreSpecific m rm
   /\ (type_msgT m -> type_msg_recv vs rm).

Hint Unfold MsgRedact.

  Theorem msg_redact_recv_iff:
   forall (vs:t) (m:msg),
    MsgRedact vs m (msg_redact_recv vs m).
  Proof.
   intros vs m. split.
   apply msg_redact_recv_specific.
   apply msg_redact_recv_type_recv.
  Qed.

Hint Resolve msg_redact_recv_iff.
  
  Lemma msg_redact_recv_identity:
   forall (vs:t) (m:msg),
   type_msg_recv vs m ->
   msg_redact_recv vs m = m.
  Proof.
   intros vs m.
   induction m; simpl; intuition; unfold exist_var in *.
   rewrite H. rewrite H2. reflexivity.
   case (type_msg_recv_dec vs (Hash m)); simpl; intuition.
   case_eq (mem (varid v) vs).
   rewrite H1. reflexivity.
   rewrite <- not_mem_iff. contradiction.
   case_eq (mem (varid v) vs).
   rewrite H1. reflexivity.
   rewrite <- not_mem_iff. contradiction.
   case_eq (mem (varid v) vs).
   rewrite H1. reflexivity.
   rewrite <- not_mem_iff. contradiction.
   case_eq (mem (varid v) vs).
   rewrite H1. reflexivity.
   rewrite <- not_mem_iff. contradiction.
   rewrite H. reflexivity.
  Qed.

Hint Rewrite -> msg_redact_recv_identity : ppl.

  Lemma msg_redact_recv_necc:
   forall (vs:t) (m:msg) (vs':t),
    Subset (message_vars_match m) vs' ->
    msg_redact_recv vs m = msg_redact_recv (inter vs vs') m.
  Proof.
   intros vs m.
   induction m; simpl; 
   intros vs' SUB12;
   intuition; unfold exist_var in *.

   rewrite <- IHm1. rewrite <- IHm2. reflexivity.
   intros a In2. apply SUB12. set_iff. right. assumption.
   intros a IN. apply SUB12. set_iff. left. assumption.

   case (type_msg_recv_dec vs (Hash m)); intros TM;
   case (type_msg_recv_dec (inter vs vs') (Hash m)); intros TM'.
   reflexivity.
   assert False. apply TM'. rewrite type_msg_recv_iff.
   rewrite type_msg_recv_iff in TM. destruct TM as [TMs TMt].
   simpl. split.
   intros a. set_iff. intro IN. split.
   apply TMs. assumption. apply SUB12. assumption.
   simpl in TMt. assumption.
   contradiction.
   assert False. apply TM. rewrite type_msg_recv_iff.
   rewrite type_msg_recv_iff in TM'. destruct TM' as [TM's TM't].
   split. simpl.
   intros a. set_iff. intros IN.
   simpl in TM's. generalize (TM's a IN).
   set_iff. intuition. assumption.
   contradiction.
   reflexivity.

   case_eq (mem (varid v) vs); intros INvs;
   case_eq (mem (varid v) (inter vs vs')); intros INvs'.
   rewrite (IHm vs'). reflexivity.
   intros a IN. apply SUB12. set_iff. right. assumption.
   assert False. 
   rewrite <- not_mem_iff in INvs'. apply INvs'.
   set_iff. split. rewrite mem_iff. assumption.
   apply SUB12. set_iff. left. reflexivity.
   contradiction.
   assert False. 
   rewrite <- not_mem_iff in INvs. apply INvs.
   rewrite <- mem_iff in INvs'. rewrite inter_iff in INvs'.
   intuition.
   contradiction.
   reflexivity.

   case_eq (mem (varid v) vs); intros INvs;
   case_eq (mem (varid v) (inter vs vs')); intros INvs'.
   rewrite (IHm vs'). reflexivity.
   intros a IN. apply SUB12. set_iff. right. assumption.
   assert False. 
   rewrite <- not_mem_iff in INvs'. apply INvs'.
   set_iff. split. rewrite mem_iff. assumption.
   apply SUB12. set_iff. left. reflexivity.
   contradiction.
   assert False. 
   rewrite <- not_mem_iff in INvs. apply INvs.
   rewrite <- mem_iff in INvs'. rewrite inter_iff in INvs'.
   intuition.
   contradiction.
   reflexivity.

   case_eq (mem (varid v) vs); intros INvs;
   case_eq (mem (varid v) (inter vs vs')); intros INvs'.
   rewrite (IHm vs'). reflexivity.
   intros a IN. apply SUB12. set_iff. right. assumption.
   assert False. 
   rewrite <- not_mem_iff in INvs'. apply INvs'.
   set_iff. split. rewrite mem_iff. assumption.
   apply SUB12. set_iff. left. reflexivity.
   contradiction.
   assert False. 
   rewrite <- not_mem_iff in INvs. apply INvs.
   rewrite <- mem_iff in INvs'. rewrite inter_iff in INvs'.
   intuition.
   contradiction.
   reflexivity.

   case_eq (mem (varid v) vs); intros INvs;
   case_eq (mem (varid v) (inter vs vs')); intros INvs'.
   rewrite (IHm vs'). reflexivity.
   intros a IN. apply SUB12. set_iff. right. assumption.
   assert False. 
   rewrite <- not_mem_iff in INvs'. apply INvs'.
   set_iff. split. rewrite mem_iff. assumption.
   apply SUB12. set_iff. left. reflexivity.
   contradiction.
   assert False. 
   rewrite <- not_mem_iff in INvs. apply INvs.
   rewrite <- mem_iff in INvs'. rewrite inter_iff in INvs'.
   intuition.
   contradiction.
   reflexivity.

   rewrite (IHm vs' SUB12). reflexivity.
  Qed.

Hint Rewrite -> msg_redact_recv_necc : ppl.

  Lemma redact_necc:
   forall (cmpv:var -> var -> Prop) (m1:msg) (m2:msg) (vs:t),
    ~ matchm cmpv (msg_redact_recv vs m1) (msg_redact_recv vs m2) <->
    let vs' := inter vs 
               (union (message_vars_match m1) (message_vars_match m2)) in
    ~ matchm cmpv (msg_redact_recv vs' m1) (msg_redact_recv vs' m2).
  Proof.
   intros cmpv m1 m2 vs.
   simpl.
   set (vs1 := message_vars_match m1).
   set (vs2 := message_vars_match m2).
   set (vs' := union vs1 vs2).

   rewrite (msg_redact_recv_necc vs m1 vs').
   rewrite (msg_redact_recv_necc vs m2 vs').
   intuition.
   unfold vs', vs1, vs2. intros a. set_iff. intuition.
   unfold vs', vs1, vs2. intros a. set_iff. intuition.
  Qed.

Hint Resolve redact_necc.

  Lemma msg_redact_recv_compat:
   forall (m:msg) (x y:t),
    Equal x y ->
    msg_redact_recv x m = msg_redact_recv y m.
  Proof.
   intros m x y EQx.
   induction m; simpl.
   reflexivity.
   reflexivity.
   reflexivity.
   rewrite IHm1. rewrite IHm2. reflexivity.

   case (type_msg_recv_dec x (Hash m)).
   case (type_msg_recv_dec y (Hash m)). reflexivity.
   rewrite EQx. contradiction.
   case (type_msg_recv_dec y (Hash m)).
   rewrite EQx. contradiction.
   reflexivity.
   
   rewrite IHm. clear IHm.
   case_eq (mem (varid v) x).
   case_eq (mem (varid v) y). reflexivity.
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   case_eq (mem (varid v) y).
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   reflexivity.

   rewrite IHm. clear IHm.
   case_eq (mem (varid v) x).
   case_eq (mem (varid v) y). reflexivity.
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   case_eq (mem (varid v) y).
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   reflexivity.

   rewrite IHm. clear IHm.
   case_eq (mem (varid v) x).
   case_eq (mem (varid v) y). reflexivity.
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   case_eq (mem (varid v) y).
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   reflexivity.

   rewrite IHm. clear IHm.
   case_eq (mem (varid v) x).
   case_eq (mem (varid v) y). reflexivity.
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   case_eq (mem (varid v) y).
   rewrite EQx. intros M. rewrite M. intros N. discriminate N.
   reflexivity.

   rewrite IHm. reflexivity.
  Qed.

Hint Resolve msg_redact_recv_compat.

(* Subsets *)

 Module SS := FSetList.Make(IdSet.IdSet).
 Module SSEP := FSetEqProperties.EqProperties(SS).
 Module SSP := SSEP.MP.
 Module SSF := SSP.FM.

  Definition Subsets s ss :=
   forall s':t,
    Subset s' s <-> SS.In s' ss.

Hint Unfold Subsets.

  Definition SSMap f s :=
  SS.fold (fun e a => SS.add (f e) a) s SS.empty.

  Lemma SSMap_AddIn:
   forall (ss:SS.t) (y:t) (e:elt),
    SS.In y (SSMap (fun (s:t) => add e s) ss) <->
    (exists x:t, Equal (add e x) y /\ SS.In x ss).
  Proof.
   intros ss y e. unfold SSMap.
   set (f := fun (s:SS.elt) a => SS.add (add e s) a).

   induction ss using SSP.set_induction.
   generalize (SSP.empty_is_empty_1 H).
   intros EQ.
   generalize (SSP.fold_1 SSF.Equal_ST SS.empty
               f H).
   intros SSEQ. rewrite SSEQ. clear SSEQ.
   SSF.set_iff.
   split. contradiction.
   intros [x [EQx Inx]].
   revert Inx. rewrite EQ. SSF.set_iff.
   trivial.

   cut (compat_op eq SS.Equal f). intros f_compat.
   cut (transpose SS.Equal f). intros f_transpose.
   rename H into NIn.
   rename H0 into ADD.
   generalize (SSP.fold_2 SSF.Equal_ST SS.empty
               f_compat f_transpose
               NIn ADD).
   intros SSEQ. rewrite SSEQ.
   clear SSEQ f_compat f_transpose.
   unfold f at 1.
   SSF.set_iff.
   rewrite IHss1.
   unfold SSP.Add in *.
   split.
   intros [EQy | [x' [EQx' Inx']]].
   exists x. split. assumption.
   rewrite ADD. left. apply Raw.eq_refl. 
   exists x'. split. assumption.
   rewrite ADD. right. assumption.
   intros [x' [EQx' Inx']].
   rewrite ADD in Inx'.
   destruct Inx' as [EQx | Inx'].
   left.
   assert (Equal x x') as EQLx. assumption.
   assert (Equal (add e x) y).
   rewrite EQLx. rewrite EQx'. reflexivity.
   assumption.
   right. exists x'. split; assumption.

   clear IHss1 H0 H ss1 ss2 x y f_compat.
   unfold f. clear f.
   unfold transpose.
   intros x y ss a.
   SSF.set_iff. intuition.

   clear IHss1 H0 H ss1 ss2 x y.
   unfold f. clear f.
   unfold compat_op.
   intros x x' y y' EQx EQy a.
   SSF.set_iff.
   rewrite EQy. clear EQy y.
   assert (Equal x x') as EQx'.
   assumption.
   intuition. left.
   assert (Equal (add e x') a).
   rewrite <- EQx'. assumption.
   assumption.
   left.
   assert (Equal (add e x) a). rewrite EQx'. assumption.
   assumption.
  Qed.

Hint Rewrite -> SSMap_AddIn : ppl.

  Lemma SSMap_AddInE:
   forall (ss:SS.t) (y:t) (e:elt),
    SS.In y (SSMap (fun (s:t) => add e s) ss) ->
    In e y.
  Proof.
   intros ss y e.
   rewrite SSMap_AddIn.
   intros [x [EQx Inx]].
   rewrite <- EQx.
   set_iff. left. reflexivity.
  Qed.

Hint Resolve SSMap_AddInE.

  Definition subset_cmb e ss :=
   SS.union ss (SSMap (fun (s:t) => add e s) ss).

  Definition subsets s :=
  fold subset_cmb s (SS.singleton empty).

  Lemma subset_cmb_compat:
   compat_op E.eq SS.Equal subset_cmb.
  Proof.
   unfold compat_op, subset_cmb.
   intros x x' ss ss' EQx EQss a.
   SSF.set_iff.
   rewrite SSMap_AddIn.
   rewrite SSMap_AddIn.
   split; intros [Ina | [a' [EQa' Ina']]].
   left. rewrite <- EQss. intuition.
   right. exists a'. rewrite <- EQss.
   rewrite <- EQx. split; assumption.
   left. rewrite EQss. assumption.
   right. exists a'. rewrite EQss. rewrite EQx. split; assumption.
  Qed.

Hint Resolve subset_cmb_compat.

  Lemma subset_cmb_transpose:
   transpose SS.Equal subset_cmb.
  Proof.
   unfold transpose.
   intros x y ss.
   unfold subset_cmb.
   intros a.
   SSF.set_iff.
   rewrite SSMap_AddIn.
   rewrite SSMap_AddIn.
   rewrite SSMap_AddIn.
   rewrite SSMap_AddIn.

   split; intros [[Ina | [a' [Eqa' Ina']]] | [a' [Eqa' Ina']]].
   left. left. assumption.

   right. exists a'. split. assumption.
   SSF.set_iff. left. assumption.

   revert Ina'. SSF.set_iff. rewrite SSMap_AddIn.
   intros [Ina' | [a'' [Eqa'' Ina'']]].
   left. right. exists a'. split; assumption.
   rewrite <- Eqa'' in Eqa'. clear Eqa'' a'.
   case (In_dec y a''); intros InY.
   cut (Equal (add y a'') a''); intros EqY.
   rewrite EqY in *. left. right. exists a''. split; assumption.
   apply (add_equal InY).
   rewrite add_add in Eqa'.
   case (In_dec x a''); intros InX.
   rewrite (add_equal InX) in Eqa'.
   right. exists a''.
   split. assumption.
   SSF.set_iff. left. assumption.
   right. exists (add x a'').
   split. assumption.
   SSF.set_iff. right.
   rewrite SSMap_AddIn.
   exists a''. split. reflexivity.
   assumption.

   left. left. assumption.

   right. exists a'.
   split. assumption.
   SSF.set_iff. left. assumption.

   revert Ina'. SSF.set_iff. rewrite SSMap_AddIn.
   intros [Ina' | [a'' [Eqa'' Ina'']]].
   left. right. exists a'. split; assumption.
   rewrite <- Eqa'' in Eqa'. clear Eqa'' a'.
   case (In_dec x a''); intros InX.
   rewrite (add_equal InX) in Eqa'.
   left. right. exists a''. split; assumption.
   rewrite add_add in Eqa'.
   case (In_dec y a''); intros InY.
   rewrite (add_equal InY) in Eqa'.
   right. exists a''.
   split. assumption.
   SSF.set_iff. left. assumption.
   right. exists (add y a'').
   split. assumption.
   SSF.set_iff. right.
   rewrite SSMap_AddIn.
   exists a''. split. reflexivity.
   assumption.
  Qed.

Hint Resolve subset_cmb_transpose.
  
  Lemma subsets_1:
   forall (s s':t),
   SS.In s' (subsets s) -> Subset s' s.
  Proof.
   intros s1. unfold Subsets.

   induction s1 using set_induction;
   intros q.
   generalize (empty_is_empty_1 H).
   intros EQ.
   generalize (fold_1 SSF.Equal_ST (SS.singleton empty)
               subset_cmb H).
   intros SSEQ. rewrite SSEQ. clear SSEQ.
   rewrite EQ. clear EQ H.
   SSF.set_iff. unfold Subset.
   intros EQ a Inq.
   assert (Equal empty q) as EQL by assumption.
   rewrite EQL. assumption.

   rename H into NIn.
   rename H0 into ADD.
   unfold subsets.
   generalize (fold_2 SSF.Equal_ST (SS.singleton empty)
               subset_cmb_compat subset_cmb_transpose
               NIn ADD).
   intros SSEQ. rewrite SSEQ. clear SSEQ.
   unfold subset_cmb at 1.
   SSF.set_iff.
   rewrite SSMap_AddIn.
 
   assert (Subset s1_1 s1_2) as SUB12.
   intros a. unfold Add in *. rewrite ADD.
   intros Ina. right. assumption.

   assert (In x s1_2) as InX2.
   unfold Add in *. rewrite ADD.
   left. reflexivity.

   case (Equal_dec s1_1 s1_2); intros EQ12.
   rewrite EQ12 in *. contradiction.

   case (Equal_dec q s1_2); intros EQq2.

   intros [Subq | [p [EQp InP]]].
   assert False. apply NIn.
   apply (IHs1_1 q Subq x).
   rewrite EQq2. assumption.
   contradiction.
   generalize (IHs1_1 p InP). clear InP.
   rewrite <- EQp in *. clear EQp q.
   rewrite EQq2. reflexivity.

   intros [Subq | [p [EQp InP]]].
   generalize (IHs1_1 q Subq). clear Subq.
   intros Subq. apply subset_trans with s1_1; assumption.
   generalize (IHs1_1 p InP). clear InP.
   rewrite <- EQp in *. clear EQp q.
   intros Subp a. set_iff.
   intros [EQx | InP].
   rewrite EQx in *. assumption.
   apply SUB12. apply Subp. assumption.
  Qed.

Hint Resolve subsets_1.

  Lemma subsets_2:
   forall (s s':t),
   Subset s' s -> SS.In s' (subsets s).
  Proof.
   intros s1. unfold Subsets.

   induction s1 using set_induction;
   intros q.
   generalize (empty_is_empty_1 H).
   intros EQ.
   generalize (fold_1 SSF.Equal_ST (SS.singleton empty)
               subset_cmb H).
   intros SSEQ. rewrite SSEQ. clear SSEQ.
   rewrite EQ. clear EQ H.
   SSF.set_iff. unfold Subset.
   intros SUB.
   assert (Empty q).
   unfold Empty. intros a IN.
   rewrite <- (empty_iff a).
   apply SUB. assumption.
   generalize (empty_is_empty_1 H).
   intros EQ. assert (Equal empty q).
   rewrite EQ. reflexivity. assumption.

   rename H into NIn.
   rename H0 into ADD.
   intros SUBq2.
   unfold subsets.
   generalize (fold_2 SSF.Equal_ST (SS.singleton empty)
               subset_cmb_compat subset_cmb_transpose
               NIn ADD).
   intros SSEQ. rewrite SSEQ. clear SSEQ.
   unfold subset_cmb at 1.
   SSF.set_iff.
   rewrite SSMap_AddIn.
 
   assert (Subset s1_1 s1_2) as SUB12.
   intros a. unfold Add in *. rewrite ADD.
   intros Ina. right. assumption.

   assert (In x s1_2) as InX2.
   unfold Add in *. rewrite ADD.
   left. reflexivity.

   case (Equal_dec s1_1 s1_2); intros EQ12.
   rewrite EQ12 in *. contradiction.

   case (Subset_dec q s1_1); intros SUBq1.
   left. apply IHs1_1. assumption.
   right. 
   case (Equal_dec q s1_2); intros EQq2.
   exists s1_1. rewrite EQq2 in *. clear EQq2 q.
   split.
   intros a. unfold Add in ADD. rewrite ADD.
   set_iff. intuition.
   apply IHs1_1. reflexivity.

   case (In_dec x q); intros InXq.
   exists (remove x q).
   split. apply (add_remove InXq).
   apply IHs1_1. intros a.
   set_iff. intros [InAq NEQ].
   generalize (SUBq2 a InAq).
   unfold Add in ADD. rewrite ADD.
   intros [EQ | InA]. contradiction.
   assumption.

   assert False.
   apply SUBq1. intros a.
   intros InAq.
   generalize (SUBq2 a InAq).
   unfold Add in ADD. rewrite ADD.
   intros [EQ | InA].
   rewrite EQ in *. contradiction.
   assumption.
   contradiction.
  Qed.

Hint Resolve subsets_2.

  Theorem subsets_iff:
   forall (s:t),
    Subsets s (subsets s).
  Proof.
   intros s. unfold Subsets.
   intros s'. split.
   apply subsets_2.
   apply subsets_1.
  Qed.

(* Min Set *)

 Import Arith.Compare_dec.
 
  Definition minset s :=
   SS.min_elt s.

Hint Unfold minset.

  Theorem minset_none_iff:
   forall (s:SS.t),
    minset s = None -> SS.Empty s.
  Proof.
   unfold minset. apply SS.min_elt_3.
  Qed.

Hint Resolve minset_none_iff.

  Theorem minset_some_iff:
   forall (s:SS.t) (x:t),
    minset s = Some x -> 
    SS.In x s /\ 
    forall (y:t),
     SS.In y s -> ~ SS.E.lt y x.
  Proof.
   intros s x.
   unfold minset. intros ME.
   split. apply SS.min_elt_1. assumption.
   intros y InY. apply (SS.min_elt_2 ME InY).
  Qed. 

Hint Resolve minset_some_iff.

(* Set opt *)
  Theorem setopt:
   forall (P:t -> Prop),
   (forall (s:t), {P s} + {~ P s}) ->
   (forall (x y:t), Equal x y -> (P x <-> P y)) ->
   forall (M:t),
   P M ->
   (forall (s:t), P s -> P (inter s M)) ->
   { s:t |
     P s
     /\ (forall (s':t), 
          let s'' := inter s' M in
          P s' -> (SS.E.eq s s'' \/ SS.E.lt s s'')) }.
  Proof.
   intros P P_dec P_compat M PM P_necc.

   set (subM := subsets M).
   generalize (subsets_iff M). intros SUB.
   unfold Subsets in SUB.
   set (f := (fun s:t => if P_dec s then true else false)).
   cut (compat_bool Equal f). intros f_compat.
   set (rM := SS.filter f subM).
   set (mrM := minset rM).
   case_eq mrM.
   intros s EQm.
   destruct (minset_some_iff rM s EQm) as [InRM MinM].
   rewrite SSF.mem_iff in InRM.
   unfold rM in InRM.
   rewrite (SSF.filter_b subM s f_compat) in InRM.
   generalize (andb_prop (SS.mem s subM) (f s) InRM).
   clear InRM. intros [InS FT]. revert FT.
   unfold f at 1. case (P_dec s). intros Ps.
   intros F. clear F.

   exists s. split. assumption.
   intros s' s'' Ps'.
   case (Equal_dec s'' s). intuition.
   intros NEQs''. right.
   unfold SS.E.lt in *.
   apply SSP.ME.le_neq.
   apply MinM.
   unfold rM. apply SS.filter_3.
   assumption.
   rewrite <- SUB.
   unfold s''.
   apply inter_subset_2.
   unfold f.
   case (P_dec s''). reflexivity.
   generalize (P_necc s' Ps'). unfold s''.
   contradiction.
   assumption.

   intros. discriminate.

   intros EqmrM.
   generalize (minset_none_iff rM EqmrM).
   unfold rM. intros EM.
   assert False. 
   unfold SS.Empty in EM. 
   apply (EM M).
   apply SS.filter_3. assumption.
   rewrite <- SUB. reflexivity.
   unfold f.
   case (P_dec M). reflexivity.
   contradiction.
   contradiction.
   
   unfold compat_bool. clear SUB subM M PM P_necc.
   unfold f. intros x y EQx.
   case (P_dec x); case (P_dec y).
   reflexivity.
   rewrite (P_compat x y EQx). contradiction.
   rewrite (P_compat x y EQx). contradiction.
   reflexivity.
  Qed.

(* Redact Opt *)
  Theorem redact_opt:
   forall (cmpv:var -> var -> Prop) (m1:msg) (m2:msg),
   (forall (u v:var), {cmpv u v} + {~ cmpv u v}) ->
   type_msgT m1 -> type_msgT m2 ->
   ~ matchm cmpv m1 m2 ->
   { vs:t | 
     ~ matchm cmpv (msg_redact_recv vs m1) (msg_redact_recv vs m2)
     /\ (forall (vs':t),
         let vs'' := (inter vs' (union (message_vars_match m1)
                             (message_vars_match m2))) in
         ~ matchm cmpv (msg_redact_recv vs' m1) (msg_redact_recv vs' m2)
         -> (SS.E.eq vs vs'' \/ SS.E.lt vs vs'')) }.
  Proof.
   intros cmpv m1 m2 cmpv_dec Tm1 Tm2 MM.
   
   set (P := fun vs => 
         ~ matchm cmpv (msg_redact_recv vs m1) (msg_redact_recv vs m2)).
   cut ((forall (s:t), {P s} + {~ P s})). intros P_dec.
   cut ((forall (x y:t), Equal x y -> (P x <-> P y))). intros P_eq.
   set (vs1 := message_vars_match m1).
   set (vs2 := message_vars_match m2).
   set (M := union vs1 vs2).
   cut (P M). intros PM.
   cut (forall (s:t), P s -> P (inter s M)). intros Pnecc.
   destruct (setopt P P_dec P_eq M PM Pnecc) as [vs [PS MIN]].

   exists vs. split.
   assumption. intros vs' vs''.
   intros Pvs'. apply (MIN vs' Pvs').
   
   intros vs. unfold P. rewrite (redact_necc cmpv m1 m2).
   simpl. intuition.

   unfold P.
   rewrite msg_redact_recv_identity.
   rewrite msg_redact_recv_identity.
   assumption.
   rewrite type_msg_recv_iff.
   split. unfold M. apply union_subset_2. assumption.
   rewrite type_msg_recv_iff.
   split. unfold M. apply union_subset_1. assumption.
   
   intros x y EQx. unfold P.
   rewrite (msg_redact_recv_compat m1 x y EQx).
   rewrite (msg_redact_recv_compat m2 x y EQx).
   intuition.

   unfold P. intros vs.
   case (matchm_dec cmpv cmpv_dec
         (msg_redact_recv vs m1) (msg_redact_recv vs m2));
   intuition.
  Qed.

(* Redactl *)
  Definition strand_msgs (s:strand) : (list msg) :=
  List.map (fun s => snd (fst s)) s.

  Definition strand_redact vs (s:strand) :=
   List.map (fun s =>
    match s with ((d,m),f) => ((d, (msg_redact_recv vs m)), f) end)
    s.

  Definition messages_vars_match ms :=
  List.fold_right (fun e a => union a (message_vars_match e)) empty ms.

  Definition initial_set_self s :=
   messages_vars_match 
    (strand_msgs (strand_chop_fst_recv (strand_ext_only s))).

  Program Definition initial_set_selfp p :=
   List.fold_right (fun e a => union a (initial_set_self (fst e)))
    empty
    (proc_strands p).

  Definition strand_share_recvs_proc_dist_self_redact_dec vs p :=
   strand_share_recvs_proc_dec 
    (fun s => 
      strand_redact vs (strand_chop_fst_recv (strand_ext_only s)))
    (fun s =>
      strand_redact vs (strand_ext_only s))
    cmpv_dist cmpv_dist_dec
    p p.

Definition initial_set_pair p1 p2 :=
 union (initial_set_selfp p1) (initial_set_selfp p2).

Definition strand_share_recvs_proc_dist_redact_dec vs p1 p2 :=
 strand_share_recvs_proc_dec 
  (fun s => 
    strand_redact vs (strand_ext_only s))
  (fun s =>
    strand_redact vs (strand_ext_only s))
  cmpv_dist cmpv_dist_dec
  p1 p2.

(* Modified protocol that shares some values *)
Definition share v : strand_desc :=
(Outgoing, (Join (MConst C.SHARE) (MVar v)), nil).

Fixpoint strand_share v s :=
match s with
| nil => nil
| sd :: s =>
  let sd_vs := strand_desc_vars_bound sd in
  if In_dec (varid v) sd_vs then
   sd :: (share v) :: s
  else
   sd :: (strand_share v s)
end.

Theorem strand_share_type:
 forall (v:var) (s:strand) (vs:t),
 type_strand vs s ->
 type_strand vs (strand_share v s).
Proof.
 intros v s.
 induction s as [|sd s]; simpl.
 trivial.
 intros vs.
 intros [Tsd Ts].
 case (In_dec (varid v) (strand_desc_vars_bound sd));
 intros INv; simpl; intuition.
 unfold exist_var. set_iff.
 intuition.
Qed.

Hint Resolve strand_share_type.

Definition strand_shares vs s :=
List.fold_right (fun e a => strand_share e a) s vs.

Theorem strand_shares_type:
 forall (V:list var) (s:strand) (vs:t),
 type_strand vs s ->
 type_strand vs (strand_shares V s).
Proof.
 intros V s.
 induction V as [|v V]; simpl; intuition.
Qed.

Hint Resolve strand_shares_type.

Definition is_shared (sd:strand_desc) :=
match sd with
| (Outgoing, (Join (MConst c) (MVar v)), nil) =>
 if const_eq_dec c C.SHARE then
  true
 else
  false
| _ =>
 false
end.

Definition shared_only := List.filter is_shared.
Definition not_shared_only := List.filter (fun sd => negb (is_shared sd)).

Lemma strand_share_same:
 forall (v:var) (s:strand),
 s = not_shared_only s ->
 s = not_shared_only (strand_share v s).
Proof.
 intros v.
 induction s as [|sd s]; intros EQs; simpl; intuition.
 case (In_dec (varid v) (strand_desc_vars_bound sd));
 intros INv; simpl; intuition;
 case_eq (negb (is_shared sd)); simpl; intros SHsd.
 case (const_eq_dec C.SHARE C.SHARE); intros EQc; simpl.
 rewrite EQs. simpl.
 case_eq (negb (is_shared sd)); intuition.
 rewrite SHsd in H. discriminate H.
 assert False by (apply EQc; intuition). contradiction.
 case (const_eq_dec C.SHARE C.SHARE); intros EQc; simpl.
 rewrite EQs. simpl.
 case_eq (negb (is_shared sd)); intuition.
 rewrite SHsd in H. discriminate H.
 assert False by (apply EQc; intuition). contradiction.
 rewrite <- IHs. reflexivity.
 revert EQs. simpl.
 case_eq (negb (is_shared sd)); intuition.
 inversion EQs. rewrite <- H1. assumption.
 rewrite SHsd in H. discriminate H.

 revert EQs. simpl.
 case_eq (negb (is_shared sd)); intuition.
 rewrite SHsd in H. discriminate H.
 clear H INv IHs.
 assert (~ List.In sd (not_shared_only s)).
 unfold not_shared_only.
 rewrite filter_In.
 intros [IN NEG].
 rewrite SHsd in NEG. discriminate NEG.
 assert False; [ idtac | contradiction ].
 apply H. clear H. rewrite <- EQs.
 simpl. left. reflexivity.
Qed.

Definition shared_perspective (s:strand) :=
List.map
 (fun sd =>
   match sd with
   | (d, m, fs) =>
     (Incoming, m, fs)
   end)
  (shared_only s).

Theorem strand_share_correct:
 forall (v:var) (s:strand),
 In (varid v) (strand_vars_bound s) ->
 In (varid v)
  (strand_vars_bound (shared_perspective (strand_share v s))).
Proof.
 intros v.
 induction s as [|sd s]; simpl.
 set_iff. trivial.
 destruct sd as [[d m] fs]. set (sd := (d,m,fs)).
 case (In_dec (varid v) (strand_desc_vars_bound sd)).
 intros INsd INs. clear IHs.
 unfold shared_perspective.
 simpl.
 case (const_eq_dec C.SHARE C.SHARE); intros EQc; simpl.
 destruct d. destruct m; simpl; set_iff; intuition.
 destruct m1; simpl; set_iff; intuition.
 destruct m2; simpl; set_iff; intuition.
 destruct fs; simpl; set_iff; intuition.
 case (const_eq_dec c C.SHARE); intros EQc'; simpl.
 set_iff; intuition.
 set_iff; intuition.
 simpl. set_iff; intuition.
 assert False by (apply EQc; intuition). contradiction.

 intros NINsd. set_iff. intros [INs | INsd].
 generalize (IHs INs). clear IHs.
 unfold shared_perspective. simpl.
 destruct d. destruct m; simpl; set_iff; intuition.
 destruct m1; simpl; set_iff; intuition.
 destruct m2; simpl; set_iff; intuition.
 destruct fs; simpl; set_iff; intuition.
 case (const_eq_dec c C.SHARE); intros EQc'; simpl.
 set_iff; intuition.
 set_iff; intuition.
 simpl. set_iff; intuition.
 contradiction.
Qed.

Hint Resolve strand_share_correct.

Theorem strand_share_subset:
 forall (v:var) (s:strand),
 Subset (strand_vars_bound s) (strand_vars_bound (strand_share v s)).
Proof.
 intros v.
 induction s as [|sd s]; simpl; intros a.
 set_iff. intuition.
 case (In_dec (varid v) (strand_desc_vars_bound sd)); simpl;
 set_iff; intuition.
Qed.

Hint Resolve strand_share_subset.

Theorem strand_shares_subset:
 forall (V:list var) (s:strand),
 Subset (strand_vars_bound s) (strand_vars_bound (strand_shares V s)).
Proof.
 induction V as [|v V]; simpl; intros s a; intuition.
 apply strand_share_subset.
 apply IHV.
 assumption.
Qed.

Hint Resolve strand_shares_subset.

Theorem strand_share_subset_sp:
 forall (v:var) (s:strand),
 Subset (strand_vars_bound (shared_perspective s))
  (strand_vars_bound (shared_perspective (strand_share v s))).
Proof.
 intros v.
 induction s as [|sd s]; simpl; intros a.
 set_iff. intuition.
 destruct sd as [[d m] fs]. set (sd := (d,m,fs)).
 case (In_dec (varid v) (strand_desc_vars_bound sd)); 
 simpl; intros INsvb.
 unfold shared_perspective. simpl.

 case (const_eq_dec C.SHARE C.SHARE); intros EQc; simpl.
 destruct d. destruct m; simpl; set_iff; intuition.
 destruct m1; simpl; set_iff; intuition.
 destruct m2; simpl; set_iff; intuition.
 destruct fs; simpl; set_iff; intuition.
 revert H.
 case (const_eq_dec c C.SHARE); intros EQc'; simpl.
 set_iff; intuition.
 set_iff; intuition.
 simpl. set_iff; intuition.
 assert False by (apply EQc; intuition). contradiction.

 unfold shared_perspective. simpl.

 case (const_eq_dec C.SHARE C.SHARE); intros EQc; simpl.
 destruct d. destruct m; simpl; set_iff; intuition.
 destruct m1; simpl; set_iff; intuition.
 destruct m2; simpl; set_iff; intuition.
 destruct fs; simpl; set_iff; intuition.
 revert H.
 case (const_eq_dec c C.SHARE); intros EQc'; simpl.
 set_iff; intuition.
 set_iff; intuition.
 simpl. set_iff; intuition.
 assert False by (apply EQc; intuition). contradiction.
Qed.

Hint Resolve strand_share_subset_sp.

Theorem strand_shares_correct:
 forall (V:list var) (s:strand),
 Subset (varlist_set V) (strand_vars_bound s) ->
 Subset (varlist_set V)
  (strand_vars_bound (shared_perspective (strand_shares V s))).
Proof.
 induction V as [|v V]; simpl.
 intros s SUB a. set_iff. intuition.
 intros s SUB a INa.
 generalize (SUB a INa).
 intros INsvb.
 revert INa.
 set_iff.
 intros [EQa | IN].
 subst a.
 apply strand_share_correct.
 apply strand_shares_subset.
 assumption.
 cut (Subset (varlist_set V) (strand_vars_bound s)).
 intros SUB'. clear SUB.
 generalize (IHV s SUB'). clear SUB' IHV.
 intros SUB. generalize (SUB a IN).
 clear SUB INsvb IN.
 apply strand_share_subset_sp.
 intros b INV.
 apply SUB. set_iff. right. assumption.
Qed.
