(* CPPLss *)
Require Export CPPLc.
Open Scope list_scope .

(* Suc strands *)
  Definition filter_faildesc ss :=
  List.filter
   (fun snv:(strand * t) =>
    let (s,v) := snv in
    if strand_desc_eq_dec (last s faildesc) faildesc then
     false
    else 
     true)
   ss.

  Lemma filter_faildesc_correct:
   forall (ss:list (strand*t)) (s:strand) (v:t),
    List.In (s,v) ss ->
    (List.In (s,v) (filter_faildesc ss)
    <-> last s faildesc <> faildesc).
  Proof.
   intros ss s v IN.
   unfold filter_faildesc.
   rewrite filter_In.
   case (strand_desc_eq_dec (last s faildesc) faildesc);
   intuition.
  Qed.

Hint Rewrite -> filter_faildesc_correct : ppl.

  Lemma filter_faildesc_preserve:
   forall (ss:list (strand*t)) (s:strand) (v:t),
    List.In (s,v) (filter_faildesc ss) ->
    List.In (s,v) ss.
  Proof.
   intros ss s v.
   unfold filter_faildesc.
   rewrite filter_In.
   intros [IN H]. assumption.
  Qed.

Hint Resolve filter_faildesc_preserve.

  Lemma filter_faildesc_not:
   forall (ss:list (strand*t)) (s:strand) (v:t),
    List.In (s,v) (filter_faildesc ss) ->
    last s faildesc <> faildesc.
  Proof.
   intros ss s v.
   unfold filter_faildesc.
   rewrite filter_In.
   case (strand_desc_eq_dec (last s faildesc) faildesc);
   intuition.
  Qed.

Hint Resolve filter_faildesc_not.

  Definition proc_suc_strands p:
   { ss:list (strand * t) |
   forall (s:strand) (v:t),
   List.In (s,v) ss <->	
   (last s faildesc <> faildesc
    /\ eval_proc p s v) }.
  Proof.
   intros p.
   destruct (proc_strands p) as [ss H].
   exists (filter_faildesc ss).
   intros s v.
   rewrite <- (H s v). clear H.
   split.
   intros FF. split.
   apply filter_faildesc_not with ss v. assumption.
   apply filter_faildesc_preserve. assumption.
   intros [LF IN].
   rewrite filter_faildesc_correct; assumption.
  Qed.

(* Strand env *)
  Definition type_strand_descT (sd:strand_desc) := 
  match sd with
  | (r, m, f) => type_msgT m
  end.

  Lemma type_strand_desc_envT:
   forall (sd:strand_desc) (env:t),
    type_strand_desc env sd -> type_strand_descT sd.
  Proof.
   intros sd env.
   destruct sd as [[r m] f]. destruct r; simpl;
   intros [TM SUB].
   apply type_msg_sendT with (union env (formulas_vars_bound f)).
   assumption.
   apply type_msg_recvT with env. assumption.
  Qed.

Hint Resolve type_strand_desc_envT.

  Lemma strand_desc_vars_smallest:	
   forall (sd:strand_desc),
    type_strand_descT sd ->
    SmallestSetWithList
     (fun env => type_strand_desc env sd)
     (strand_desc_vars sd) (strand_desc_varsL sd).
  Proof.
   destruct sd as [[r m] f].
   destruct r; simpl; intros TM.

   destruct (message_vars_smallest m TM) as [SS VLS].
   unfold SmallestSetWithList, SmallestSet, VarListSet in *.
   destruct SS as [TMS SUB].
   split. split. split.
   apply type_msg_send_subset with (message_vars m).
   assumption.
   intro. set_iff. intuition.
   case (In_dec a (formulas_vars_bound f)); intuition.
   intro. set_iff. intuition.
   intros s'. intros [TMS' SUB'].
   intro. set_iff. intros [[InM InFVB] | InFVP].
   assert (Subset (message_vars m)
           (union s' (formulas_vars_bound f))) as SUB1.
   apply SUB. assumption.
   assert (In a (union s' (formulas_vars_bound f))) as IN1.
   apply SUB1. assumption.
   revert IN1. set_iff. intuition.
   apply SUB'; assumption.
   intro a. set_iff. split.
   intros [[InM NInB] | InFVP].
   rewrite VLS in InM. destruct InM as [t' InM].
   exists t'. rewrite <- in_app_iff.
   left. rewrite filter_In.
   split. assumption. simpl.
   unfold negb.
   case_eq (mem a (formulas_vars_bound f)).
   rewrite <- mem_iff. intros F. contradiction.
   intros. reflexivity.

   destruct (formulas_vars_to_proveL f a InFVP) as [t' InF].
   exists t'. rewrite <- in_app_iff.
   right. assumption.
   intros [t' IN]. rewrite <- in_app_iff in IN.
   destruct IN as [InM | InFVP].
   left. rewrite VLS.
   rewrite filter_In in InM.
   destruct InM as [InM NEG].
   split. exists t'. assumption.
   revert NEG. unfold negb. simpl.
   case_eq (mem a (formulas_vars_bound f)).
   intros MEM EQ. discriminate EQ.
   rewrite <- not_mem_iff. intros. assumption.

   right. rewrite formulas_vars_proveL_equiv.
   exists t'. assumption.

   destruct (message_vars_match_smallest m TM) as [SS VLS].
   unfold SmallestSetWithList, SmallestSet, VarListSet in *.
   destruct SS as [TMS SUB].
   split. split. split.
   apply type_msg_recv_subset with (message_vars_match m).
   assumption.
   intro. set_iff. intuition.
   intro. set_iff. intuition.
   case (In_dec a (message_vars_bound m)); intuition.
   intros s'. intros [TMS' SUB'].
   intro. set_iff. intros [InM | [InFVP NInB]].
   apply SUB; assumption.
   assert (In a (union s' (message_vars_bound m))) as IN1.
   apply SUB'. assumption.
   revert IN1. set_iff. intuition.

   intro a. set_iff. split.
   intros [InM | [InFVP NInB]].
   rewrite VLS in InM. destruct InM as [t' InM].
   exists t'. rewrite <- in_app_iff.
   right. assumption.
   destruct (formulas_vars_to_proveL f a InFVP) as [t' InF].
   exists t'. rewrite <- in_app_iff.
   left. rewrite filter_In.
   split. assumption.
   unfold negb. simpl.
   case_eq (mem a (message_vars_bound m)).
   rewrite <- mem_iff. intros. contradiction.
   intros. reflexivity.
   intros [t' IN]. rewrite <- in_app_iff in IN.
   destruct IN as [InM | InFVP].
   rewrite filter_In in InM.
   destruct InM as [InFVP NEG].
   right. split.
   rewrite formulas_vars_proveL_equiv.
   exists t'. assumption.
   rewrite not_mem_iff. revert NEG.
   simpl. case_eq (mem a (message_vars_bound m)); simpl.
   intros EQ F. discriminate F.
   intros. reflexivity.
   left. rewrite VLS. exists t'. assumption.
  Qed.

Hint Resolve strand_desc_vars_smallest.

  Fixpoint type_strandT s :=
  match s with
  | nil => True
  | sd :: s' => (type_strand_descT sd) /\ (type_strandT s')
  end.

  Lemma type_strand_envT:
   forall (s:strand) (env:t),
    type_strand env s -> type_strandT s.
  Proof.
   induction s as [|sd s]; simpl.
   trivial.
   intros env [TSD TS].
   split. apply type_strand_desc_envT with env. assumption.
   apply IHs with (union env (strand_desc_vars_bound sd)).
   assumption.
  Qed.

Hint Resolve type_strand_envT.

  Lemma strand_vars_smallest:
   forall (s:strand),
    type_strandT s ->
    SmallestSetWithList
     (fun env => type_strand env s)
     (strand_vars s) (strand_varsL s).
  Proof.
   induction s as [|sd s]; simpl.

   intros T. split. split. trivial.
   intros s' T' a.  set_iff. intros. contradiction.
   split. set_iff. intros. contradiction. simpl.
   intros [vt IN]. contradiction.

   intros [TSDT TST].
   destruct (IHs TST) as [[TS Ss] Hs]. clear IHs.
   destruct (strand_desc_vars_smallest sd TSDT)
   as [[TSD Ssd] Hsd].
   unfold SmallestSetWithList, SmallestSet, VarListSet.

   split. split. split.
   apply type_strand_desc_subset with (strand_desc_vars sd).
   unfold Subset. intro a. set_iff. intuition.
   assumption.
   apply type_strand_subset with (strand_vars s).
   unfold Subset. intro a. set_iff. intuition.
   case (In_dec a (strand_desc_vars_bound sd)); intuition.
   assumption.
   intros s' [TSD' TS'].
   unfold Subset. intro a. set_iff.
   intros [InE' | [InE NInB]]. 
   apply Ssd; assumption.
   cut (In a (union s' (strand_desc_vars_bound sd))).
   set_iff. intros [InS' | InB ].
   assumption. contradiction.
   apply Ss; assumption.
   
   clear TS TSDT TST Ss TSD Ssd.
   unfold VarListSet in *.
   intros a. set_iff.
   rewrite Hs. rewrite Hsd. clear Hs Hsd.
   split. intros [[vt InV'] | [[vt InV] NB]].
   exists vt. apply in_or_app. left. assumption.
   exists vt. apply in_or_app. right.
   rewrite filter_In. split.
   assumption. simpl. unfold negb.
   case_eq (mem a (strand_desc_vars_bound sd)).
   rewrite <- mem_iff. intros F. contradiction.
   intros. reflexivity.

   intros [vt InAPP]. rewrite <- in_app_iff in InAPP.
   destruct InAPP as [InV | InV'].
   left. exists vt. assumption.
   right. rewrite filter_In in InV'.
   destruct InV' as [InV NEG].
   split. exists vt. assumption.
   revert NEG. simpl. unfold negb.
   case_eq (mem a (strand_desc_vars_bound sd)).
   intros M EQ. discriminate EQ.
   rewrite not_mem_iff. intros. assumption.
  Qed.

Hint Resolve strand_vars_smallest.

(* Backup strand *)
  Definition formula_vars f :=
   union (formula_vars_bound f) (formula_vars_prove f).
  Definition formulas_vars f :=
   union (formulas_vars_bound f) (formulas_vars_prove f).

  Definition formula_db f :=
  match f with
  | Formula n vs =>
    Formula n (List.map (fun v => DataV v) (fvars vs))
  end.

  Definition formulas_db fs := List.map formula_db fs.

  Definition strandd_trust_db (sd:strand_desc) :=
  match sd with (_, _, fs) => (formulas_db fs) end.

  Definition strand_trust_db (s:strand) :=
  List.fold_right
   (fun (sd:strand_desc) a => (strandd_trust_db sd) ++ a) 
   nil
   s.

  Definition NecessaryFormula env f :=
  match f with
  | Formula n vs =>
    (exists v:var, List.In v (fvars vs) /\ In (varid v) env)
  end.

  Definition necessary_formula env f :=
  match f with
  | Formula n vs =>
    List.existsb (fun v => mem (varid v) env) (fvars vs)
  end.

  Lemma necessary_formula_iff:
   forall (env:t) (f:formula),
    NecessaryFormula env f <->
    (necessary_formula env f = true).
  Proof.
   intros env f.
   destruct f as [n vs]. simpl.
   rewrite existsb_exists.
   split; intros [v [LIN IN]];
   exists v;
   try (rewrite mem_iff in *);
   (split; assumption).
  Qed.

Hint Rewrite <- necessary_formula_iff : ppl.

  Lemma not_necessary_formula_iff:
   forall (env:t) (f:formula),
    ~ NecessaryFormula env f <->
    (necessary_formula env f = false).
  Proof.
   intros env f.
   rewrite necessary_formula_iff.
   case_eq (necessary_formula env f);
   intuition.
  Qed.

Hint Rewrite <- not_necessary_formula_iff : ppl.

  Inductive TrustDBPart :
  t -> list formula -> list formula -> list formula -> Prop :=
  | TDBP_NIL:
    forall (env:t),
    TrustDBPart env nil nil nil
  | TDBP_CONS_NEC_BEFORE:
    forall (env:t) (f:formula) (fs nfs ufs:list formula),
    NecessaryFormula env f ->
    TrustDBPart (union env (formula_vars f)) fs nfs ufs ->
    TrustDBPart env (f::fs) (f::nfs) ufs
  | TDBP_CONS_NEC_AFTER:
    forall (env:t) (f:formula)
           (fs nfs nfs' ufs ufs':list formula),
    TrustDBPart env fs nfs ufs ->
    ~ NecessaryFormula env f ->
    NecessaryFormula (union env (formulas_vars nfs)) f ->
    TrustDBPart (union (union env (formulas_vars nfs))
                  (formula_vars f))
     ufs nfs' ufs' ->
    TrustDBPart env (f::fs) (f::(nfs++nfs')) ufs'
  | TDBP_CONS_UNEC_AFTER:
    forall (env:t) (f:formula) (fs nfs ufs:list formula),
    TrustDBPart env fs nfs ufs ->
    ~ NecessaryFormula env f ->
    ~ NecessaryFormula (union env (formulas_vars nfs)) f ->
    TrustDBPart env (f::fs) nfs (f::ufs).

Hint Constructors TrustDBPart.

  Fixpoint trustdb_part n env fs
  : ((list formula) * (list formula)) :=
  match n with
  | O => (nil,nil)
  | S n =>
  match fs with
  | nil => (nil, nil)
  | f :: fs =>
    if necessary_formula env f then
       let env' := union env (formula_vars f) in
       let (nfs, ufs) := trustdb_part n env' fs in
       (f :: nfs, ufs)
    else
       let (nfs, ufs) := trustdb_part n env fs in
       let env' := union env (formulas_vars nfs) in
       if necessary_formula env' f then
          let env'' := union env' (formula_vars f) in
          let (nfs', ufs') := trustdb_part n env'' ufs in
          (f :: (nfs ++ nfs'), ufs')
       else
          (nfs, (f::ufs))
  end
  end.

  Lemma trustdb_part_len:
   forall (n:nat) (env:t) (fs nfs ufs:list formula),
   List.length fs <= n ->
   trustdb_part n env fs = (nfs,ufs) ->
   List.length (nfs ++ ufs) = List.length fs.
  Proof.
   intros n.
   induction n as [|n]; simpl;
   intros env fs nfs ufs LE EQ.

   inversion EQ; subst. simpl. clear EQ.
   destruct fs as [|f fs]; simpl.
   reflexivity. simpl in LE.
   absurd_hyp LE; [idtac | assumption].
   apply le_Sn_O.
   revert EQ. destruct fs as [|f fs]; simpl.
   intros EQ.
   inversion EQ; subst. simpl. reflexivity.
   cut (length fs <= n). clear LE. intros LE.
   case (necessary_formula env f).
   set (env' := (union env (formula_vars f))).
   case_eq (trustdb_part n env' fs).
   intros nfs' ufs' EQ.
   intros EQ'. inversion EQ'. subst. clear EQ'.
   simpl.
   rewrite (IHn env' fs nfs' ufs LE EQ).
   reflexivity.
   case_eq (trustdb_part n env fs).
   intros nfs' ufs' EQ.
   set (env' := (union env (formulas_vars nfs'))).
   case (necessary_formula env' f).
   set (env'' := union env' (formula_vars f)).
   case_eq (trustdb_part n env'' ufs').
   intros nfs'' ufs'' EQ'.
   intros EQ''. inversion EQ''. subst. clear EQ''.
   rewrite <- (IHn env fs nfs' ufs' LE EQ).
   rewrite <- (IHn env fs nfs' ufs' LE EQ) in LE.
   clear fs EQ.
   cut (length ufs' <= n). intros LE2.
   simpl. rewrite app_length.
   rewrite app_length. rewrite app_length.
   rewrite <- (IHn env'' ufs' nfs'' ufs LE2 EQ').
   rewrite app_length.
   rewrite Arith.Plus.plus_assoc. reflexivity.
   apply le_trans with (length (nfs' ++ ufs')).
   rewrite app_length.
   apply Arith.Plus.le_plus_r.
   assumption.
   intros EQ'. inversion EQ'. subst. clear EQ'.
   rewrite <- (IHn env fs nfs ufs' LE EQ).
   rewrite app_length.
   rewrite app_length.
   simpl. rewrite <- Arith.Plus.plus_Snm_nSm.
   simpl. reflexivity.
   apply le_S_n. simpl in LE.
   assumption.
  Qed.

Hint Resolve trustdb_part_len.

  Lemma trustdb_part_iff:
   forall (n:nat) (env:t) (fs nfs ufs:list formula),
   List.length fs <= n ->
   ((trustdb_part n env fs = (nfs,ufs))
   <-> TrustDBPart env fs nfs ufs).
  Proof.
   intros n. induction n as [|n]; simpl;
   intros env fs nfs ufs LE.
   destruct fs as [|f fs]; simpl in LE.
   split. intro EQ. inversion EQ. apply TDBP_NIL.
   intro EQ. inversion EQ. reflexivity.
   absurd_hyp LE; [idtac | assumption].
   apply le_Sn_O.
   destruct fs as [|f fs]; simpl; simpl in LE.
   split. intro EQ. inversion EQ. apply TDBP_NIL.
   intro EQ. inversion EQ. reflexivity.
   cut (length fs <= n). clear LE. intro LE.
   split.

   (* Case 1 *)
   case_eq (necessary_formula env f).
   rewrite <- necessary_formula_iff.
   intros NEC.
   case_eq (trustdb_part n (union env (formula_vars f)) fs).
   intros nfs' ufs'. 
   rewrite (IHn (union env (formula_vars f))
                fs nfs' ufs' LE).
   intros TDBP.
   intros EQ. inversion EQ. subst. clear EQ.
   apply TDBP_CONS_NEC_BEFORE.
   assumption. assumption.

   rewrite <- not_necessary_formula_iff.
   intros NNEC.
   case_eq (trustdb_part n env fs).
   intros nfs' ufs'. intros TDBP.
   cut (length fs = length nfs' + length ufs'). intros EQlen.
   rewrite (IHn env fs nfs' ufs' LE) in TDBP.
   case_eq (necessary_formula
    (union env (formulas_vars nfs')) f).
   rewrite <- necessary_formula_iff.
   intros NEC.
   set (env' := union (union env (formulas_vars nfs'))
                 (formula_vars f)).
   case_eq (trustdb_part n env' ufs').
   intros nfs'' ufs'' TDBP'.
   cut (length ufs' = length nfs'' + length ufs'').
   intros EQlen'.
   rewrite (IHn env' ufs' nfs'' ufs'') in TDBP'.
   intros EQ. inversion EQ. subst nfs ufs. clear EQ.
   apply TDBP_CONS_NEC_AFTER with ufs'; assumption.
 
   apply le_trans with (length fs).
   rewrite EQlen. apply Arith.Plus.le_plus_r.
   assumption. rewrite <- app_length.
   symmetry.
   apply trustdb_part_len with n env'.
   apply le_trans with (length fs).
   rewrite EQlen. apply Arith.Plus.le_plus_r.
   assumption. assumption.

   rewrite <- not_necessary_formula_iff.
   intros NNECA.
   intros EQ. inversion EQ. subst. clear EQ.
   apply TDBP_CONS_UNEC_AFTER; assumption.
   symmetry. rewrite <- app_length.
   apply trustdb_part_len with n env; assumption.

   (* Case 2 *)
   intros TDBP. inversion TDBP; subst.
   case_eq (necessary_formula env f).
   rewrite <- necessary_formula_iff.
   intros F. clear F.
   set (env' := (union env (formula_vars f))).   
   case_eq (trustdb_part n env' fs).
   intros nfs' ufs'.
   rewrite <- (IHn env' fs nfs0 ufs LE) in H5.
   intros EQ. subst env'. rewrite H5 in EQ.
   clear H5. inversion EQ. subst nfs0 ufs. clear EQ.
   reflexivity.
   rewrite <- not_necessary_formula_iff.
   intros F. contradiction.
   
   case_eq (necessary_formula env f).
   rewrite <- necessary_formula_iff.
   intros F. contradiction.
   intros F. clear F.
   case_eq (trustdb_part n env fs).
   intros nfs'' ufs'' EQ.
   rewrite <- (IHn env fs nfs0 ufs0 LE) in H1.
   rewrite H1 in EQ. inversion EQ. subst nfs0 ufs0.
   clear EQ.
   set (env' := (union env (formulas_vars nfs''))).
   case_eq (necessary_formula env' f).
   intro F. clear F.
   set (env'' := (union env' (formula_vars f))).
   case_eq (trustdb_part n env'' ufs'').
   intros nfs''' ufs''' EQ.
   rewrite <- (IHn env'' ufs'' nfs' ufs) in H7.
   rewrite H7 in EQ. inversion EQ. subst nfs''' ufs'''.
   clear H7 EQ.
   reflexivity.   
   apply le_trans with (length fs).
   rewrite <- (trustdb_part_len n env fs nfs'' ufs'').
   rewrite app_length. apply Arith.Plus.le_plus_r.
   assumption. assumption. assumption.
   rewrite <- not_necessary_formula_iff.
   intros F. contradiction.

   case_eq (necessary_formula env f).
   rewrite <- necessary_formula_iff.
   intros F. contradiction.
   intros F. clear F.
   case_eq (trustdb_part n env fs).
   intros nfs'' ufs'' EQ.
   rewrite <- (IHn env fs nfs ufs0 LE) in H1.
   rewrite H1 in EQ. inversion EQ. subst nfs ufs0.
   clear EQ.
   set (env' := (union env (formulas_vars nfs''))).
   case_eq (necessary_formula env' f).
   rewrite <- necessary_formula_iff.
   intros F. contradiction.
   intro F. clear F.
   reflexivity.

   apply le_S_n. assumption.
  Qed.

Hint Rewrite -> trustdb_part_iff : ppl.

  Lemma trustdb_part_part:
   forall (n:nat) (env:t) (fs nfs ufs:list formula),
   List.length fs <= n ->
   trustdb_part n env fs = (nfs,ufs) ->
   forall (f:formula),
   List.In f fs <->
   ((List.In f nfs) \/ (List.In f ufs)).
  Proof.
   intros n. induction n as [|n]; simpl;
   intros env fs nfs ufs LE.
   destruct fs as [|f fs]; simpl in LE.
   intro EQ. inversion EQ. subst. simpl.
   intuition.
   absurd_hyp LE; [idtac | assumption].
   apply le_Sn_O.

   destruct fs as [|f fs]; simpl; simpl in LE.
   intros EQ. inversion EQ. subst. simpl. intuition.
   cut (length fs <= n). clear LE. intro LE.

   case (necessary_formula env f).
   set (env' := (union env (formula_vars f))).
   case_eq (trustdb_part n env' fs).
   intros nfs' ufs' TDBP EQ. 
   inversion EQ. subst. clear EQ.
   intros f'.
   rewrite (IHn env' fs nfs' ufs LE TDBP).
   simpl. intuition.

   case_eq (trustdb_part n env fs).
   intros nfs' ufs'. intros TDBP.
   case (necessary_formula
    (union env (formulas_vars nfs')) f).
   set (env' := union (union env (formulas_vars nfs'))
                 (formula_vars f)).
   case_eq (trustdb_part n env' ufs').
   intros nfs'' ufs'' TDBP'.
   intros EQ. inversion EQ. subst nfs ufs. clear EQ.
   intros f'.
   rewrite (IHn env fs nfs' ufs' LE TDBP).
   cut (List.length ufs' <= n). intros LE2.
   rewrite (IHn env' ufs' nfs'' ufs'' LE2 TDBP').
   simpl. rewrite <- in_app_iff.
   intuition.

   apply le_trans with (length fs).
   rewrite <- (trustdb_part_len n env fs nfs' ufs' LE TDBP).
   rewrite app_length. apply Arith.Plus.le_plus_r.
   assumption.

   intros EQ. inversion EQ. subst. clear EQ.
   intros f'. 
   rewrite (IHn env fs nfs ufs' LE TDBP).
   simpl. intuition.

   apply le_S_n. assumption.
  Qed.
   
Hint Rewrite <- trustdb_part_part : ppl.

  Definition trustdb_part_auto env fs :=
   trustdb_part (List.length fs) env fs.

Hint Unfold trustdb_part_auto.

  Lemma trustdb_part_auto_iff:
   forall (env:t) (fs nfs ufs:list formula),
   ((trustdb_part_auto env fs = (nfs,ufs))
   <-> TrustDBPart env fs nfs ufs).
  Proof.
   intros env fs nfs ufs. 
   unfold trustdb_part_auto.
   rewrite trustdb_part_iff.
   apply le_refl.
   split; trivial.
  Qed.

Hint Rewrite -> trustdb_part_auto_iff : ppl.

  Lemma type_strand_cut:
   forall (s s':strand) (env:t),
    type_strand env (s ++ s') <->
    type_strand env s
    /\ type_strand (union env (strand_vars_bound s)) s'.
  Proof.
   induction s as [|sd s]; simpl.
   intros s' env. split.
   intros TS. split. trivial. apply type_strand_subset with env.
   unfold Subset. intro a. set_iff. intuition. assumption.
   intros [T TS]. apply type_strand_subset with (union env empty).
   unfold Subset. intro a. set_iff. intuition.
   assumption.
   
   intros s' env. split.
   intros [TSD TS].
   rewrite IHs in TS. destruct TS as [TS TS'].
   split. split. assumption. assumption.
   apply type_strand_subset with
    (union (union env (strand_desc_vars_bound sd))
     (strand_vars_bound s)).
   unfold Subset. intro a. set_iff. intuition.
   assumption.
   intros [[TSD TS] TS']. rewrite IHs.
   split. assumption.
   split. assumption.
   apply type_strand_subset with
    (union env (union (strand_vars_bound s)
     (strand_desc_vars_bound sd))).
   unfold Subset. intro a. set_iff. intuition.
   assumption.
  Qed.

Hint Rewrite -> type_strand_cut : ppl.

  Fixpoint list_dedup eT e_eq_dec l :=
  match l with
  | nil => nil
  | e :: l =>
    let l' := list_dedup eT e_eq_dec l in
    if @List.In_dec eT e_eq_dec e l' then
       l'
    else
       e :: l'
  end.

  Lemma list_dedup_NoDup:
   forall (eT:Type) 
    (e_eq_dec:(forall (x y:eT), {x = y} + {x <> y}))
    (l:list eT),
   NoDup (list_dedup eT e_eq_dec l).
  Proof.
   intros eT e_eq_dec.
   induction l as [|e l]; simpl.
   apply NoDup_nil.
   case (List.In_dec e_eq_dec e (list_dedup eT e_eq_dec l)).
   intros. assumption.
   intros NIN.
   apply NoDup_cons; assumption.
  Qed.

  Definition vars_dedup := list_dedup var var_eq_dec.

  Lemma vars_dedup_varlist_set:
   forall (vsL:list var),
   Equal (varlist_set vsL)
    (varlist_set (vars_dedup vsL)).
  Proof.
   induction vsL as [|v vsL]; simpl.
   reflexivity.
   unfold vars_dedup.
   simpl.
   case (List.In_dec var_eq_dec v
    (list_dedup var var_eq_dec vsL)).
   intros IN.
   rewrite <- IHvsL.
   intros a. set_iff. intuition.
   unfold E.eq, IOT.eq in H0.
   subst a.
   rewrite IHvsL. clear IHvsL.
   unfold vars_dedup. revert IN.
   set (l' := list_dedup var var_eq_dec vsL).
   assert (VarListSet l' (varlist_set l')) as VLS.
   apply varlist_set_in.
   unfold VarListSet in VLS.
   destruct v as [vi vt].
   rewrite (VLS vi). intros IN.
   exists vt. assumption.
   simpl. intros NIN.
   unfold vars_dedup in IHvsL.
   rewrite <- IHvsL.
   reflexivity.
  Qed.

Hint Rewrite <- vars_dedup_varlist_set : ppl.

  Definition backup_strandd dir sl sr : strand_desc :=
  let n := length sl in
  let sr_vsL := strand_varsL sr in
  let sr_vs := strand_vars sr in
  let all_sr_vs := union sr_vs (strand_vars_bound sr) in
  let (nfs, ufs) :=
     trustdb_part_auto all_sr_vs (strand_trust_db sl) in
  let vs := union sr_vs (formulas_vars_prove nfs) in
  let vsL := vars_dedup (sr_vsL ++ (formulas_vars_proveL nfs)) in
  let bmsg := Join (MConst C.BACKUP) 
   (Join (MConst (C.from_nat n)) (vars_msg vsL)) in
  match dir with
  | Outgoing =>
    (Outgoing, bmsg, nfs)
  | Incoming =>
    (Incoming, bmsg, nfs)
  end.

  Lemma varlistset_formulas_prove:
   forall (fs:list formula),
    VarListSet 
     (formulas_vars_proveL fs)
     (formulas_vars_prove fs).
  Proof.
   intros fs. unfold VarListSet.
   intro a.
   rewrite formulas_vars_proveL_equiv.
   split; trivial.
  Qed.

Hint Resolve varlistset_formulas_prove.

  Lemma backup_strandd_type_recv :
   forall (vs:t) (sl sr:strand),
   type_strand vs (sl ++ sr) ->
   type_strand_desc vs (backup_strandd Incoming sl sr).
  Proof.
   intros vs sl sr TS.
   unfold backup_strandd.
   destruct (trustdb_part_auto
              (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)) as [nfs ufs].
   simpl. split. split. trivial.
   split. trivial.
   apply type_msg_recv_vars_msg.
   rewrite vars_msg_bound.
   rewrite <- vars_dedup_varlist_set.
   rewrite <- varlist_set_unionapp.
   cut (VarListSet 
        (formulas_vars_proveL nfs)
        (formulas_vars_prove nfs)).
   intros EQ. rewrite varlistset_eq in EQ.
   rewrite EQ. clear EQ.
   intro. set_iff. intuition.
   apply varlistset_formulas_prove.
  Qed.

Hint Resolve backup_strandd_type_recv.

  Lemma listin_prop:
   forall (A:Type) (l1 l2 l3:list A) (P:A -> Prop),
   (forall (a:A),
    List.In a l1 <-> ((List.In a l2) \/ (List.In a l3))) ->
   (forall (a:A), List.In a l1 -> P a) ->
   (forall (a:A), List.In a l2 -> P a).
  Proof.
   intros A l1 l2 l3 P SL P1.
   intros a IN2.
   apply P1. rewrite SL. left.
   assumption.
  Qed.

  Lemma formula_db_vars:
   forall (f:formula),
   Equal (formula_vars_prove (formula_db f))
         (union (formula_vars_prove f)
                (formula_vars_bound f)).
  Proof.
   intros f. destruct f as [n vs].
   simpl.
   induction vs as [|[v|v] vs]; simpl;
   try (rewrite IHvs);
   intro; set_iff; intuition.
  Qed.

Hint Rewrite -> formula_db_vars : ppl.

  Lemma formula_db_vars_bound:
   forall (f:formula),
   Equal (formula_vars_bound (formula_db f))
         empty.
  Proof.
   intros f. destruct f as [n vs].
   simpl.
   induction vs as [|[v|v] vs]; simpl;
   try (rewrite IHvs);
   intro; set_iff; intuition.
  Qed.  

Hint Rewrite -> formula_db_vars_bound : ppl.

  Lemma formulas_db_vars_bound:
   forall (fs:list formula),
   Equal (formulas_vars_bound (formulas_db fs))
         empty.
  Proof.
   induction fs as [|f fs]; simpl.
   reflexivity.
   rewrite IHfs.
   rewrite formula_db_vars_bound.
   intro; set_iff; intuition.
  Qed.

Hint Rewrite -> formulas_db_vars_bound : ppl.

  Lemma trustdb_subset:
   forall (vs:t) (s:strand),
   type_strand vs s ->
   forall (f:formula),
   List.In f (strand_trust_db s) ->
   Subset (formula_vars_prove f)
          (union vs (strand_vars_bound s)).
  Proof.
   intros vs s. revert vs.
   induction s as [|sd s]; simpl.
   intros. contradiction.
   destruct sd as [[r m] fs].
   destruct r; simpl.

   (* Case 1 *)
   intros vs [[TMS SUB] TS] f'.
   rewrite <- in_app_iff.
   intros [INF | INST].
   clear TMS TS IHs.
   revert SUB INF.
   induction fs as [|f fs]; simpl.
   intros. contradiction.
   intros SUB [EQ | INF].
   subst f'. rewrite formula_db_vars.
   clear IHfs. intro. set_iff. intuition.
   left. apply SUB. set_iff. left. assumption.
   apply subset_trans with
    (union vs (union (strand_vars_bound s)
      (formulas_vars_bound fs))).
   apply IHfs. intro. set_iff. intro IN.
   apply SUB. set_iff. right. assumption.
   assumption.
   intro. set_iff. intuition.
   apply subset_trans with 
    (union (union vs (formulas_vars_bound fs))
     (strand_vars_bound s)).
   apply IHs. assumption.
   assumption. intro. set_iff. intuition.

   (* Case 2 *)
   intros vs [[TMS SUB] TS] f'.
   rewrite <- in_app_iff.
   intros [INF | INST].
   clear TMS TS IHs.
   revert SUB INF.
   induction fs as [|f fs]; simpl.
   intros. contradiction.
   intros SUB [EQ | INF].
   subst f'. rewrite formula_db_vars.
   clear IHfs. intro. set_iff. intuition.
   cut (In a (union vs (message_vars_bound m))).
   set_iff. intuition.
   apply SUB. set_iff. left. assumption.
   apply subset_trans with
    (union vs (union (strand_vars_bound s)
      (union (message_vars_bound m)
      (formulas_vars_bound fs)))).
   apply IHfs. intro. intro IN.
   apply SUB. set_iff. right. assumption.
   assumption.
   intro. set_iff. intuition.
   apply subset_trans with 
    (union (union vs (union
      (message_vars_bound m) (formulas_vars_bound fs)))
     (strand_vars_bound s)).
   apply IHs. assumption.
   assumption. intro. set_iff. intuition.
  Qed.

Hint Resolve trustdb_subset.

  Lemma formulas_vars_prove_subset:
   forall (s:t) (fs:list formula),
   (forall (f:formula),
    List.In f fs ->
    Subset (formula_vars_prove f) s) ->
   Subset (formulas_vars_prove fs) s.
  Proof.
   intros s fs.
   induction fs as [|f fs]; simpl.
   intros SUB a. set_iff. intuition.
   intros SUB a. set_iff.
   intros [INF | INFS].
   apply SUB with f. left. reflexivity.
   assumption.
   apply IHfs. intros f' IN.
   apply SUB. right. assumption.
   assumption.
  Qed.

Hint Resolve formulas_vars_prove_subset.

  Lemma trustdb_type_subset:
   forall (vs vs':t) (s:strand) (nfs ufs:list formula),
   type_strand vs s ->
   TrustDBPart vs' (strand_trust_db s) nfs ufs ->
   Subset (formulas_vars_prove nfs)
          (union vs (strand_vars_bound s)).
  Proof.
   intros vs vs' s nfs ufs TS TDBP.
   apply formulas_vars_prove_subset.
   apply (listin_prop
          formula (strand_trust_db s) nfs ufs
          (fun f => Subset (formula_vars_prove f)
           (union vs (strand_vars_bound s)))).
   intro f.
   rewrite <- trustdb_part_auto_iff in TDBP.
   unfold trustdb_part_auto in TDBP.
   rewrite (trustdb_part_part (length (strand_trust_db s))
    vs' (strand_trust_db s)
    nfs ufs).
   apply le_refl. assumption.
   split; trivial.
   apply trustdb_subset.
   assumption.
  Qed.

Hint Resolve trustdb_type_subset.

  Lemma backup_strandd_type_send:
   forall (vs:t) (sl sr:strand),
   type_strand vs (sl ++ sr) ->
   type_strand_desc (union vs (strand_vars_bound sl))
    (backup_strandd Outgoing sl sr).
  Proof.
   intros vs sl sr TS.
   rewrite type_strand_cut in TS.
   destruct TS as [TSl TSr].
   unfold backup_strandd.
   case_eq (trustdb_part_auto
              (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)).
   intros nfs ufs TDBP. simpl.
   destruct (strand_vars_smallest sr) as [SS VLS].
   apply type_strand_envT with (union vs (strand_vars_bound sl)).
   assumption.
   unfold SmallestSet in SS.
   destruct SS as [TSsr' SUB].

   split. split. trivial.
   split. trivial.
   rewrite type_msg_send_vars_msg.
   rewrite <- vars_dedup_varlist_set.
   rewrite <- varlist_set_unionapp.
   rewrite varlistset_eq in VLS. rewrite <- VLS.
   clear VLS.

   cut (VarListSet 
        (formulas_vars_proveL nfs)
        (formulas_vars_prove nfs)).
   intros EQ. rewrite varlistset_eq in EQ.
   rewrite <- EQ. clear EQ.

   apply union_subset_3.
   apply subset_trans with (union vs (strand_vars_bound sl)).
   apply SUB. assumption.
   intro. set_iff. intuition.

   apply subset_trans with (union vs (strand_vars_bound sl)).
   apply trustdb_type_subset 
   with (union (strand_vars sr) (strand_vars_bound sr))
        ufs.
   assumption.
   rewrite <- trustdb_part_auto_iff.
   assumption.

   intro. set_iff. intuition.
   apply varlistset_formulas_prove.

   apply trustdb_type_subset 
   with (union (strand_vars sr) (strand_vars_bound sr))
        ufs.
   assumption.
   rewrite <- trustdb_part_auto_iff.
   assumption.
  Qed.

Hint Resolve backup_strandd_type_send.

  Lemma type_strand_cons:
   forall (vs:t) (s:strand) (sd:strand_desc),
   type_strandT s ->
   type_strand_desc vs sd ->
   Subset (strand_vars s) (strand_desc_vars_bound sd) ->
   type_strand vs (sd :: s).
  Proof.
   intros vs s sd TST TSD SUB.
   simpl. split. assumption.
   destruct (strand_vars_smallest s) as [SS VLS].
   assumption.
   unfold SmallestSet in SS.
   destruct SS as [TSsr' SUB'].
   apply type_strand_subset with (strand_vars s).
   intro. set_iff. intuition.
   assumption.
  Qed.

Hint Resolve type_strand_cons.

  Lemma backup_strandd_type:
   forall (vs:t) (sl sr:strand),
   type_strand vs (sl ++ sr) ->
   type_strand vs 
    ((backup_strandd Incoming sl sr) :: sr).
  Proof.
   intros vs sl sr TS.
   cut (type_strandT sr). intro TST.
   apply type_strand_cons.
   assumption.
   apply backup_strandd_type_recv. assumption.
   unfold backup_strandd.
   case_eq (trustdb_part_auto
             (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)).
   intros nfs ufs TDBP.
   simpl. rewrite vars_msg_bound.
   destruct (strand_vars_smallest sr) as [SS VLS].
   assumption.
   rewrite varlistset_eq in VLS. rewrite VLS.
   clear VLS.
   rewrite <- vars_dedup_varlist_set.
   rewrite <- varlist_set_unionapp.
   intro. set_iff. intuition.

   rewrite type_strand_cut in TS.
   destruct TS as [TSl TSr]. 
   apply type_strand_envT with (union vs (strand_vars_bound sl)).
   assumption.
  Qed.

Hint Resolve backup_strandd_type.

  Lemma formulas_db_formula:
   forall (fs:list formula),
   (forall (f:formula), List.In f fs ->
   formula_db f = f) ->
   formulas_db fs = fs.
  Proof.
   intros fs H.
   induction fs as [|f fs].
   simpl. reflexivity.
   simpl. rewrite H.
   rewrite IHfs. reflexivity.
   intros f'. intros IN.
   rewrite H. reflexivity.
   simpl. right. assumption.
   simpl. left. reflexivity.
  Qed.

Hint Rewrite -> formulas_db_formula : ppl.

  Lemma formula_db_idem:
   forall (f:formula),
    formula_db (formula_db f) = formula_db f.
  Proof.
   destruct f as [n vs]. simpl.
   unfold fvars.
   rewrite map_map.
   rewrite map_map.
   rewrite map_map.
   reflexivity.
  Qed.

Hint Rewrite -> formula_db_idem : ppl.

  Lemma formula_db_formulas:
   forall (f:formula) (fs:list formula),
    List.In f (formulas_db fs) ->
    formula_db f = f.
  Proof.
   intros f fs.
   induction fs as [|f' fs]; simpl.
   intros. contradiction.
   intros [EQ | IN].
   subst f. clear IHfs.
   apply formula_db_idem.
   apply IHfs. assumption.
  Qed.

Hint Rewrite -> formula_db_formulas : ppl.

  Lemma strand_trust_db_formula:
   forall (s:strand) (f:formula),
    List.In f (strand_trust_db s) ->
    formula_db f = f.
  Proof.
   induction s as [|[[r m] fs] s]; simpl.
   intros. contradiction.
   intros f. rewrite <- in_app_iff.
   intros [INFD | INST].
   apply formula_db_formulas with fs.
   assumption.
   apply IHs. assumption.
  Qed.

Hint Rewrite -> strand_trust_db_formula : ppl.

  Lemma backup_strandd_trust_db:
   forall (sl sr:strand),
   strandd_trust_db (backup_strandd Incoming sl sr) = 
      fst (trustdb_part_auto
           (union (strand_vars sr) (strand_vars_bound sr))
           (strand_trust_db sl)).
  Proof.
   intros sl sr.
   unfold backup_strandd.
   case_eq (trustdb_part_auto
            (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)).
   intros nfs ufs TDBP.
   simpl. apply formulas_db_formula.
   intros f IN.
   apply (listin_prop
          formula (strand_trust_db sl) nfs ufs
          (fun f => formula_db f = f)).
   intros f'.
   unfold trustdb_part_auto in TDBP.
   rewrite (trustdb_part_part (length (strand_trust_db sl))
    (union (strand_vars sr) (strand_vars_bound sr))
    (strand_trust_db sl)
    nfs ufs).
   apply le_refl. assumption.
   split; trivial.
   apply strand_trust_db_formula.
   assumption.
  Qed.

Hint Rewrite -> backup_strandd_trust_db : ppl.

  Lemma backup_strandd_fs:
   forall (sl sr:strand),
   snd (backup_strandd Incoming sl sr) = 
      fst (trustdb_part_auto
           (union (strand_vars sr) (strand_vars_bound sr))
           (strand_trust_db sl)).
  Proof.
   intros sl sr.
   unfold backup_strandd.
   case_eq (trustdb_part_auto
             (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)).
   intros nfs ufs TDBP.
   simpl. reflexivity.
  Qed.

Hint Rewrite -> backup_strandd_fs : ppl.

  Lemma formulas_vars_bound_ind:
   forall (fs:list formula),
   (forall (f:formula),
    List.In f fs -> Equal (formula_vars_bound f) empty) ->
   Equal (formulas_vars_bound fs) empty.
  Proof.
   intros fs.
   induction fs as [|f fs]; simpl.
   reflexivity.
   intros IH.
   rewrite IH. left. reflexivity.
   rewrite IHfs.
   intros f' IN. apply IH. right. assumption.
   intuition.
  Qed.

Hint Rewrite -> formulas_vars_bound_ind : ppl.

  Lemma backup_strandd_infsbound:
   forall (sl sr:strand),
   Equal (formulas_vars_bound
          (snd (backup_strandd Incoming sl sr)))
         empty.
  Proof.
   intros sl sr.
   unfold backup_strandd.
   case_eq (trustdb_part_auto
             (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)).
   intros nfs ufs TDBP.
   simpl.
   apply formulas_vars_bound_ind.
   apply (listin_prop
          formula (strand_trust_db sl) nfs ufs
          (fun f => Equal (formula_vars_bound f) empty)).
   intro f.
   unfold trustdb_part_auto in TDBP.
   rewrite (trustdb_part_part (length (strand_trust_db sl))
    (union (strand_vars sr) (strand_vars_bound sr))
    (strand_trust_db sl)
    nfs ufs).
   apply le_refl. assumption.
   split; trivial.

   clear sr nfs ufs TDBP.
   intros f.
   induction sl as [|[[r m] f'] sl].
   simpl. contradiction.
   simpl. rewrite <- in_app_iff.
   intros [L | R].
   clear sl IHsl. revert L.
   induction f' as [|f' fs].
   simpl. contradiction.
   simpl. intros [L | R].
   subst f. apply formula_db_vars_bound.
   apply IHfs. assumption.
   apply IHsl. assumption.
  Qed.

Hint Rewrite -> backup_strandd_infsbound : ppl.
  
  Definition BackupStrandD vs sl sr bsdi bsdo :=
   type_strand vs (sl ++ sr) ->
   type_strand_desc (union vs (strand_vars_bound sl)) bsdo
   /\ type_strand vs (bsdi :: sr)
   /\ snd bsdi = 
      fst (trustdb_part_auto
           (union (strand_vars sr) (strand_vars_bound sr))
           (strand_trust_db sl))
   /\ fst (fst bsdi) = Incoming
   /\ fst (fst bsdo) = Outgoing
   /\ snd bsdi = snd bsdo
   /\ snd (fst bsdi) = snd (fst bsdo).

  Theorem backup_strandd_is_BackupStrandD:
   forall (vs:t) (sl sr:strand),
    BackupStrandD vs sl sr
     (backup_strandd Incoming sl sr)
     (backup_strandd Outgoing sl sr).
  Proof.
   intros vs sl sr.
   unfold BackupStrandD.
   intros TS.
   split. apply backup_strandd_type_send. assumption.
   split. apply backup_strandd_type. assumption.
   (* split. apply backup_strandd_trust_db. *)

   unfold backup_strandd.
   case_eq (trustdb_part_auto 
            (union (strand_vars sr) (strand_vars_bound sr))
              (strand_trust_db sl)).
   intros nfs ufs TDBP. simpl.
   intuition.
  Qed.

Hint Resolve backup_strandd_is_BackupStrandD.

  Theorem backup_strandd_distinct:
   forall (sl1 sr1 sl2 sr2:strand),
    List.length sl1 <> List.length sl2 ->
    let m1 := snd (fst (backup_strandd Incoming sl1 sr1)) in
    let m2 := snd (fst (backup_strandd Incoming sl2 sr2)) in
    forall (cmpvL cmpvR:var -> var -> Prop),
    ~ matchm_either cmpvL cmpvR m1 m2.
  Proof.
   intros sl1 sr1 sl2 sr2 NEQ.
   unfold backup_strandd.
   case_eq (trustdb_part_auto
              (union (strand_vars sr1) (strand_vars_bound sr1))
              (strand_trust_db sl1)).
   intros nfs1 ufs1 TDBP1.
   case_eq (trustdb_part_auto
              (union (strand_vars sr2) (strand_vars_bound sr2))
              (strand_trust_db sl2)).
   intros nfs2 ufs2 TDBP2.
   intros cmpvL cmpvR.
   unfold matchm_either.
   simpl.
   intros [[BEQ [NMEQ H]] | [BEQ [NMEQ H]]]; clear H.
   apply C.from_nat_neq with (length sl1) (length sl2).
   assumption. symmetry. assumption.
   apply C.from_nat_neq with (length sl1) (length sl2);
   assumption.
  Qed.

Hint Resolve backup_strandd_distinct.

  Lemma varlist_set_mkv:
   forall (is:list id) (getT:id -> type),
    Equal (varlist_set (List.map (fun i => Var i (getT i)) is))
     (List.fold_right (fun e a => add e a) empty is).
  Proof.
   intros is getT.
   induction is as [|i is]; simpl.
   reflexivity.
   rewrite IHis.
   reflexivity.
  Qed.

Hint Rewrite -> varlist_set_mkv : ppl.

  Definition SmallestBackupStrandD vs sl sr bsdi bsdo :=
  (BackupStrandD vs sl sr bsdi bsdo) /\
  (forall (xbsdi:strand_desc) (xbsdo:strand_desc),
   (BackupStrandD vs sl sr xbsdi xbsdo) ->
   type_strand vs (sl ++ sr) ->
   Subset (strand_desc_vars_bound bsdi)
          (union vs (strand_desc_vars_bound xbsdi))).

  Lemma type_strand_cons2:
   forall (vs:t) (s:strand) (sd:strand_desc),
   type_strand vs (sd :: s) ->
   Subset (strand_vars s)
          (union vs (strand_desc_vars_bound sd)).
  Proof.
   intros vs s sd.
   simpl. intros [TSD TS].
   destruct (strand_vars_smallest s) as [SS VLS].
   apply type_strand_envT
   with (union vs (strand_desc_vars_bound sd)).
   assumption.
   unfold SmallestSet in SS.
   destruct SS as [TSsr SUB].
   apply SUB. assumption.
  Qed.   

Hint Resolve type_strand_cons2.

  Theorem backup_strandd_smallest:
   forall (vs:t) (sl sr:strand),
   SmallestBackupStrandD vs sl sr    
    (backup_strandd Incoming sl sr)
    (backup_strandd Outgoing sl sr).
  Proof.
   intros vs sl sr.
   unfold SmallestBackupStrandD.
   split. apply backup_strandd_is_BackupStrandD.
   intros [[xid xim] xif].
   intros [[xod xom] xof].
  
   unfold BackupStrandD.
   intros HBSD TS.
   destruct (HBSD TS) as [TSDo [TSi [TDB1 [D1 [D2 [TDB2 TMeq]]]]]].
   clear HBSD.
   simpl in TMeq. subst xom.
   simpl in TDB1. subst xif.
   simpl in TDB2. subst xof.
   simpl in D1. simpl in D2.
   subst xod. subst xid.

   cut (Equal (formulas_vars_bound
             (snd (backup_strandd Incoming sl sr))) empty);
   [ idtac | apply backup_strandd_infsbound].

   unfold backup_strandd.
   destruct (trustdb_part_auto
     (union (strand_vars sr) (strand_vars_bound sr))
     (strand_trust_db sl)) as [nfs ufs].
   simpl. 
   rewrite vars_msg_bound.
   rewrite <- vars_dedup_varlist_set.
   rewrite <- varlist_set_unionapp.

   destruct (strand_vars_smallest sr) as [SS VLS].
   rewrite type_strand_cut in TS. 
   apply type_strand_envT with (union vs (strand_vars_bound sl)).
   intuition.
   rewrite varlistset_eq in VLS. rewrite <- VLS.
   clear VLS.
   cut (VarListSet 
        (formulas_vars_proveL nfs)
        (formulas_vars_prove nfs));
   [ intros VLS | apply varlistset_formulas_prove ].   
   rewrite varlistset_eq in VLS. rewrite <- VLS.
   clear VLS.

   revert TSi TSDo. simpl.

   intros [[TRmi Sfvp1] TSr].
   intros [TSmi Sfvp2].
   intros EQfvb.
   rewrite EQfvb in *. clear EQfvb.
   rewrite (@empty_union_1 empty). intuition.
   rewrite (@empty_union_1 empty). intuition.
   rewrite (@empty_union_2 empty). intuition.
   rewrite (@empty_union_2 empty). intuition.
   rewrite (@empty_union_2 empty) in TSmi;
   [ idtac | intuition ].
   rewrite (@empty_union_2 empty) in TSr;
   [ idtac | intuition ].

   apply union_subset_3.

   unfold SmallestSet in SS.
   destruct SS as [TSsr SUB].
   apply SUB. assumption.
   assumption.
  Qed.

Hint Resolve backup_strandd_smallest.

  (* XXX Could refine formulas by looking at ones relied-upon in
         cont, before needing some vars *)

(* Measurement *)
  Definition strand_desc_size (sd:strand_desc) :=
  match sd with
  | (d, m, fs) =>
    cardinal (message_vars_bound m) + (List.length fs)
  end.

  Definition strand_backup i s :=
  let sl := List.firstn i s in
  let sr := List.skipn i s in
   backup_strandd Incoming sl sr.

  Definition strand_all_backups s :=
  List.map (fun i => strand_backup i s)
   (List.seq 0 (List.length s)).

  Definition strand_all_backups_sizes s :=
  List.map strand_desc_size (strand_all_backups s).

(* Same backup *)
  Definition backup_eq (sd0:strand_desc) (sd1:strand_desc) :=
  let m0 := snd (fst sd0) in
  let m1 := snd (fst sd1) in
  let fs0 := snd sd0 in
  let fs1 := snd sd1 in
  Equal (message_vars_bound m0) (message_vars_bound m1)
  /\ (fs0 = fs1).

  Lemma backup_eq_dec:
   forall (u v:strand_desc),
   {backup_eq u v} + {~ backup_eq u v}.
  Proof.
   destruct u as [[du mu] fu].
   destruct v as [[dv mv] fv].
   unfold backup_eq.
   simpl.
   destruct (list_eq_dec formula_eq_dec fu fv); intuition.
   destruct (Equal_dec (message_vars_bound mu) (message_vars_bound mv));
   intuition.
  Qed.

  Definition backups_eq sd0 sds :=
  List.existsb
   (fun sd1 => if backup_eq_dec sd0 sd1 then true else false)
   sds.

  Fixpoint backup_dup sds :=
  match sds with
  | nil => false
  | sd0 :: sds =>
    backups_eq sd0 sds || backup_dup sds
  end.
