Require Export Util DOrderedType.
Open Scope list_scope .

(* IOT *)
 Definition IOTt := DOrderedType.strt.
 Definition IOTeq := @eq IOTt.
 Parameter IOTlt : IOTt -> IOTt -> Prop.
 Definition IOTeq_refl := @refl_equal IOTt.
 Definition IOTeq_sym := @sym_eq IOTt.
 Definition IOTeq_trans := @trans_eq IOTt.
 Axiom IOTlt_trans :
  forall x y z : IOTt, IOTlt x y -> IOTlt y z -> IOTlt x z.
 Axiom IOTlt_not_eq :
  forall x y : IOTt, IOTlt x y -> ~ IOTeq x y.
 Parameter IOTcompare :
  forall x y : IOTt, Compare IOTlt IOTeq x y.

 Parameter new_id : unit -> IOTt .
 Parameter d1 : IOTt.
	
 Parameter d2 : IOTt.
 Parameter role_chanid : IOTt -> IOTt.
 Parameter nonce_id : IOTt -> IOTt.
(* / IOT *)

Module IOT : DUsualOrderedType.
 Definition t := IOTt.
	
 Definition str (s:IOTt) : t := s.
 Definition rts (s:t) : IOTt := s.
 Definition eq := @eq t.
 Definition eq_refl := @refl_equal t.
 Definition eq_sym := @sym_eq t.
 Definition eq_trans := @trans_eq t.
 Definition lt := IOTlt.
 Definition lt_trans := IOTlt_trans.
 Definition lt_not_eq := IOTlt_not_eq.
 Definition compare := IOTcompare.
 Definition new_id := new_id.
 Definition d1 := d1.
 Definition d2 := d2.
 Definition role_chanid := role_chanid.
 Definition nonce_id := nonce_id.
End IOT.

Export IOT.

Module Export IdSet := FSetList.Make(IOT).
Module Export IEP := FSetEqProperties.EqProperties(IdSet).
Module IP := MP.
	
Export IP.
Export IP.FM.

Definition SmallestSet P s :=
 P s /\ (forall s':t, P s' -> Subset s s').

Hint Unfold SmallestSet.

Lemma Subset_dec :
 forall (a b:t),
  {Subset a b} + {~ Subset a b}.
Proof.
 intros a b.
 case_eq (subset a b); intro S.
 left. rewrite subset_iff. assumption.
 right. intro S'. rewrite subset_iff in S'.
 absurd (false = true).
 discriminate. apply trans_eq with (subset a b).
 symmetry. assumption. assumption.
Qed.

Hint Resolve Subset_dec.

  Lemma Equal_dec:
   forall (u v:t), {Equal u v} + {~ Equal u v}.
  Proof.
   intros u v.
   case_eq (equal u v).
   intros EQ. left.
   rewrite equal_iff. intuition.
   intros NEQ. right.
   rewrite equal_iff. intros EQ.
   absurd (false = true).
   discriminate.
   rewrite <- EQ. rewrite <- NEQ.
 reflexivity.
Qed.

Hint Resolve Equal_dec.