Require Export Coq.Lists.List.
Require Export Coq.FSets.FSets.
Require Export Coq.Arith.EqNat.
Require Export Coq.Arith.Le Coq.Arith.Lt.
Require Export Coq.Bool.Bool.
Require Export Coq.Arith.Max.

(* Tactics *)
  Ltac nil_hypo :=
   match goal with
   | H : nil = ?car :: ?cdr  |- _ =>
     absurd (nil = car :: cdr);
     [ apply nil_cons
     | exact H]
   | H : nil = ?l1 ++ ?car :: ?cdr |- _ =>
     absurd (nil = l1 ++ car :: cdr);
     [ apply app_cons_not_nil
     | exact H]
   | H : ?car :: ?cdr = nil |- _ =>
     absurd (nil = car :: cdr);
     [ apply nil_cons
     | symmetry; exact H]
   | H : ?l1 ++ ?car :: ?cdr = nil |- _ =>
     absurd (nil = l1 ++ car :: cdr);
     [ apply app_cons_not_nil
     | symmetry; exact H]
   end.

  Ltac rewrite_one :=
   match goal with
   | H : ?lhs = ?rhs |- _ =>
     rewrite <- H in *;
     clear rhs H
   end.

  Ltac rewrite_every :=
   repeat rewrite_one.

  Ltac head_type t :=
  match t with
  | ?x -> ?y => head_type y
  | ?x => x
  end.
 
  Ltac assert_head t := 
  let m := type of t in 
  let u := head_type m in
  assert u; [apply t | idtac].

(* Lists *)
 Definition sublist A l1 l2 :=
 forall e:A, List.In e l1 -> List.In e l2.

Hint Unfold sublist.

 Lemma in_app_iff:
  forall (A:Type) (l m:list A) (a:A),
  List.In a l \/ List.In a m <-> List.In a (l ++ m).
 Proof.
  intros A l m a.
  split. intros H. apply in_or_app. assumption.
  intros H. apply in_app_or. assumption.
 Qed.

Hint Rewrite <- in_app_iff : datatypes.

(* Nat *)
 Lemma plus_gt_or:
  forall (n m:nat),
   n + m > 0 -> {n > 0} + {m > 0}.
 Proof.
  intros n m.
  induction n as [|n]; simpl.
  intros M. right. assumption.
  intros SNM. left. apply Arith.Gt.gt_Sn_O.
 Qed.

Hint Resolve plus_gt_or.

 Lemma plus_or_gt:
  forall (n m:nat),
   (n > 0) \/ (m > 0) ->
   n + m > 0.
 Proof.
  intros n m.
  destruct n as [|n]; 
  destruct m as [|m]; simpl.
  intros [F|F]; assumption.
  intros [F|G].
  assert (False). apply Arith.Gt.gt_irrefl with 0.
  assumption. contradiction.
  assumption.
  rewrite Arith.Plus.plus_0_r.
  intros [G|G]. assumption.
  assert (False). apply Arith.Gt.gt_irrefl with 0.
  assumption. contradiction.
  intros [G|G]; apply Arith.Gt.gt_Sn_O.
 Qed.

Hint Resolve plus_or_gt.