(* WPPLi *)
Require Export WPPL.
Open Scope list_scope.

Fixpoint action_missing_vars a {struct a} :=
  match a with
    | Return =>
      empty
    | Comm from_r to_r ffs msg tfs sa =>
      (union (message_vars msg)
        (action_missing_vars sa))
    | New r v t sa =>
      (remove (varid v)
        (action_missing_vars sa))
    | Bind r v msg sa =>
      (union (message_vars msg)
        (remove (varid v) (action_missing_vars sa)))
    | Match r v msg sa =>
      (union (add (varid v) (message_vars_match msg))
        (action_missing_vars sa))
    | Derive r f sa =>
      (union (formulas_vars_prove f)
        (diff (action_missing_vars sa)
          (formulas_vars_bound f)))
  end.

 Lemma action_missing_to_varsL:
  forall (a:action) (vi:id),
   In vi (action_missing_vars a) ->
    exists vt:type,
     List.In (Var vi vt) (action_varsL a).
 Proof.
  simple induction a; clear a; simpl;
  [ idtac
  | intros fr tr ffs m tfs a IHa
  | intros ar v nt a IHa
  | intros ar v m a IHa
  | intros ar v m a IHa
  | intros ar f a IHa ];
  intros vi; set_iff.

  intro F. contradiction.

  intros [ InM | InAM ].
  destruct (message_to_varsL m vi InM) as [ vt InML ].
  exists vt. apply in_or_app. left. assumption.
  destruct (IHa vi InAM) as [ vt InAL ].
  exists vt. apply in_or_app. right. assumption.

  intros [ InAM NEQ ].
  destruct (IHa vi InAM) as [ vt InAL ].
  exists vt. right. assumption.

  intros [ InM | [ InAM NEQ ] ].
  destruct (message_to_varsL m vi InM) as [ vt InML ].
  exists vt. right. apply in_or_app. left. assumption.
  destruct (IHa vi InAM) as [ vt InAL ].
  exists vt. right. apply in_or_app. right. assumption.

  intros [ [ EQv | InMM ] | InAM ].
  unfold E.eq, IOT.eq in *.
  destruct v as [ vi' vt ].
  simpl in EQv. subst vi'.
  exists vt. left. reflexivity.
  cut (In vi (message_vars m)).
  intro InM.
  destruct (message_to_varsL m vi InM) as [ vt InML ].
  exists vt. right. apply in_or_app. left. assumption.
  apply message_vars_match_subset. assumption.
  destruct (IHa vi InAM) as [ vt InAL ].
  exists vt. right. apply in_or_app. right. assumption.

  intros [ InFVP | [InAM NInFV] ].
  destruct (formulas_vars_to_proveL f vi InFVP) as [vt InFVPL].
  exists vt. apply in_or_app. left. assumption.
  destruct (IHa vi InAM) as [ vt InAL ].
  exists vt. apply in_or_app. right. assumption.
 Qed.

Hint Resolve action_missing_to_varsL.

 Definition spec_missing_vars s :=
 match s with
 | Spec rspecs a =>
   let all_env :=
    fold_left
     (fun all_env rspec =>
      match rspec with
      | RSpec _ start_decls _ =>
	union all_env (varlist_set start_decls)
      end)
     rspecs
     empty
   in
   (diff (action_missing_vars a)
         all_env)
 end.
	

 Definition gen_var env r v :=	
  match v with
  | Var id t =>
    if mem id env then
       (env, (fun a => a))
    else
       match type_to_new_type t with
       | Some nt => 
         let env' := add id env in
         (env', (fun a => New r v nt a))
       | None =>
         (env, (fun a => a))
       end
  end.

 Lemma gen_var_missing:
  forall (env env':t) (r:id) (vi:id) (vt:type)
    (mka:action -> action),
   gen_var env r (Var vi vt) = (env', mka) ->
   forall (a:action),
    List.In (Var vi vt) (action_varsL a) ->
    (~ In vi (diff (action_missing_vars (mka a)) env))
    \/ (type_to_new_type vt = None).
 Proof.
  intros env env' r vi vt mka.
  simpl.
  case_eq (mem vi env).
  intros MEM EQp a InAL.
  inversion EQp. subst env' mka. clear EQp.
  left. set_iff. rewrite <- mem_iff in MEM.
  intros [ F1 F2 ]. contradiction.
  intros NMEM. rewrite <- not_mem_iff in NMEM.
  case (type_to_new_type vt).
  intros n EQp a InAL.
  inversion EQp. subst env' mka. clear EQp.
  left. simpl. set_iff.
  intuition.
  intros EQp a InAL.
  right. reflexivity.
 Qed.

Hint Resolve gen_var_missing.

 Lemma gen_var_in:
  forall (env env':IdSet.t) (r:id) (vi:id) (vt:type) (nt:new_type)
         (mka:action -> action),
   gen_var env r (Var vi vt) = (env', mka) ->
   type_to_new_type vt = Some nt ->
   In vi env'.
 Proof.
  intros env env' r vi vt nt mka.
  simpl.
  case_eq (mem vi env); [ intro MEM | intro NMEM ].
  intros EQ T.
  inversion EQ. subst env.
  rewrite <- mem_iff in MEM.
  assumption.
  case_eq (type_to_new_type vt); [ intros nt' EQs | intro EQn ].
  intros EQ T.
  inversion EQ.
  set_iff. left. reflexivity.
  intros EQ T.
  discriminate T.
 Qed.

Hint Resolve gen_var_in.

 Lemma gen_var_none:
  forall (env env':IdSet.t) (r:id) (vi:id) (vt:type)
         (mka:action -> action),
   gen_var env r (Var vi vt) = (env', mka) ->
   ~ In vi env' ->
   type_to_new_type vt = None.
 Proof.
  intros env env' r vi vt mka.
  simpl.
  case_eq (mem vi env); [ intro MEM | intro NMEM ].
  intros EQ T.
  inversion EQ. subst env.
  rewrite <- mem_iff in MEM.
  contradiction.
  case_eq (type_to_new_type vt); [ intros nt' EQs | intro EQn ].
  intros EQ T.
  inversion EQ. subst env' mka. clear EQ.
  absurd (In vi (add vi env)). assumption. set_iff. left. reflexivity.
  intros EQ T. reflexivity.
 Qed.

Hint Resolve gen_var_none.

 Lemma gen_var_in2:
  forall (env env':IdSet.t) (r:id) (v:var)
         (mka:action -> action),
   gen_var env r v = (env', mka) ->
   forall (vi:id),
    ~ In vi env ->
    In vi env' ->
    exists vt:type, exists nt:new_type,
     v = Var vi vt
     /\ type_to_new_type vt = Some nt.
 Proof.
  intros env env' r v mka.
  destruct v as [ vi' vt' ].
  simpl.
  case_eq (mem vi' env); [ intro MEM | intro NMEM ].
  intros EQp vi NInE InE'.
  inversion EQp. subst env. contradiction.
  case_eq (type_to_new_type vt'); [ intros nt EQs | intro EQn ].
  intros EQp vi NInE InE'.
  inversion EQp. subst env' mka. clear EQp.
  exists vt'. exists nt.
  rewrite add_iff in InE'.
  unfold E.eq, IOT.eq in *.
  destruct InE' as [ EQ | InE ].
  subst vi. split. reflexivity. assumption.
  contradiction.
  intros EQp vi NInE InE'.
  inversion EQp. subst env. contradiction.
 Qed.

Hint Resolve gen_var_in2.
 
 Lemma gen_var_notin_1:
  forall (env env':IdSet.t) (r:id) (vi:id) (vt:type)
         (mka:action -> action),
   gen_var env r (Var vi vt) = (env', mka) ->
    forall (vi':id),
     vi' <> vi ->
     ~ In vi' env ->
     ~ In vi' env'.
 Proof.
  intros env env' r vi vt mka.
  simpl.
  case_eq (mem vi env); [ intro MEM | intro NMEM ].
  intros EQ vi' EQv NIN.
  inversion EQ. subst env.
  trivial.
  case_eq (type_to_new_type vt); [ intros nt' EQs | intro EQn ].
  intros EQ vi' EQv NIN.
  inversion EQ. subst env' mka. clear EQ.
  set_iff.
  unfold E.eq, IOT.eq.
  intros [ EQvv | IN ].
  apply EQv. symmetry. assumption.
  contradiction.
  intros EQ vi' EQv NIN.
  inversion EQ. subst env. assumption.
 Qed.

Hint Resolve gen_var_notin_1.

 Lemma gen_var_notin_2:
  forall (env env':IdSet.t) (r:id) (vi:id) (vt:type)
         (mka:action -> action),
   gen_var env r (Var vi vt) = (env', mka) ->
   type_to_new_type vt = None ->
   ~ In vi env ->
   ~ In vi env'.
 Proof.
  intros env env' r vi vt mka.
  simpl.
  case_eq (mem vi env); [ intro MEM | intro NMEM ].
  intros EQ T NIN.
  inversion EQ. subst env.
  assumption.
  case_eq (type_to_new_type vt); [ intros nt' EQs | intro EQn ].
  intros EQ T NIN.
  discriminate T.
  intros EQ. inversion EQ.
  intro EQs. trivial.
 Qed.

Hint Resolve gen_var_notin_2.

 Lemma gen_var_monotone:
  forall (env env':IdSet.t) (r:id) (v:var) (mka:action -> action),
   gen_var env r v = (env', mka) ->
   forall (i:id),
    In i env -> In i env'.
 Proof.
  intros env env' r v mka.
  destruct v as [ vi vt ].
  simpl.
  case (mem vi env).
  intros EQ i IN.
  inversion EQ. subst env. assumption.
  case (type_to_new_type vt).
  intros n EQ i IN.
  inversion EQ. set_iff. right. assumption.
  intros EQ i IN.
  inversion EQ. subst env. assumption.
 Qed.

Hint Resolve gen_var_monotone.

 Lemma gen_var_type: 
  forall (env env':IdSet.t) (r:id) (v:var) (mka:action -> action),
   gen_var env r v = (env', mka) ->
   forall (a:action) (ns:IdSet.t) (es:list var),
   type_action ns env' r es a ->
   type_action ns env r es (mka a).
 Proof.
  intros env env' r v mka.
  destruct v as [vi vt].
  simpl.
  case (mem vi env).
  intros EQp a ns es Ta.
  inversion EQp. assumption.

  case_eq (type_to_new_type vt).
  intros n EQn EQp a ns es Ta.
  inversion EQp.
  subst env' mka. clear EQp.
  simpl.
  case (ME.eq_dec r r); [ intro EQ; clear EQ | intro NEQ ].
  split.
 
  revert EQn.
  destruct vt; simpl; intro EQn;
  try (discriminate EQn);
  try (inversion EQn; subst n; clear EQn);
  simpl;
  [ case (type_eq_dec Nonce Nonce)
  | case (type_eq_dec Symkey Symkey)
  | case (type_eq_dec Pubkey Pubkey) ];
  intro EQt;
  try (trivial);
  try (apply EQt; reflexivity).

  assumption.
  absurd_hyp NEQ.
  intro F. apply F. reflexivity.
  assumption.

  intros EQn EQp a ns es Ta.
  inversion EQp. assumption.
 Qed.

Hint Resolve gen_var_type.

 Lemma gen_var_correct:
  forall (env env':IdSet.t) (a:action) (r vi:id) (vt:type)
         (mka:action->action),
   gen_var env r (Var vi vt) = (env', mka) ->
    In vi (diff (action_missing_vars (mka a)) env) ->
     (type_to_new_type vt = None)
     /\ (exists vt:type, List.In (Var vi vt) (action_varsL a)).
 Proof.
  intros env env' a r vi vt mka.
  simpl.
  case_eq (mem vi env).
 
  intros MEM EQp. inversion EQp.
  subst env'. clear EQp H1 mka.
  intro IND. absurd_hyp IND.
  rewrite <- mem_iff in MEM.
  set_iff. intros [ F1 F2 ].
  contradiction. assumption.

  intros NMEM. clear NMEM. 
  case (type_to_new_type vt); [ intro n | idtac ];
  intro EQp; set_iff.

  inversion EQp. subst env' mka. clear EQp.
  simpl. intros [ IN NIN ].
  absurd_hyp IN.
  set_iff. intros [ IN' NEQ ].
  apply NEQ. reflexivity.
  assumption.

  inversion EQp. subst env' mka. clear EQp.
  intros [ IN NIN ]. clear NIN.
  split. reflexivity.
  apply action_missing_to_varsL.
  assumption.
 Qed.

Hint Resolve gen_var_correct.

 Lemma gen_var_inmissing:
  forall (env:IdSet.t) (r:id) (v:var) (env':IdSet.t) (mka:action -> action),
   gen_var env r v = (env', mka) ->
   forall (vi:id) (a:action),
    ~ In vi env' ->
    In vi (action_missing_vars (mka a)) ->
    In vi (action_missing_vars a).
 Proof.
  intros env r v env' mka.
  destruct v as [ vi vt ].
  simpl.
  case_eq (mem vi env); intro MEMvi.
  intro EQ. inversion EQ. subst env' mka.
  trivial.
  rewrite <- not_mem_iff in MEMvi.
  case_eq (type_to_new_type vt); [ intro nt | idtac ]; intro EQvt;
  intro EQ; inversion EQ; subst env' mka; clear EQ;
  [ idtac | trivial ].
  intros vi' a.
  simpl.
  set_iff.
  intuition.
 Qed.

Hint Resolve gen_var_inmissing.

 Lemma gen_var_missingvars:
  forall (env:IdSet.t) (r:id) (v:var) (env':IdSet.t) (mka:action -> action) (a:action),
   gen_var env r v = (env', mka) ->
   Equal (action_missing_vars (mka a))
         (diff (action_missing_vars a) (diff env' env)).
  Proof.
  intros env r v env' mka a.
  destruct v as [ vi vt ].
  simpl.
  case_eq (mem vi env); intro MEMvi.
  intro EQ. inversion EQ. subst env' mka. clear EQ.
  unfold Equal. intro b. set_iff. intuition.
  rewrite <- not_mem_iff in MEMvi.
  case_eq (type_to_new_type vt); [ intro nt | idtac ]; intro EQvt;
  intro EQ; inversion EQ; subst env' mka; clear EQ;
  simpl; unfold Equal; intro b; set_iff;
  unfold E.eq, IOT.eq; intuition.
  subst b.
  apply H3. assumption.
 Qed.

Hint Rewrite -> gen_var_missingvars : ppl.

 Fixpoint gen_vars env r vs {struct vs} : (prod t (action -> action)):=
  match vs with
  | nil => (env, (fun a => a))
  | v :: vs' =>
    let (env', mka) := gen_vars env r vs' in
    let (env'', mka') := gen_var env' r v in
     (env'', (fun a => mka (mka' a)))
  end.

 Lemma gen_vars_in:
  forall (env env':IdSet.t) (r:id) (vs:list var)
         (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
   (forall (vi:id) (vt:type) (nt:new_type),
     List.In (Var vi vt) vs ->
     type_to_new_type vt = Some nt ->
     In vi env').
  Proof.
   intros env env' r vs. revert vs env env'.
   induction vs as [| v vs]; simpl;
   intros env env' mka EQg vi vt nt IN EQt.
   contradiction.
   revert EQg.
   case_eq (gen_vars env r vs). intros env'0 mka0 EQg.
   case_eq (gen_var env'0 r v). intros env'' mka' EQgv.
   intro EQp.
   inversion EQp. clear EQp. subst env' mka.
   rename env'0 into env'.
   rename mka0 into mka.
   destruct IN as [ EQv | IN ].
   subst v.
   apply gen_var_in with env' r vt nt mka'; assumption.

   apply gen_var_monotone with env' r v mka'.
   assumption.
   apply IHvs with env mka vt nt; assumption.
  Qed.

Hint Resolve gen_vars_in.

 Lemma gen_vars_notin_1:
  forall (env env':IdSet.t) (r:id) (vs:list var)
         (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
    forall (vi:id),
     ~ In vi env ->
     ~ (exists vt:type, List.In (Var vi vt) vs) ->
     ~ In vi env'.
 Proof.
  intros env env' r vs. revert vs env env'.
  induction vs as [| v vs ]; simpl;
  intros env env' mka EQg vi NINe NINv.
  inversion EQg. subst env. assumption.
  revert EQg.
  case_eq (gen_vars env r vs). intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v). intros env'' mka' EQgv.
  intro EQp.
  inversion EQp. clear EQp. subst env' mka.
  rename env'0 into env'.
  rename mka0 into mka.
  destruct v as [ vi' vt ].
  case (id_eq_dec vi vi').
  intro EQ. subst vi'.
  intro INe. apply NINv.
  exists vt. left. reflexivity.
  intro NEQ.
  apply gen_var_notin_1 with env' r vi' vt mka'.
  assumption. assumption.
  apply IHvs with env mka.
  assumption. assumption.
  intros [ vt' InVS ].
  apply NINv.
  exists vt'. right. assumption.
 Qed.

Hint Resolve gen_vars_notin_1.

 Lemma gen_vars_in2:
  forall (env env':IdSet.t) (r:id) (vs:list var) (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
   forall (vi:id),
    In vi env' ->
    ~ In vi env ->
     exists vt:type, exists nt:new_type,
     List.In (Var vi vt) vs /\
     type_to_new_type vt = Some nt.
 Proof.
 intros env env' r vs. revert vs env env'.
 induction vs as [| v vs]; simpl;
 intros env env' mka EQgvs vi InE' NInE.
 inversion EQgvs. subst env'. contradiction.
 revert EQgvs.
 case_eq (gen_vars env r vs). intros env'0 mka0 EQg.
 case_eq (gen_var env'0 r v). intros env'' mka' EQgv.
 intro EQp.
 inversion EQp. clear EQp. subst env' mka.
 rename env'0 into env'.
 rename mka0 into mka.
 rename InE' into InE''.
 case (In_dec vi env').
 intro InE'.
 destruct (IHvs env env' mka EQg vi) as [ vt [ nt [ InVS EQnt ] ] ].
 assumption. assumption.
 exists vt. exists nt.
 split. right. assumption. assumption.
 intro NInE'.
 destruct (gen_var_in2 env' env'' r v mka' EQgv vi) as [ vt [ nt [ EQv EQnt ] ] ].
 assumption. assumption.
 exists vt. exists nt.
 split. left. assumption. assumption.
 Qed.
  
Hint Resolve gen_vars_in2.

 Lemma gen_vars_notin_2:
  forall (env env':IdSet.t) (r:id) (vs:list var)
         (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
   (forall (vi:id),
     ~ In vi env ->
     (forall vt:type,
      List.In (Var vi vt) vs ->
      type_to_new_type vt = None) ->
     ~ In vi env').
 Proof.
  intros env env' r vs. revert vs env env'.
  induction vs as [| v vs]; simpl;
  intros env env' mka EQg vi InH EQt.
  inversion EQg. subst env. assumption.
  revert EQg.
  case_eq (gen_vars env r vs). intros env'0 mka0 EQg.
  case_eq (gen_var env'0 r v). intros env'' mka' EQgv.
  intro EQp.
  inversion EQp. clear EQp. subst env' mka.
  rename env'0 into env'.
  rename mka0 into mka.
  
  destruct v as [vi' vt'].
  case (id_eq_dec vi vi').
  intro EQv. subst vi.

  apply gen_var_notin_2 with env' r vt' mka'.
  assumption. apply EQt. left. reflexivity.
  apply IHvs with env mka.
  assumption. assumption.
  intros vt InVS.
  apply EQt. right. assumption.

  intro NEQv.
  apply gen_var_notin_1 with env' r vi' vt' mka'.
  assumption. assumption.
  apply IHvs with env mka.
  assumption. assumption.
  intros vt InVS.
  apply EQt. right. assumption.
 Qed.

Hint Resolve gen_vars_notin_2.

 Lemma gen_vars_type: 
  forall (env env':IdSet.t) (r:id) (vs:list var) (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
   forall (a:action) (ns:IdSet.t) (es:list var),
   type_action ns env' r es a ->
   type_action ns env r es (mka a).
 Proof.
  intros env env' r vs. revert env env'.
  induction vs as [| v vs ]; simpl.

  intros env env' mka EQp a ns es Ta.
  inversion EQp. assumption.

  intros env env' mka.
  case_eq (gen_vars env r vs); intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v); intros env'' mka' EQgv.
  intros EQp a ns es Ta.
  inversion EQp. 
  subst env' mka. clear EQp.
  rename env'0 into env'.
  rename mka0 into mka.
  apply IHvs with env'.
  assumption.
  apply gen_var_type with env'' v;
  assumption.
 Qed.

Hint Resolve gen_vars_type.

 Lemma gen_vars_inmissing:
  forall (env:IdSet.t) (r:id) (vs:list var) (env':IdSet.t) (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
   forall (vi:id) (a:action),
    ~ In vi env' ->
    In vi (action_missing_vars (mka a)) ->
    In vi (action_missing_vars a).
 Proof.
  intros env r vs. revert env r.
  induction vs as [| v vs ]; simpl;
  intros env r env' mka.
  intro EQ. inversion EQ. subst env' mka. clear EQ.
  trivial.
  case_eq (gen_vars env r vs); intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v); intros env'' mka' EQgv.
  intros EQp.
  inversion EQp. 
  subst env' mka. clear EQp.
  rename env'0 into env'.
  rename mka0 into mka.
  intros vi a NIn InAM.
  apply gen_var_inmissing with env' r v env'' mka'.
  assumption. assumption.
  cut (~ In vi env').
  intro NIn'.
  apply IHvs with env r env' mka.
  assumption. assumption.
  assumption.
  clear InAM IHvs.
  intro F.
  apply NIn.
  apply gen_var_monotone with env' r v mka'.
  assumption.
  assumption.
 Qed.

Hint Resolve gen_vars_inmissing.

 Lemma gen_vars_monotone:
  forall (env env':IdSet.t) (r:id) (vs:list var) (mka:action -> action),
   gen_vars env r vs = (env', mka) ->
   forall (i:id),
    In i env -> In i env'.
 Proof.
  intros env env' r vs. revert env env' r.
  induction vs as [| v vs ]; simpl;
  intros env env' r mka.
  intro EQ. inversion EQ. subst env' mka. clear EQ.
  trivial.
  case_eq (gen_vars env r vs); intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v); intros env'' mka' EQgv.
  intros EQp.
  inversion EQp. 
  subst env' mka. clear EQp.
  rename env'0 into env'.
  rename mka0 into mka.
  intros i InE.
  apply gen_var_monotone with env' r v mka'.
  assumption.
  apply IHvs with env r mka.
  assumption.
  assumption.
 Qed.

Hint Resolve gen_vars_monotone.

 Lemma gen_vars_missingvars:
  forall (env:IdSet.t) (r:id) (vs:list var) (env':IdSet.t) (mka:action -> action) (a:action),
   gen_vars env r vs = (env', mka) ->
   Equal (action_missing_vars (mka a))
         (diff (action_missing_vars a) (diff env' env)).
 Proof.
  intros env r vs. revert env r.
  induction vs as [| v vs ]; simpl;
  intros env r env' mka a.
  intro EQ. inversion EQ. subst env' mka. clear EQ.
  unfold Equal. intro b. set_iff. intuition.
  case_eq (gen_vars env r vs); intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v); intros env'' mka' EQgv.
  intros EQp.
  inversion EQp. 
  subst env' mka. clear EQp.
  rename env'0 into env'.
  rename mka0 into mka.
  rewrite (IHvs env r env' mka).
  assumption.
  rewrite (gen_var_missingvars env' r v env'' mka').
  assumption.
  unfold Equal. intro b. set_iff.
  split.
  intros [ [ InBAM NIn1 ] NIn2 ].
  split. assumption.
  intros [ InBE'' NInB ].
  case (In_dec b env'); [ intro InBE' | intro NInBE' ].
  apply NIn2. split; assumption.
  apply NIn1. split; assumption.
  intros [ InBAM NIn1 ].
  split. split. assumption.
  intros [ InBE'' NInB' ].
  apply NIn1; clear NIn1. split. assumption.
  intro InBE.
  absurd (In b env'). assumption.
  apply gen_vars_monotone with env r vs mka; assumption.
  intros [ InBE' NInB ].
  apply NIn1; clear NIn1.
  split; [ idtac | assumption ].
  apply gen_var_monotone with env' r v mka'; assumption.
 Qed.

Hint Rewrite -> gen_vars_missingvars : ppl.

 Lemma gen_vars_correct:
  forall (env env':IdSet.t) (a:action) (r:id) (vs:list var)
         (mka:action->action),
   gen_vars env r vs = (env', mka) ->
    forall (vi:id) (vt:type),
     List.In (Var vi vt) vs ->
     (count_occ var_eq_dec vs (Var vi vt) = S O) ->
     (forall (vt':type), vt <> vt' -> ~ List.In (Var vi vt') vs) ->
     In vi (diff (action_missing_vars (mka a)) env) ->
      (type_to_new_type vt = None)
      /\ (exists vt:type, List.In (Var vi vt) (action_varsL a)).
 Proof.
  intros env env' a r vs. revert vs env env' a.
  induction vs as [| v vs ]; simpl.
  intros. contradiction.
  intros env env' a mka.
  case_eq (gen_vars env r vs); intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v); intros env'' mka' EQgv.
  intros EQp.
  inversion EQp. 
  subst env' mka. clear EQp.
  rename env'0 into env'.
  rename mka0 into mka.
  intros vi vt.

  case (var_eq_dec v (Var vi vt));
  [ intro EQ; subst v
  | intro NEQ ].
  intro IN. clear IN.
  intros OCC NIN InD.
  apply gen_var_correct with env' env'' r mka'.
  assumption.
  revert InD.
  set_iff.
  intros [ InAM NInE ].
  cut (~ In vi env').
  intro NInE'.
  split; [ idtac | assumption ].
  apply gen_vars_inmissing with env r vs env' mka;
  assumption.

  apply gen_vars_notin_1 with env r vs mka.
  assumption. assumption.
  intros [ vt' InVS ].
  case (type_eq_dec vt vt'); [ intro EQ; subst vt' | intro NEQ ].
  absurd (count_occ var_eq_dec vs (Var vi vt) > 0).
  inversion OCC.
  apply Arith.Gt.gt_irrefl.
  rewrite <- count_occ_In. assumption.
  unfold not in NIN.
  apply NIN with vt'.
  intro F. contradiction.
  right. assumption.

  intros [ EQv | InVS ].
  contradiction.
  intros OCC NIN InD.
  apply IHvs with env env' mka; clear IHvs.
  assumption. assumption.
  assumption.
  intros vt' NEQvt.
  intro InVSt.
  unfold not in NIN.
  apply NIN with vt'.
  unfold not in NEQvt. assumption.
  right. assumption.
  clear NIN.
  revert InD.
  set_iff.
  intros [ InAM NInE ].
  split; [ idtac | assumption ].

  revert InAM.
  rewrite (gen_vars_missingvars env r vs env').
  assumption.
  rewrite (gen_var_missingvars env' r v env'').
  assumption.
  rewrite (gen_vars_missingvars env r vs env').
  assumption.

  set_iff. intuition.
 Qed.

Hint Resolve gen_vars_correct.

 Lemma gen_vars_none:
  forall (env env':IdSet.t) (r:id) (vs:list var)
         (mka:action->action),
   gen_vars env r vs = (env', mka) ->
    forall (vi:id) (vt:type),
     List.In (Var vi vt) vs ->
     ~ In vi env' ->
     type_to_new_type vt = None.
 Proof.
  intros env env' r vs. revert env env' r.
  induction vs as [|v vs]; simpl.
  intros. contradiction.
  intros env env' r mka.
  case_eq (gen_vars env r vs); intros env'0 mka0 EQgvs.
  case_eq (gen_var env'0 r v); intros env'' mka' EQgv.
  intros EQp.
  inversion EQp. 
  subst env' mka. clear EQp.
  rename env'0 into env'.
  rename mka0 into mka.
  intros vi vt.
  intros [ EQv | InVS ] NInE''.
  subst v.
  apply gen_var_none with env' env'' r vi mka'; assumption.
  apply IHvs with env env' r mka vi.
  assumption. assumption.
  intro InE'. apply NInE''.  
  apply gen_var_monotone with env' r v mka'; assumption.
 Qed.

Hint Resolve gen_vars_none.

 Definition genable_var v :=
 match v with
 | Var _ vt =>
 match type_to_new_type vt with
 | Some nt => true
 | None => false
 end
 end.

 Lemma gen_var_envT:
  forall (env env':IdSet.t) (r:id) (v:var) (mka:action -> action),   
   gen_var env r v = (env', mka) ->
   genable_var v = true ->
   Equal env' (add (varid v) env).
 Proof.
  intros env env' r v mka.
  destruct v as [vi vt].
  simpl.
  case_eq (mem vi env).
  intro InE.
  rewrite <- mem_iff in InE.
  intros EQp. inversion EQp. subst. clear EQp.
  case (type_to_new_type vt).
  intros nt.
  symmetry. apply add_equal. assumption.
  intros EQp. discriminate EQp.
  intro NInE.
  case (type_to_new_type vt).
  intros nt.
  intros EQp. inversion EQp. subst. clear EQp.
  intros EQt. apply eq_refl.
  intros EQp. inversion EQp. subst. clear EQp.
  intro F. discriminate F.
 Qed.

Hint Rewrite -> gen_var_envT : ppl.

 Lemma gen_var_envF:
  forall (env env':IdSet.t) (r:id) (v:var) (mka:action -> action),   
   gen_var env r v = (env', mka) ->
   genable_var v = false ->
   Equal env' env.
 Proof.
  intros env env' r v mka.
  destruct v as [vi vt].
  simpl.
  case_eq (mem vi env).
  intro InE.
  rewrite <- mem_iff in InE.
  intros EQp. inversion EQp. subst. clear EQp.
  case (type_to_new_type vt).
  intros nt.
  intro F. discriminate F.
  intro EQf. apply eq_refl.
  intros NInE.
  case (type_to_new_type vt).
  intros nt.
  intros EQp. inversion EQp. subst. clear EQp.
  intro F. discriminate F.
  intros EQp. inversion EQp. subst. clear EQp.
  intro EQf. apply eq_refl.
 Qed.

Hint Rewrite -> gen_var_envF : ppl.

Definition genable_vars := List.filter genable_var.

Lemma gen_vars_env:
  forall (env env':IdSet.t) (r:id) (vs:list var) (mka:action -> action),   
   gen_vars env r vs = (env', mka) ->
   Equal env' (union env (varlist_set (genable_vars vs))).
 Proof.
  intros env env' r vs. revert vs env env' r.
  induction vs as [|v vs]; simpl; intros env env' r mka.
  intro EQp. inversion EQp. subst. clear EQp.
  unfold Equal. intro a. set_iff. intuition.
  case_eq (gen_vars env r vs).
  intros _env' _mka EQgvs.
  case_eq (gen_var _env' r v).
  intros _env'' _mka' EQgv.
  intros EQp. inversion EQp. subst. clear EQp.
  case_eq (genable_var v); intro EQgnv.
  simpl.
  rewrite (gen_var_envT _env' env' r v _mka').
  assumption. assumption.
  rewrite (IHvs env _env' r _mka).
  assumption.
  unfold Equal. intro b. set_iff.
  intuition.
  rewrite (gen_var_envF _env' env' r v _mka').
  assumption. assumption.
  rewrite (IHvs env _env' r _mka).
  assumption.
  unfold Equal. intro b. set_iff. intuition.
 Qed.

Hint Rewrite -> gen_vars_env : ppl.

Definition SetNonOverlap s s' :=
  forall x, In x s -> ~ In x s'.

Theorem SetNonOverlap_dec:
  forall s s', {SetNonOverlap s s'} + {~ SetNonOverlap s s'}.
Proof.
  unfold SetNonOverlap.
  induction s using set_induction; intros s'.
  left. intros x.
  cut (Equal s empty). intros EQ.
  rewrite EQ. set_iff. contradiction.
  apply empty_is_empty_1. assumption.
  unfold Add in *. rename H0 into A12.
  case (In_dec x s'); intros IN'.
  right. intros Op'.
  apply Op' with x. rewrite A12.
  intuition. assumption.
  destruct (IHs1 s') as [IH | NIH].
  left. intros y. rewrite A12.
  intros [EQx | IN1].
  subst y. assumption.
  apply IH. assumption.
  right. intros Op'.
  apply NIH.
  intros y IN1. apply Op'.
  rewrite A12. right. assumption.
Qed.

 Fixpoint action_norecurbind a :=
  match a with
    | Return =>
      True
    | Comm from_r to_r ffs msg tfs sa =>
      action_norecurbind sa
    | New r v t sa =>
      action_norecurbind sa
    | Bind r v msg sa =>
      ~ In (varid v) (message_vars msg)
      /\ action_norecurbind sa
    | Match r v msg sa =>
      action_norecurbind sa
    | Derive r f sa =>
      SetNonOverlap (formulas_vars_prove f) (formulas_vars_bound f)
      /\ action_norecurbind sa
  end.

 Lemma action_norecurbind_dec:
  forall (a:action),
  {action_norecurbind a} + {~ action_norecurbind a}.
 Proof.
  simple induction a; clear a; simpl;
  [ idtac
  | intros fr tr ffs m tfs a IHa
  | intros ar v nt a IHa
  | intros ar v m a IHa
  | intros ar v m a IHa
  | intros ar f a IHa ];
  try (left; trivial; fail);
  try (apply IHa).

  destruct (In_dec (varid v) (message_vars m)) as [InV | NInV].
  intuition.
  destruct IHa as [Ha | NHa]; intuition.

  destruct IHa as [Ha | NHa]; [ idtac | intuition ].
  set (fvp := (formulas_vars_prove f)).
  set (fvb := (formulas_vars_bound f)).
  destruct (SetNonOverlap_dec fvp fvb); intuition.
Qed.

(* Last actions are ignored, because a protocol goal should not be a
   `new' value *)
 Fixpoint gen_action all_env a {struct a} : action :=
  match a with
    | Return =>
      Return
    | Comm from_r to_r ffs msg tfs sa =>
      let missing := message_varsL msg in
      let (all_env', mka) := gen_vars all_env from_r missing in
      let sa' := gen_action all_env' sa in
     mka (Comm from_r to_r ffs msg tfs sa')
  | New r v t sa =>
    let all_env' := add (varid v) all_env in
    let sa' := gen_action all_env' sa in 
     New r v t sa'
  | Bind r v msg sa =>
    let missing := message_varsL msg in
    let all_env' := add (varid v) all_env in
    let (all_env'', mka) := gen_vars all_env' r missing in
    let sa' := gen_action all_env'' sa in
     mka (Bind r v msg sa')
  | Match r v msg sa =>
    let missing := v :: (message_vars_matchL msg) in
    let (all_env', mka) := gen_vars all_env r missing in
    let sa' := gen_action all_env' sa in
     mka (Match r v msg sa')
  | Derive r f sa =>
    let missing := formulas_vars_proveL f in
    let new_decls := formulas_vars_bound f in
    let all_env' := union new_decls all_env in
    let (all_env'', mka) := gen_vars all_env' r missing in
    let sa' := gen_action all_env'' sa in
     mka (Derive r f sa')
  end.

 Lemma gen_action_correct:
  forall (env:IdSet.t) (a:action) (vi:id),
   action_norecurbind a ->
   In vi (diff (action_missing_vars (gen_action env a)) env) ->
    exists vt:type,
     (type_to_new_type vt = None)
     /\ (List.In (Var vi vt) (action_varsL a)).
 Proof.
  intros env a. revert a env.
  simple induction a; clear a; simpl;
  [ idtac
  | intros fr tr ffs m tfs a IHa
  | intros ar v nt a IHa
  | intros ar v m a IHa
  | intros ar v m a IHa
  | intros ar f a IHa ];
  intros env vi ANRB;
  set_iff.

  (* Case 1 *)
  intros [F NInE]. contradiction.

  (* Case 2 *)
  case_eq (gen_vars env fr (message_varsL m));
  intros env' mka EQgvs.
  rewrite (gen_vars_missingvars env fr (message_varsL m) env' mka).
  assumption.
  simpl. set_iff.
  intros [ [[InMV | InAM] NIn1] NInE ].

  (* Case 2a *)
  clear IHa.
  destruct (message_to_varsL m vi) as [vt InML]. assumption.
  exists vt.
  split; [ idtac | apply in_or_app; left; assumption ].
  apply gen_vars_none with env env' fr (message_varsL m) mka vi.
  assumption. assumption.
  intro InE'. apply NIn1. split; assumption.

  (* Case 2b *)
  destruct (IHa env' vi) as [vt [EQn InAL]].
  assumption.
  set_iff. split. assumption.
  intro InE'. apply NIn1. split; assumption.
  exists vt;
  split;
  [ assumption
  | apply in_or_app; right; assumption ].

  (* Case 3 *)
  intros [[InAm NEQv] NInE].
  destruct (IHa (add (varid v) env) vi) as [vt [ EQn InAL ] ].
  assumption.
  set_iff.
  split. assumption. intros [EQv | InE]; contradiction.
  exists vt.
  split. assumption. right. assumption.

  (* Case 4 *)
  case_eq (gen_vars (add (varid v) env) ar (message_varsL m));
  intros env' mka EQgvs.
  rewrite (gen_vars_missingvars (add (varid v) env) ar (message_varsL m) env' mka).
  assumption.
  simpl.
  assert (In (varid v) env') as InE'.
  apply gen_vars_monotone with (add (varid v) env) ar (message_varsL m) mka.
  assumption. set_iff. left. reflexivity.
  destruct v as [ va vat ]. simpl in *.
  rename vi into vb.
  set_iff. unfold E.eq, IOT.eq.
  intros [ [InOR NIn1] NInE ].
  destruct ANRB as [ NInMV ANRB ].
  destruct InOR as [InMV | [InAM NEQv]].

  (* Case 4a *)
  clear IHa.
  destruct (message_to_varsL m vb) as [vbt InML]. assumption.
  exists vbt.
  split; [ idtac | right; apply in_or_app; left; assumption ].
  apply gen_vars_none with (add va env) env' ar (message_varsL m) mka vb.
  assumption. assumption.
  intro InBE'. apply NIn1; clear NIn1.
  split. assumption.
  intros [ EQv | InE ]; [ idtac | contradiction ].
  subst vb. clear InBE'.
  contradiction.

  (* Case 4b *)
  destruct (IHa env' vb) as [vbt [EQn InAL]].
  assumption.
  set_iff.
  split. assumption.
  intro InBE'. apply NIn1; clear NIn1.
  split. assumption.
  intros [ EQv | InE ].
  subst vb. clear InBE'.
  absurd (va = va). assumption. reflexivity.
  contradiction.
  exists vbt.
  split. assumption.
  right. apply in_or_app. right. assumption.

  (* Case 5 *)
  case_eq (gen_vars env ar (message_vars_matchL m));
  intros env' mka EQgvs.
  case_eq (gen_var env' ar v);
  intros env'' mka' EQgv.
  rewrite (gen_vars_missingvars env ar (message_vars_matchL m) env' mka).
  assumption.
  rewrite (gen_var_missingvars env' ar v env'' mka').
  assumption.
  simpl. set_iff.
  intros [ [ [ [ [ EQvi | InMVM ] | InAM ] NIn2 ] NIn1 ] NInE ].
  
  (* Case 5a *)
  unfold E.eq, IOT.eq in EQvi; subst vi.
  revert NIn2 NIn1 NInE.
  destruct v as [vi vt].
  simpl.
  intros NIn2 NIn1 NInE.
  case (In_dec vi env'); [ intro InE' | intro NInE' ].
  assert False. apply NIn1. split; assumption. contradiction.
  case (In_dec vi env''); [ intro InE'' | intro NInE'' ].  
  assert False. apply NIn2. split; assumption. contradiction.
  exists vt.
  split; [ idtac | left; reflexivity ].
  apply gen_var_none with env' env'' ar vi mka'.
  assumption. assumption.

  (* Case 5b *)
  case (In_dec vi env'); [ intro InE' | intro NInE' ].
  assert False. apply NIn1. split; assumption. contradiction.
  case (In_dec vi env''); [ intro InE'' | intro NInE'' ].  
  assert False. apply NIn2. split; assumption. contradiction.
  clear NIn1 NIn2.
  destruct (message_to_match_varsL m vi) as [vt InML]. assumption.
  exists vt.
  split.
  apply gen_vars_none with env env' ar (message_vars_matchL m) mka vi.
  assumption. assumption. assumption.
  right. apply in_or_app. left.
  apply message_match_in_varsL. assumption.

  (* Case 5c *)
  case (In_dec vi env'); [ intro InE' | intro NInE' ].
  assert False. apply NIn1. split; assumption. contradiction.
  case (In_dec vi env''); [ intro InE'' | intro NInE'' ].  
  assert False. apply NIn2. split; assumption. contradiction.
  clear NIn1 NIn2.
  destruct (IHa env'' vi) as [vt [EQn InAL]].
  assumption.
  set_iff. split; assumption.
  exists vt.
  split. assumption.
  right. apply in_or_app. right. assumption.

  (* Case 6 *)
  case_eq (gen_vars (union (formulas_vars_bound f) env)
            ar (formulas_vars_proveL f));
  intros env' mka EQgvs.
  rewrite (gen_vars_missingvars (union (formulas_vars_bound f) env)
   ar (formulas_vars_proveL f) env' mka).
  assumption. simpl. set_iff.
  destruct ANRB as [NoRec ANRB].
  intros [ [[InFVP | [InAM NInFVB]] H] NInE ].

  (* Case 6a *)
  destruct (formulas_vars_to_proveL f vi InFVP) as [vt InFL].
  exists vt.
  split.
  apply gen_vars_none with (union (formulas_vars_bound f) env)
   env' ar (formulas_vars_proveL f) mka vi.
  assumption. assumption. 
  intro InE'. apply H. clear H. split. assumption.
  intros [InFVB | InE].
  apply NoRec with vi; assumption.
  contradiction.
  apply in_or_app. left. assumption.

  (* Case 6b *)
  destruct (IHa env' vi ANRB)
  as [vt [EQn InAL]].
  set_iff. split. assumption.
  intro InE'. apply H.
  split. assumption.
  intros [InFVB | InE]; contradiction.

  exists vt. split. assumption.
  apply in_or_app. right. assumption.
 Qed.

Hint Resolve gen_action_correct.

 Definition gen_spec s :=
 match s with
 | Spec rspecs a =>
   let all_env :=
    fold_left
     (fun all_env rspec =>
      match rspec with
      | RSpec _ start_decls _ =>
	union all_env (varlist_set start_decls)
      end)
     rspecs
     empty
   in
   let a' := gen_action all_env a in
    Spec rspecs a'
 end.

 Lemma gen_spec_correct:
  forall (rss:list role_spec) (a:action) (vi:id),
   action_norecurbind a ->
   In vi (spec_missing_vars (gen_spec (Spec rss a))) ->
    exists vt:type,
     (type_to_new_type vt = None)
     /\ (List.In (Var vi vt) (spec_varsL (Spec rss a))).
 Proof.
  intros rss a vi ANRB.
  simpl.
  set (env := fold_left
     (fun all_env rspec =>
      match rspec with
      | RSpec _ start_decls _ =>
	union all_env (varlist_set start_decls)
      end)
     rss
     empty).
  intros InD.
  destruct (gen_action_correct env a vi) as [vt [EQn InAVL]].
  assumption. assumption.
  exists vt.
  split. assumption.
  apply in_or_app. right. assumption.
 Qed.
Hint Resolve gen_spec_correct.

(* Bindc *)
Fixpoint message_replace i rm m { struct m } :=
  if msg_eq_dec rm m then
    MVar (Var i Msg)
    else
      match m with 
        | MNil => MNil
        | MVar v => MVar v
        | MConst c => MConst c
        | Join m1 m2 =>
          Join (message_replace i rm m1) (message_replace i rm m2)
        | Hash m => 
          Hash (message_replace i rm m)
        | PubEnc v m => 
          PubEnc v (message_replace i rm m)
        | PubSign v m =>
          PubSign v (message_replace i rm m)
        | SymEnc v m => 
          SymEnc v (message_replace i rm m)
        | SymSign v m =>
          SymSign v (message_replace i rm m)
        | Common.Bind v m =>
          Common.Bind v (message_replace i rm m)
      end.

Lemma message_replace_type_send:
  forall (i:id) (rm:msg) (m:msg) (vs:IdSet.t),
    type_msg_send vs m ->
    type_msg_send (add i vs) (message_replace i rm m).
Proof.
  intros i rm m vs.

   revert m.
   simple induction m; clear m;
   [ idtac
   | intros v
   | intros c
   | intros m1 IHm1 m2 IHm2
   | intros m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1 ];
   simpl;
   [ case (msg_eq_dec rm MNil)
   | case (msg_eq_dec rm (MVar v))
   | case (msg_eq_dec rm (MConst c))
   | case (msg_eq_dec rm (Join m1 m2))
   | case (msg_eq_dec rm (Hash m1))
   | case (msg_eq_dec rm (SymEnc v m1))
   | case (msg_eq_dec rm (SymSign v m1))
   | case (msg_eq_dec rm (PubEnc v m1))
   | case (msg_eq_dec rm (PubSign v m1))
   | case (msg_eq_dec rm (Common.Bind v m1)) ];
   simpl;
   unfold exist_var;
   simpl; set_iff;
   intros EQ Tm;
   intuition.
  Qed.

Hint Resolve message_replace_type_send.

  Lemma message_replace_type_recv:
   forall (i:id) (rm:msg) (m:msg) (vs:IdSet.t),
    type_msg_recv vs m ->
    type_msg_recv (add i vs) (message_replace i rm m).
  Proof.
   intros i rm m vs.

   revert m.
   simple induction m; clear m;
   [ idtac
   | intros v
   | intros c
   | intros m1 IHm1 m2 IHm2
   | intros m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1 ];
   simpl;
   [ case (msg_eq_dec rm MNil)
   | case (msg_eq_dec rm (MVar v))
   | case (msg_eq_dec rm (MConst c))
   | case (msg_eq_dec rm (Join m1 m2))
   | case (msg_eq_dec rm (Hash m1))
   | case (msg_eq_dec rm (SymEnc v m1))
   | case (msg_eq_dec rm (SymSign v m1))
   | case (msg_eq_dec rm (PubEnc v m1))
   | case (msg_eq_dec rm (PubSign v m1))
   | case (msg_eq_dec rm (Common.Bind v m1)) ];
   simpl;
   unfold exist_var;
   simpl; set_iff;
   intros EQ Tm;
   intuition.
  Qed.

Hint Resolve message_replace_type_recv.

  Fixpoint message_replace_action i rm a {struct a} :=
  match a with
  | Return => 
    Return
  | Comm f t ffs m tfs a =>
    Comm f t ffs (message_replace i rm m)  tfs
     (message_replace_action i rm a)
  | New ar v nt a =>
    New ar v nt
      (message_replace_action i rm a)
  | Bind ar v m a =>
    Bind ar v (message_replace i rm m) 
      (message_replace_action i rm a)
  | Match ar v m a =>
    Match ar v (message_replace i rm m) 
      (message_replace_action i rm a)
  | Derive ar f a =>
    Derive ar f
      (message_replace_action i rm a)
  end.

  Fixpoint messages_replace_action rs a :=
  match rs with
  | nil => a
  | (i,rm) :: rs => 
    let a' := message_replace_action i rm a in
    messages_replace_action rs a'
  end.

  Fixpoint bindc_msg new_id m {struct m} 
  : (prod (list (prod id msg)) msg) :=	
  match m with 
  | MNil =>	
    (nil, MNil)
  | MVar v =>
    (nil, MVar v)
  | MConst c =>
    (nil, MConst c)
  | Join m1 m2 =>
    let (mrs1, m1') := bindc_msg new_id m1 in
    let (mrs2, m2') := bindc_msg new_id m2 in
     (mrs1 ++ mrs2, Join m1' m2')
  | Hash m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg new_id m in
     ((ni, Hash m') :: mrs, MVar (Var ni Msg))
  | SymEnc v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg new_id m in
     ((ni, SymEnc v m') :: mrs, MVar (Var ni Msg))
  | SymSign v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg new_id m in
     ((ni, SymSign v m') :: mrs, MVar (Var ni Msg))
  | PubEnc v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg new_id m in
     ((ni, PubEnc v m') :: mrs, MVar (Var ni Msg))
  | PubSign v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg new_id m in
     ((ni, PubSign v m') :: mrs, MVar (Var ni Msg))
  | Common.Bind v m =>
    let (mrs, m') := bindc_msg new_id m in
     (mrs, Common.Bind (Var (varid v) Msg) m')
  end.

  Fixpoint bindc_replaces r drms a {struct drms} :=
  match drms with
  | nil => a
  | (i, rm) :: drms =>
    bindc_replaces r drms
     (Bind r (Var i Msg) rm a)
  end.

  Fixpoint bindc_action n new_id a {struct n} :=
  match n with 
  | O => a
  | S n =>
  match a with
  | Return =>
      Return
  | Comm f t ffs m tfs a =>
   let (replaces, m') := bindc_msg new_id m in
   let a' := messages_replace_action replaces a in
   let a'' := bindc_action n new_id a' in
    bindc_replaces f replaces (Comm f t ffs m' tfs a'')
  | New r v t a =>
   let a' := bindc_action n new_id a in
    New r v t a'
  | Bind r v m a =>
   let (replaces, m') := bindc_msg new_id m in
   let a' := messages_replace_action replaces a in
   let a'' := bindc_action n new_id a' in
    bindc_replaces r replaces (Bind r v m' a'')
  | Match r v m a =>
   let (replaces, m') := bindc_msg new_id m in
   let a' := messages_replace_action replaces a in
   let a'' := bindc_action n new_id a' in
    bindc_replaces r replaces (Match r v m' a'')
  | Derive r t a =>
   let a' := bindc_action n new_id a in
    Derive r t a'
  end
  end.
  
  Fixpoint action_size a {struct a} :=
  match a with
  | Return => O
  | Comm f t ffs m tfs a => S (action_size a)
  | New r v t a => S (action_size a)
  | Bind r v m a => S (action_size a)
  | Match r v m a => S (action_size a)
  | Derive r t a => S (action_size a)
  end.

  Lemma message_replace_action_size:
   forall (i:id) (rm:msg) (a:action),
    action_size a = action_size (message_replace_action i rm a).
  Proof.
   intros i rm.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar v nt a IHa
   | intros ar v m a IHa
   | intros ar v m a IHa
   | intros ar f a IHa ];
   try (rewrite IHa);
   reflexivity.
  Qed.

Hint Rewrite <- message_replace_action_size : ppl.

  Lemma messages_replace_action_size:
   forall (replaces:list (prod id msg)) (a:action),
    action_size a
    = action_size (messages_replace_action replaces a).
  Proof.
   induction replaces as [|r rs]; simpl.
   trivial.
   destruct r as [i rm].
   intro a.
   rewrite (message_replace_action_size i rm a).
   apply IHrs.
  Qed.

Hint Rewrite <- messages_replace_action_size : ppl.

  Definition bindc_action_auto new_id a :=
   bindc_action (action_size a) new_id a.

  Fixpoint type_action_recv r renv a :=
  match a with 	
  | Return => True
  | Comm fr tr ffs m tfs a =>
    if id_eq_dec r tr then
     let renv' := union renv (formulas_vars_bound ffs) in
     let renv'' := union renv' (message_vars_bound m) in
     (type_msg_recv renv m)
     /\ (type_action_recv r renv'' a)
    else
     type_action_recv r renv a
  | New ar v nt a =>
    if id_eq_dec r ar then
     let renv' := add (varid v) renv in
     type_action_recv r renv' a
    else
     type_action_recv r renv a
  | Bind ar v m a =>
    if id_eq_dec r ar then
     let renv' := add (varid v) renv in
     type_action_recv r renv' a
    else
     type_action_recv r renv a
  | Match ar v m a =>
    if id_eq_dec r ar then
     let renv' := union renv (message_vars_bound m) in
     type_action_recv r renv' a
    else
     type_action_recv r renv a
  | Derive ar phi a =>
    if id_eq_dec r ar then
     let renv' := union renv (formulas_vars_bound phi) in
     type_action_recv r renv' a
    else
     type_action_recv r renv a
  end.

  Lemma type_action_recv_subset:
   forall (a:action) (r:id) (renv renv':IdSet.t),
    Subset renv renv' ->
    type_action_recv r renv a ->
    type_action_recv r renv' a.
  Proof.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar v nt a IHa
   | intros ar v m a IHa
   | intros ar v m a IHa
   | intros ar f a IHa ];
   intros r renv renv' SUB.

   trivial.

   case (id_eq_dec r tr); [ intro EQ; subst tr | intro NEQ; apply IHa; assumption ].
   intros [ Tmr Tar ].
   split.
   apply type_msg_recv_subset with renv; assumption.
   apply IHa with (union (union renv (formulas_vars_bound ffs))
     (message_vars_bound m)).
   unfold Subset. intro b. set_iff. unfold Subset in SUB.
   intuition.
   assumption.
   
   case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ; apply IHa; assumption ].
   intros Tar.
   apply IHa with (add (varid v) renv).
   unfold Subset. intro b. set_iff. unfold Subset in SUB.
   intuition.
   assumption.

   case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ; apply IHa; assumption ].
   intros Tar.
   apply IHa with (add (varid v) renv).
   unfold Subset. intro b. set_iff. unfold Subset in SUB.
   intuition.
   assumption.

   case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ; apply IHa; assumption ].
   intros Tar.
   apply IHa with (union renv (message_vars_bound m)).
   unfold Subset. intro b. set_iff. unfold Subset in SUB.
   intuition.
   assumption.

   case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ; apply IHa; assumption ].
   intros Tar.
   apply IHa with (union renv (formulas_vars_bound f)).
   unfold Subset. intro b. set_iff. unfold Subset in SUB.
   intuition.
   assumption.
  Qed.

Hint Resolve type_action_recv_subset.
  
  Lemma bindc_replaces_type:
   forall (a:action) (r1 r2:id) (renv:IdSet.t)
          (replaces:list (prod id msg)),
    type_action_recv r1 renv a ->
    type_action_recv r1 renv (bindc_replaces r2 replaces a).
  Proof.
   intros a r1 r2 renv replaces. revert a r1 r2 renv.
   induction replaces as [|r rs]; simpl.
   trivial.
   destruct r as [i rm].
   intros a r1 r2 renv Ta.
   apply IHrs.
   simpl.
   case (id_eq_dec r1 r2).
   intro EQ. subst r2.
   apply type_action_recv_subset with renv.
   unfold Subset. intro b. set_iff. intuition.
   assumption.
   intro NEQ.
   assumption.
  Qed.

Hint Resolve bindc_replaces_type.

  Lemma bindc_msg_type:
   forall (m m':msg) (new_id:unit -> id)
     (replaces:list (id*msg)) (renv:IdSet.t),
    bindc_msg new_id m = (replaces, m') ->
    type_msg_recv renv m'.
  Proof.
   simple induction m; clear m;
   [ idtac
   | intros v
   | intros c
   | intros m1 IHm1 m2 IHm2
   | intros m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1
   | intros v m1 IHm1 ];
   intros m' newid replaces renv;
   simpl;
   try (case_eq (bindc_msg newid m1); intros mrs1 m1' EQm1');
   try (case_eq (bindc_msg newid m2); intros mrs2 m2' EQm2');
   intro EQ; inversion EQ; subst replaces m'; clear EQ;
   simpl;
   try (trivial).

   split. apply IHm1 with newid mrs1. assumption.
   apply IHm2 with newid mrs2. assumption.

   split. case (type_eq_dec Msg Msg); intro C.
   trivial. absurd (Msg = Msg). assumption. reflexivity.
   apply IHm1 with newid mrs1. assumption.
  Qed.

Hint Resolve bindc_msg_type.

  Lemma bindc_action_type:
   forall (n:nat) (a:action) (r:id) (renv:IdSet.t) (new_id:unit -> id),
    action_size a <= n ->
    type_action_recv r renv (bindc_action n new_id a).
  Proof.
   induction n as [|n].
 
   simpl.
   destruct a; intros r renv newid; simpl; intro LE;
   try (trivial);
   (absurd (S (action_size a) <= 0);
   [ apply Arith.Le.le_Sn_O
   | assumption ]).

   destruct a as 
   [ 
   | fr tr ffs m tfs a 
   | ar v nt a
   | ar v m a
   | ar v m a
   | ar f a ]; simpl;
   intros r renv newid LE;
   try (trivial);
   assert (action_size a <= n) as LE2
   by (apply Arith.Le.le_S_n; assumption);
   clear LE;
   try (case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ ]);
   try (apply IHn; assumption);
   case_eq (bindc_msg newid m); intros replaces m' BEQ;
   apply bindc_replaces_type;
   simpl;
   [ case (id_eq_dec r tr); [ intro EQ; subst tr | intro NEQ ]
   | case (id_eq_dec r r); [ intro EQ; clear EQ | intro NEQ; absurd (r = r); [ assumption | reflexivity ] ]
   | case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ' ]
   | case (id_eq_dec r r); [ intro EQ; clear EQ | intro NEQ; absurd (r = r); [ assumption | reflexivity ] ]
   | case (id_eq_dec r ar); [ intro EQ; subst ar | intro NEQ' ] ];
   try (apply IHn; rewrite <- messages_replace_action_size; assumption).

   split. apply bindc_msg_type with m newid replaces.
   assumption.
   apply IHn.
   rewrite <- messages_replace_action_size.
   assumption.
  Qed.

Hint Resolve bindc_action_type.

  Lemma bindc_action_auto_type:
   forall (a:action) (r:id) (renv:IdSet.t) (new_id:unit -> id),
    type_action_recv r renv (bindc_action_auto new_id a).
  Proof.
   unfold bindc_action_auto.
   intros a r renv newid.
   apply bindc_action_type.
   apply Arith.Le.le_refl.
  Qed.

Hint Resolve bindc_action_auto_type.

  Definition bindc_spec new_id s :=
  match s with
  | Spec rss a =>
    let a' := bindc_action_auto new_id a in
     Spec rss a'
  end.

  Definition type_spec_recv r s :=
  match s with
  | Spec rss a =>
    (type_rspecs r rss)
    /\ (match rspec_find r rss with
        | None => False
        | Some (RSpec r bs es) =>
          type_action_recv r (varlist_set bs) a
        end)
  end.

  Lemma bindc_spec_type:
   forall (r:id) (rss:list role_spec) (a:action) (new_id:unit -> id),
    type_rspecs r rss ->
    type_spec_recv r (bindc_spec new_id (Spec rss a)).
  Proof.
   intros r rss a newid Tr.
   simpl.
   split. assumption.
   case_eq (rspec_find r rss).
   intros rs. destruct rs as [ r' bs es ].
   intro EQr.
   apply bindc_action_auto_type.
   intro EQr.
   unfold type_rspecs in Tr.
   destruct Tr as [ bs [ es EQr' ] ].
   absurd (Some (RSpec r bs es) = None).
   discriminate.
   apply trans_eq with (rspec_find r rss).
   symmetry. assumption. assumption.
  Qed.

Hint Resolve bindc_spec_type.

  Fixpoint type_spec_recv_roles rs s :=
  match rs with
  | nil => True
  | r :: rs =>
    (type_spec_recv r s) /\ (type_spec_recv_roles rs s)
  end.

  Lemma bindc_spec_type_roles:
   forall (rs:list id) (rss:list role_spec) (a:action) (new_id:unit -> id),
    (forall r:id, List.In r rs -> type_rspecs r rss) ->
    type_spec_recv_roles rs (bindc_spec new_id (Spec rss a)).
  Proof.
   induction rs as [| r rs ].
   simpl. trivial.
   intros rss a newid IN.
   red.
   split.
   apply bindc_spec_type.
   apply IN. left. reflexivity.
   apply IHrs. intro r'.
   intro INr'.
   apply IN. simpl. right. assumption.
  Qed.

Hint Resolve bindc_spec_type_roles.

  (* XXX spec_roles (bindc_spec new_id s) *)
  Lemma bindc_spec_type_all:
   forall (s:spec) (new_id:unit -> id),
    type_spec_recv_roles (spec_roles s) (bindc_spec new_id s).
  Proof.
   intros s newid.
   destruct s as [ rss a ].
   apply bindc_spec_type_roles.
   apply spec_roles_type.
  Qed.

Hint Resolve bindc_spec_type_all.

(* Open *)
Require Export ListOrder.

Definition LO_f vs (bmsg:id * msg) :=
  let (msgi, msgv) := bmsg in
    if type_msg_recv_dec vs msgv then
      mem msgi vs
    else
      false.

Lemma LO_f_equal:
  forall (s s':t) a,
    Equal s s' -> LO_f s a = LO_f s' a.
Proof.
  intros s s' [msgi msgv] EQs.
  unfold LO_f.
  case (type_msg_recv_dec s msgv); intros TMs;
  case (type_msg_recv_dec s' msgv); intros TMs';
  case_eq (mem msgi s); intros Ms;
  case_eq (mem msgi s'); intros Ms';
    try (reflexivity);
    try (rewrite <- mem_iff in *);
    try (rewrite <- not_mem_iff in *);
    rewrite EQs in *;
    contradiction.
Qed.

Definition LO_C (bmsg:id * msg) :=
  message_vars_bound (snd bmsg).

Definition open_bmsgs (bmsgs:list (id*msg)) (vs:t) :=
  list_order_auto (id * msg) LO_f LO_C vs bmsgs.

Definition bmsg_vars_bound (bmsgs:list (id*msg)) : IdSet.t :=
  LC (id*msg) LO_C bmsgs.

Lemma open_bmsgs_correct_perm:
  forall bmsgs vs,
    let oc := open_bmsgs bmsgs vs in
      Permutation bmsgs ((fst oc) ++ (snd oc)).
Proof.
  intros bmsgs vs oc.
  unfold oc.
  unfold open_bmsgs.
  apply list_order_auto_perm.
Qed.

Lemma open_bmsgs_closed:
  forall bmsgs vs o c,
    open_bmsgs bmsgs vs = (o,c) ->
    let vs' := union vs (bmsg_vars_bound o) in
    forall msgi msgv,
      List.In (msgi,msgv) c ->
      ((~ In msgi vs') \/ (~ type_msg_recv vs' msgv)).
Proof.
  intros bmsgs vs o c EQ vs' msgi msgv IN.
  generalize (list_order_auto_closed 
    (id*msg) LO_f LO_f_equal LO_C
    vs bmsgs o c EQ (msgi,msgv) IN).
  unfold LO_f.
  unfold bmsg_vars_bound in *.
  fold vs'.
  case (type_msg_recv_dec vs' msgv); intros TMRv.
  rewrite <- not_mem_iff. intuition.
  intuition.
Qed.

Lemma LO_f_subset:
  forall (s s':t) (a:id*msg),
    LO_f s a = true -> Subset s s' -> LO_f s' a = true.
Proof.
  intros s s' [msgi msgv].
  unfold LO_f.
  case (type_msg_recv_dec s msgv); intros TMs;
  case (type_msg_recv_dec s' msgv); intros TMs';
  try (rewrite <- mem_iff);
  try (rewrite <- not_mem_iff);
  try (rewrite <- mem_iff);
  try (rewrite <- not_mem_iff);
  intros INs SUB.
  apply SUB. assumption.
  assert False. apply TMs'.
  apply type_msg_recv_subset with s; intuition.
  contradiction.
  discriminate INs.
  discriminate INs.
Qed.

Theorem open_bmsgs_opened:
  forall bmsgs vs o c,
    open_bmsgs bmsgs vs = (o,c) ->
    forall msgi msgv pre post,
      o = pre ++ (msgi,msgv) :: post ->
      (In msgi (union vs (bmsg_vars_bound pre)))
      /\ (type_msg_recv (union vs (bmsg_vars_bound pre)) msgv).
Proof.
  intros bmsgs vs o c EQ msgi msgv pre post EQo.
  generalize (list_order_auto_opened
    (id*msg) LO_f LO_f_subset LO_f_equal LO_C
    vs bmsgs o c EQ (msgi,msgv) pre post EQo).
  unfold LO_f. fold (bmsg_vars_bound pre).
  case (type_msg_recv_dec (union vs (bmsg_vars_bound pre)) msgv).
  rewrite mem_iff. intuition.
  intros T F. discriminate F.
Qed.
  
Definition build_match (r:id) (opened:list (id*msg)) (a:action) :=
  fold_right
  (fun (bmsg:id*msg) (a:action) =>
    let (msgi,msgv) := bmsg in
      (Match r (Var msgi Msg) msgv a))
  a opened.

Lemma build_match_type :
  forall vs bmsgs o c,
    open_bmsgs bmsgs vs = (o,c) ->
    let vs' := (union vs (bmsg_vars_bound o)) in
      forall ns r es a,
        type_action ns vs' r es a ->
        type_action ns vs r es (build_match r o a).
Proof.
  intros vs bmsgs o c EQ vs' ns r es a Ta.
  generalize (open_bmsgs_opened bmsgs vs o c EQ).
  intros Hm.
  clear EQ. unfold vs' in *. clear vs'.
  revert vs Ta Hm.

  induction o as [|[msgi msgv] o]; simpl;
    intros vs Ta Hm.
  rewrite empty_union_2 in Ta; intuition.

  case (ME.eq_dec r r); intros EQr; [ clear EQr | 
    absurd (r = r); intuition ].
  case (type_eq_dec Msg Msg); intros TEQm; [ clear TEQm |
    absurd (Msg = Msg); intuition ].

  unfold exist_var. simpl.
  destruct (Hm msgi msgv nil o) as [IN TMR].
  simpl. reflexivity. simpl in *.
  rewrite empty_union_2 in IN; intuition.
  rewrite empty_union_2 in TMR; intuition.
  apply IHo. unfold LO_C in Ta. simpl in Ta.
  rewrite union_assoc. assumption.

  clear IN TMR Ta IHo.
  intros msgi' msgv' pre post EQ'.
  destruct (Hm msgi' msgv' ((msgi,msgv)::pre) post) as [IN TMR].
  rewrite EQ'. simpl. reflexivity.
  simpl in *. unfold LO_C in *. simpl in *.
  rewrite union_assoc. intuition.
Qed.
Hint Resolve build_match_type.
        
  Fixpoint open_action r vs bmsgs a {struct a} :=
  match a with
  | Return => 
    Return
  | Comm f t ffs m tfs a =>
    if ME.eq_dec r f then
     let vs' := union vs (formulas_vars_bound ffs) in
     let (o,c) := open_bmsgs bmsgs vs' in
     let vs'' := union vs' (bmsg_vars_bound o) in
       build_match r o
        (Comm f t ffs m tfs (open_action r vs'' c a))
    else if ME.eq_dec r t then
     let vs' := union (union vs (message_vars_bound m))
                 (formulas_vars_bound tfs)
     in
     let (o,c) := open_bmsgs bmsgs vs' in
     let vs'' := union vs' (bmsg_vars_bound o) in
       (Comm f t ffs m tfs 
        (build_match r o (open_action r vs'' c a)))
    else
     Comm f t ffs m tfs (open_action r vs bmsgs a)
  | New ar v nt a =>
    if ME.eq_dec r ar then
     New ar v nt (open_action r (add (varid v) vs) bmsgs a)
    else
     New ar v nt (open_action r vs bmsgs a)
  | Bind ar v m a =>
    if ME.eq_dec r ar then
     let vs' := (add (varid v) vs) in
     let (o,c) := open_bmsgs bmsgs vs' in
     let vs'' := union vs' (bmsg_vars_bound o) in
       (Bind ar v m 
         (build_match r o (open_action r vs'' c a)))
    else
     let bmsgs' := ((varid v),m) :: bmsgs in
     Bind ar v m (open_action r vs bmsgs' a)
  | Match ar v m a =>
    if ME.eq_dec r ar then
     let vs' := (union vs (message_vars_bound m)) in
     let (o,c) := open_bmsgs bmsgs vs' in
     let vs'' := union vs' (bmsg_vars_bound o) in
       (Match ar v m
        (build_match r o (open_action r vs'' c a)))
    else
     Match ar v m (open_action r vs bmsgs a)
  | Derive ar f a =>
    if ME.eq_dec r ar then
     let vs' := (union vs (formulas_vars_bound f)) in
     let (o,c) := open_bmsgs bmsgs vs' in
     let vs'' := union vs' (bmsg_vars_bound o) in
       (Derive ar f
        (build_match r o (open_action r vs'' c a)))
    else
     Derive ar f (open_action r vs bmsgs a)
  end.
	
  (* XXX Property idea: Unopened bmsgs and environment *)

  Definition open_spec r s :=
  match s with
  | Spec rss a =>
    match rspec_find r rss with
    | None => Spec rss a
    | Some (RSpec r bs es) =>
      Spec rss (open_action r (varlist_set bs) nil a)
    end
  end.

  Fixpoint open_spec_roles rs s :=
   match rs with
   | nil => s
   | r :: rs =>
     open_spec_roles rs (open_spec r s)
   end.

  Definition open_spec_all s :=
   open_spec_roles (spec_roles s) s.