(* WPPLi *)
Require Export WPPL.
Open Scope list_scope.

(* Gen *)
Definition genable_var v :=
  match v with
    | Var vi vt =>
      match type_to_new_type vt with
        | Some nt => true
        | None => false
      end
  end.

Definition gen_var r v a :=
  match v with
    | Var vi vt =>
      match type_to_new_type vt with
        | Some nt =>
          New r v nt a
        | None =>
          a
      end
  end.

Lemma gen_var_type:
  forall v ns vs r es a,
    genable_var v = true ->
    type_action ns (add (varid v) vs) r es a ->
    type_action ns vs r es (gen_var r v a).
Proof.
  intros [vi vt] ns vs r es a.
  simpl. destruct vt; simpl;
  try (intro F; discriminate F);
  intros T; clear T;
  (case (ME.eq_dec r r); intros EQ; [ clear EQ | absurd (r = r); intuition ]).
  case (type_eq_dec Nonce Nonce); intros EQ;
    [ intuition | absurd (Nonce = Nonce); intuition ].
  case (type_eq_dec Symkey Symkey); intros EQ;
    [ intuition | absurd (Symkey = Symkey); intuition ].
  case (type_eq_dec Pubkey Pubkey); intros EQ;
    [ intuition | absurd (Pubkey = Pubkey); intuition ].
Qed. 

Hint Resolve gen_var_type.

Lemma gen_var_type_pres:
  forall v ns vs r es a,
    type_action ns vs r es a ->
    type_action ns vs r es (gen_var r v a).
Proof.
  intros [vi vt] ns vs r es a Ta.
  simpl. destruct vt; simpl;
  intuition;
  (case (ME.eq_dec r r); intros EQ; [ clear EQ | absurd (r = r); intuition ]).
  case (type_eq_dec Nonce Nonce); intros EQ;
    [ intuition | absurd (Nonce = Nonce); intuition ].
  apply type_action_subset with vs; intuition.
  case (type_eq_dec Symkey Symkey); intros EQ;
    [ intuition | absurd (Symkey = Symkey); intuition ].
  apply type_action_subset with vs; intuition.
  case (type_eq_dec Pubkey Pubkey); intros EQ;
    [ intuition | absurd (Pubkey = Pubkey); intuition ].
  apply type_action_subset with vs; intuition.
Qed.

Hint Resolve gen_var_type_pres.

Lemma gen_var_type_pres':
  forall v ns vs r r' es a,
    type_action ns vs r es a ->
    type_action ns vs r es (gen_var r' v a).
Proof.
  intros [vi vt] ns vs r r' es a Ta.
  case (ME.eq_dec r r'); intros EQ.
  subst r'. intuition.
  simpl. destruct (type_to_new_type vt) as [nt|].
  simpl. case (ME.eq_dec r r'); intuition.
  intuition.
Qed.

Hint Resolve gen_var_type_pres'.

Lemma gen_var_connected:
  forall v r a,
    connected_action (Some r) a ->
    connected_action (Some r) (gen_var r v a).
Proof.
  intros [vi vt] r Ca. simpl.
  destruct (type_to_new_type vt) as [nt|];
  simpl; intuition.
Qed.

Hint Resolve gen_var_connected.

Lemma connected_some_none:
  forall r a,
    connected_action (Some r) a ->
    connected_action None a.
Proof.
  induction a; simpl; intuition.
Qed.

Hint Resolve connected_some_none.

Lemma gen_var_connected_none:
  forall r v a,
    connected_action (Some r) a ->
    connected_action None (gen_var r v a).
Proof.
  intros r [vi vt] a. simpl.
  destruct (type_to_new_type vt) as [nt|];
  simpl; intuition. 
  apply (connected_some_none r). intuition.
Qed.

Hint Resolve gen_var_connected_none.

Definition genable_vars := List.forallb genable_var.

Hint Unfold genable_vars.

Definition gen_vars r vs a := List.fold_right (gen_var r) a vs.

Hint Unfold gen_vars.

Lemma gen_vars_type:
  forall vsL ns vs r es a,
    genable_vars vsL = true ->
    type_action ns (union (varlist_set vsL) vs) r es a ->
    type_action ns vs r es (gen_vars r vsL a).
Proof.
  induction vsL as [|v vsL]; simpl;
    intros ns vs r es a EQt Ta.
  cut (Equal vs (union empty vs)). intros EQ.
  rewrite EQ. assumption. intuition.
  generalize (andb_prop (genable_var v) (genable_vars vsL) EQt).
  intros [EQtv EQtvL].
  apply gen_var_type. assumption.
  apply IHvsL; intuition.
  cut (Equal (union (varlist_set vsL) (add (varid v) vs))
    (union (add (varid v) (varlist_set vsL)) vs)). intros EQ.
  rewrite EQ. assumption.
  intros e. set_iff. intuition.
Qed.

Hint Resolve gen_vars_type.

Lemma gen_vars_type_pres:
  forall vsL ns vs r es a,
    type_action ns vs r es a ->
    type_action ns vs r es (gen_vars r vsL a).
Proof.
  induction vsL as [|v vsL]; simpl; intuition.
Qed.

Hint Resolve gen_vars_type_pres.

Lemma gen_vars_type_pres':
  forall vsL ns vs r r' es a,
    type_action ns vs r es a ->
    type_action ns vs r es (gen_vars r' vsL a).
Proof.
  induction vsL; simpl; intuition.
Qed.

Hint Resolve gen_vars_type_pres'.

Lemma gen_vars_connected:
  forall vsL r a,
    connected_action (Some r) a ->
    connected_action (Some r) (gen_vars r vsL a).
Proof.
  induction vsL as [|v vsL]; simpl; intuition.
Qed.

Hint Resolve gen_vars_connected.

Lemma gen_vars_connected_none:
  forall vL r a,
    connected_action (Some r) a ->
    connected_action None (gen_vars r vL a).
Proof.
  induction vL as [|v vL]; simpl; intros r a Ca.
  apply (connected_some_none r); intuition.
  intuition.
Qed.

Hint Resolve gen_vars_connected_none.

Fixpoint wait_until r F a :=
  match a with
    | Return =>
      a
    | Comm fr tr msg sa =>
      if ME.eq_dec r fr then
        F a
      else
        Comm fr tr msg (wait_until r F sa)
    | New ar v t sa =>
      if ME.eq_dec r ar then
        F a
      else
        New ar v t (wait_until r F sa)
    | Bind ar v msg sa =>
      if ME.eq_dec r ar then
        F a
      else
        Bind ar v msg (wait_until r F sa)
    | Match ar v msg sa =>
      if ME.eq_dec r ar then
        F a
      else        
        Match ar v msg (wait_until r F sa)
    | Derive ar f sa =>
      if ME.eq_dec r ar then
        F a
      else
        Derive ar f (wait_until r F sa)
  end.

Lemma wait_until_type_pres:
  forall F r,
    (forall ns vs es a, type_action ns vs r es a -> type_action ns vs r es (F a)) ->
    forall a ns vs es,
    type_action ns vs r es a ->
    type_action ns vs r es (wait_until r F a).
Proof.
  intros F r F_pres.

  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros ns vs es;
      [ idtac
        | case (ME.eq_dec r fr); [intro EQ; subst fr | intro NEQ;
          case (ME.eq_dec r tr); [intros EQ'; subst tr | intros NEQ']]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]];
      simpl.

  trivial.
  intros [IN [TMS Ta]]. apply F_pres. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].
  case (ME.eq_dec r fr); intros EQ; [intuition | clear EQ].
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].
  case (ME.eq_dec r fr); intros EQ; [intuition | clear EQ].
  case (ME.eq_dec r tr); intros EQ; [intuition | clear EQ].
  intuition. 

  intros [Tnt Ta]. apply F_pres. simpl. intuition.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ]. intuition.
  intros [Tv [TMS Ta]]. apply F_pres. simpl. intuition.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ]. intuition.
  intuition. apply F_pres. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ]. intuition.
  intuition. apply F_pres. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ]. intuition.
Qed.

Hint Resolve wait_until_type_pres.

Lemma wait_until_type_pres':
  forall F r r',
    (forall ns vs es a, type_action ns vs r es a -> type_action ns vs r es (F a)) ->
    forall a ns vs es,
    type_action ns vs r es a ->
    type_action ns vs r es (wait_until r' F a).
Proof.
  intros F r' r F_pres.

  case (ME.eq_dec r' r); intros EQr.
  subst r'. intuition.

  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros ns vs es;
      [ idtac
        | case (ME.eq_dec r fr); [intro EQ; subst fr | intro NEQ;
          case (ME.eq_dec r tr); [intros EQ'; subst tr | intros NEQ']]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]];
      simpl.

  intuition.
  case (ME.eq_dec r' r); intros EQ; [ intuition | clear EQ ].
  case (ME.eq_dec r' tr); [intros EQ'; subst tr | intros NEQ'].
  intros [TMR Ta].
  apply F_pres. simpl. 
  case (ME.eq_dec r' r); intros EQ; [ intuition | clear EQ ].
  case (ME.eq_dec r' r'); intros EQ; [intuition | absurd (r' = r'); intuition].
  intros Ta. apply F_pres. simpl.
  case (ME.eq_dec r' r); intros EQ; [ intuition | clear EQ ].
  case (ME.eq_dec r' tr); intros EQ; [intuition | intuition].

  case (ME.eq_dec r' r); intros EQ; [ intuition | clear EQ ].
  case (ME.eq_dec r' fr); intros EQ; [intuition | intuition].

  case (ME.eq_dec r' fr); intros EQ; [ intuition | clear EQ ].
  case (ME.eq_dec r' tr); intros EQ; [intuition | intuition].

  case (ME.eq_dec r' r); intros EQ; [ intuition | clear EQ ].
  intros Ta. apply F_pres. simpl.
  case (ME.eq_dec r' r); intros EQ; [ intuition | intuition ].

  case (ME.eq_dec r' ar); intros EQ; [intuition | intuition].

  case (ME.eq_dec r' r); intros EQ; [intuition | clear EQ].
  intros Ta. apply F_pres. simpl.
  case (ME.eq_dec r' r); intros EQ; [ intuition | intuition ].

  case (ME.eq_dec r' ar); intros EQ; [intuition | intuition].

  case (ME.eq_dec r' r); intros EQ; [intuition | clear EQ].
  intros Ta. apply F_pres. simpl.
  case (ME.eq_dec r' r); intros EQ; [ intuition | intuition ].

  case (ME.eq_dec r' ar); intros EQ; [intuition | intuition].

  case (ME.eq_dec r' r); intros EQ; [intuition | clear EQ].
  intros Ta. apply F_pres. simpl.
  case (ME.eq_dec r' r); intros EQ; [ intuition | intuition ].

  case (ME.eq_dec r' ar); intros EQ; [intuition | intuition].
Qed.

Hint Resolve wait_until_type_pres'.

Lemma gen_vars_wait_connected:
  forall r vsL a mr,
    connected_action mr a ->
    connected_action mr (wait_until r (gen_vars r vsL) a).
Proof.
  intros r.

  induction vsL as [|v vsL]; simpl.

  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros mr.
  intuition.
  case (ME.eq_dec r fr); intros EQ; [ subst fr | idtac ].
  simpl. intuition. simpl. intuition.
  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  simpl. intuition. simpl. intuition.
  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  simpl. intuition. simpl. intuition.
  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  simpl. intuition. simpl. intuition.
  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  simpl. intuition. simpl. intuition.

  rename v into gv.
  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros mr.
  intuition.

  case (ME.eq_dec r fr); intros EQ; [ subst fr | idtac ].
  destruct mr as [r'|]. simpl. intros [EQ Ca]. subst r'.
  apply gen_var_connected. apply gen_vars_connected.
  simpl. intuition.
  intros [CR Ca]. apply (gen_var_connected_none r).
  apply gen_vars_connected. simpl. intuition.
  simpl. intuition.

  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  destruct mr as [r'|]. simpl. intros [EQ Ca]. subst r'.
  apply gen_var_connected. apply gen_vars_connected.
  simpl. intuition.
  intros [CR Ca]. apply (gen_var_connected_none r).
  apply gen_vars_connected. simpl. intuition.
  simpl. intuition.

  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  destruct mr as [r'|]. simpl. intros [EQ Ca]. subst r'.
  apply gen_var_connected. apply gen_vars_connected.
  simpl. intuition.
  intros [CR Ca]. apply (gen_var_connected_none r).
  apply gen_vars_connected. simpl. intuition.
  simpl. intuition.

  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  destruct mr as [r'|]. simpl. intros [EQ Ca]. subst r'.
  apply gen_var_connected. apply gen_vars_connected.
  simpl. intuition.
  intros [CR Ca]. apply (gen_var_connected_none r).
  apply gen_vars_connected. simpl. intuition.
  simpl. intuition.

  case (ME.eq_dec r ar); intros EQ; [ subst ar | idtac ].
  destruct mr as [r'|]. simpl. intros [EQ Ca]. subst r'.
  apply gen_var_connected. apply gen_vars_connected.
  simpl. intuition.
  intros [CR Ca]. apply (gen_var_connected_none r).
  apply gen_vars_connected. simpl. intuition.
  simpl. intuition.
Qed.
Hint Resolve gen_vars_wait_connected.

(* XXX Up until here is good, the problem is that we need to ensure
   that we don't instantiate the variables of other parties or that
   are initially known by someone *)

Fixpoint action_varsL r a {struct a} :=
  match a with
    | Return =>
      nil
    | Comm fr tr m sa =>
      if ME.eq_dec r fr then
        (message_varsL m) ++ (action_varsL r sa)
      else if ME.eq_dec r tr then
        action_varsL r sa
        (* (message_vars_matchL m) ++ (action_varsL r sa) *)
      else
        action_varsL r sa
    | New ar v nt sa =>
      if ME.eq_dec r ar then
        v :: (action_varsL r sa)
      else
        action_varsL r sa
    | Bind ar v m sa =>
      if ME.eq_dec r ar then
        v :: (message_varsL m) ++ (action_varsL r sa)
      else
        action_varsL r sa
    | Match ar v m sa =>
      if ME.eq_dec r ar then
        v :: (message_varsL m) ++ (action_varsL r sa)
      else
        action_varsL r sa
    | Derive ar f sa =>
      if ME.eq_dec r ar then
        (formulas_vars_proveL f) ++ (action_varsL r sa)
      else
        action_varsL r sa
  end.

Fixpoint action_missing_vars r a :=
  match a with
    | Return =>
      empty
    | Comm fr tr m sa =>
      if ME.eq_dec r fr then        
        (union (message_vars m)
          (action_missing_vars r sa))
      else if ME.eq_dec r tr then
        action_missing_vars r sa
        (*(union (message_vars_match m)
          (diff (action_missing_vars r sa)
            (message_vars_bound m)))*)
      else
        action_missing_vars r sa
    | New ar v nt sa =>
      (action_missing_vars r sa)
    | Bind ar v m sa =>
      if ME.eq_dec r ar then
        (union (message_vars m)
          (action_missing_vars r sa))
      else
        action_missing_vars r sa
    | Match ar v m sa =>
      if ME.eq_dec r ar then
        (union (message_vars_match m)
          (diff (action_missing_vars r sa)
            (message_vars_bound m)))
      else
        action_missing_vars r sa
    | Derive ar f sa =>
      if ME.eq_dec r ar then
        (union (formulas_vars_prove f)
          (diff (action_missing_vars r sa)
            (formulas_vars_bound f)))
      else
        action_missing_vars r sa
  end.

Definition action_missing_varsL r a :=
  List.filter 
   (fun v =>
     mem (varid v) (action_missing_vars r a))
   (action_varsL r a).

Lemma action_missing_varsL_correct:
  forall v r a,
    List.In v (action_missing_varsL r a) ->
    In (varid v) (action_missing_vars r a).
Proof.
  intros v r a.
  unfold action_missing_varsL.
  rewrite filter_In. rewrite <- mem_iff.
  intuition.
Qed.

Definition gen_only := List.filter genable_var.

Hint Unfold gen_only.

Lemma gen_only_able:
  forall vsL,
    genable_vars (gen_only vsL) = true.
Proof.
  induction vsL; simpl; intuition.
  case_eq (genable_var a); simpl.
  intros EQ. rewrite EQ. simpl. assumption.
  intuition.
Qed.

Hint Resolve gen_only_able.

Definition gen_action r a :=
  (wait_until r (gen_vars r (gen_only (action_missing_varsL r a))) a).

Theorem filter_app:
  forall A (f:A->bool) l l',
    List.filter f (l ++ l') = (List.filter f l) ++ (List.filter f l').
Proof.
  intros A f.
  induction l as [|e l]; simpl.
  intuition. intros l'. rewrite IHl.
  case (f e); simpl; reflexivity.
Qed.

Theorem filter_filter:
  forall A (f g:A->bool) l,
    List.filter f (List.filter g l) =
    List.filter (fun e => andb (f e) (g e)) l.
Proof.
  intros A f g.
  induction l as [|e l]; simpl.
  reflexivity.
  case (g e); simpl; case (f e); simpl.
  rewrite IHl. reflexivity.
  assumption. assumption. assumption.
Qed.

Fixpoint type_after ns vs vs' r es a :=
  let P := type_action ns vs r es in
  let F := type_action ns (union vs' vs) r es in
  match a with
    | Return =>
      P a
    | Comm fr tr m sa =>
      if ME.eq_dec r fr then
        F a
      else if ME.eq_dec r tr then
        type_msg_recv vs m 
        /\ type_after (add fr ns) (union vs (message_vars_bound m)) vs' r es sa
      else
        type_after ns vs vs' r es sa
    | New ar v t sa =>
      if ME.eq_dec r ar then
        F a
      else
        type_after ns vs vs' r es sa
    | Bind ar v msg sa =>
      if ME.eq_dec r ar then
        F a
      else
        type_after ns vs vs' r es sa
    | Match ar v msg sa =>
      if ME.eq_dec r ar then
        F a
      else
        type_after ns vs vs' r es sa
    | Derive ar f sa =>
      if ME.eq_dec r ar then
        F a
      else
        type_after ns vs vs' r es sa
  end.

Lemma gen_action_type:
  forall r a ns vs es,
    type_after ns vs (varlist_set (gen_only (action_missing_varsL r a))) r es a -> 
    type_action ns vs r es (gen_action r a).
Proof.
  unfold gen_action. intros r.

  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros ns vs es;
      [ idtac
        | case (ME.eq_dec r fr); [intro EQ; subst fr | intro NEQ;
          case (ME.eq_dec r tr); [intros EQ'; subst tr | intros NEQ']]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]];
      simpl; intros Ta.

  intuition.

  apply gen_vars_type. apply gen_only_able. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].

  revert Ta. unfold action_missing_varsL. simpl.
  case (ME.eq_dec r fr); intros EQ; [intuition | clear EQ].
  case (ME.eq_dec r r); intros EQ; [clear EQ | absurd (r = r); intuition].
  intuition.

  revert Ta. unfold action_missing_varsL. simpl.
  case (ME.eq_dec r fr); intros EQ; [intuition | clear EQ].
  case (ME.eq_dec r tr); intros EQ; [intuition | clear EQ].
  apply IHa.

  apply gen_vars_type. apply gen_only_able. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].

  revert Ta. unfold action_missing_varsL. simpl.
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ].
  apply IHa.

  apply gen_vars_type. apply gen_only_able. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].

  revert Ta. unfold action_missing_varsL. simpl.
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ].
  apply IHa.

  apply gen_vars_type. apply gen_only_able. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].

  revert Ta. unfold action_missing_varsL. simpl.
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ].
  apply IHa.

  apply gen_vars_type. apply gen_only_able. simpl.
  case (ME.eq_dec r r); intros EQ; [intuition | absurd (r = r); intuition].

  revert Ta. unfold action_missing_varsL. simpl.
  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ].
  apply IHa.
Qed.

Hint Resolve gen_action_type.

Lemma gen_action_type_pres:
  forall ns vs r es a,
    type_action ns vs r es a ->
    type_action ns vs r es (gen_action r a).
Proof.
  unfold gen_action. intuition.
Qed.

Hint Resolve gen_action_type_pres.

Lemma gen_action_type_pres':
  forall ns vs r r' es a,
    type_action ns vs r es a ->
    type_action ns vs r es (gen_action r' a).
Proof.
  unfold gen_action. intuition. 
Qed.

Hint Resolve gen_action_type_pres'.

Lemma gen_action_connected:
  forall mr r a,
    connected_action mr a ->
    connected_action mr (gen_action r a).
Proof.
  unfold gen_action. intuition.
Qed.

Hint Resolve gen_action_connected.

(* spec_lift *)
Definition spec_lift F r s :=
  match s with
    | Spec rss a =>
      match rspec_find r rss with
        | Some (RSpec r' bs es) =>
          Spec rss (F r' bs es a)
        | None =>
          Spec rss a
      end
  end.

Lemma spec_lift_roles_eq:
  forall F r s,
    spec_roles (spec_lift F r s) = spec_roles s.
Proof.
  intros F r [rss a]. simpl.
  destruct (rspec_find r rss) as [[r' bs es]|]; simpl; intuition.
Qed.

Hint Resolve spec_lift_roles_eq.

Lemma spec_lift_type_pres:
  forall F r s,
    (forall ns bs' vs es' es r r' a, 
      type_action ns vs r es a ->
      type_action ns vs r es (F r' bs' es' a)) ->
    type_spec r s ->
    type_spec r (spec_lift F r s).
Proof.
  intros F r s F_pres.
  destruct s as [rss a]. simpl.
  case_eq (rspec_find r rss); simpl.
  intros [r' bs es]. simpl. intros EQ.
  rewrite EQ. intuition.
  intros EQ. rewrite EQ. intuition.
Qed.

Hint Resolve spec_lift_type_pres.

Lemma spec_lift_type_pres':
  forall F r r' s,
    (forall ns bs' vs es' es r r' a, 
      type_action ns vs r es a ->
      type_action ns vs r es (F r' bs' es' a)) ->
    type_spec r s ->
    type_spec r (spec_lift F r' s).
Proof.
  intros F r r' s F_pres.
  destruct s as [rss a]. simpl.
  case_eq (rspec_find r rss); simpl.
  intros [r'' bs es]. simpl. intros EQ.
  case_eq (rspec_find r' rss); simpl.
  intros [r''' bs' es']. simpl. intros EQ'.
  rewrite EQ. intuition.
  rewrite EQ. intuition.
  intros EQ.
  case_eq (rspec_find r' rss); simpl.
  intros [r''' bs' es']. simpl. intros EQ'.
  rewrite EQ. intuition.
  rewrite EQ. intuition.
Qed.

Hint Resolve spec_lift_type_pres'.

Lemma spec_lift_connected:
  forall F r s,
    (forall mr r bs es a, connected_action mr a -> connected_action mr (F r bs es a)) ->
    connected_spec s ->
    connected_spec (spec_lift F r s).
Proof.
  intros F r s F_conn.
  destruct s as [rss a]. simpl.
  destruct (rspec_find r rss) as [[r' bs es]|]; simpl; intuition.
Qed.

Hint Resolve spec_lift_connected.

Definition spec_all_lift F s :=
  List.fold_right (fun r s => (F r s)) s (spec_roles s).

Lemma spec_all_lift_roles_eq:
  forall F s,
    (forall r s, spec_roles (F r s) = spec_roles s) -> 
    spec_roles (spec_all_lift F s) = spec_roles s.
Proof.
  intros F s F_eq.
  unfold spec_all_lift.

  destruct s as [rss a]. simpl. revert a.
  destruct rss as [|[r bs es] rss]; simpl.
  reflexivity. intros a.
  rewrite F_eq.

  cut (forall s' l, spec_roles (fold_right (fun r s => F r s)  s' l) = spec_roles s').
  intros EQ. rewrite EQ. simpl. reflexivity.
  clear r bs es rss a. intros s rs.
  induction rs as [|r rs]; simpl.
  reflexivity. 
  rewrite F_eq. rewrite IHrs. reflexivity.
Qed.

Hint Resolve spec_all_lift_roles_eq.

Lemma spec_all_lift_type_roles_pres:
  forall F rs s,
    (forall r r' s, type_spec r s -> type_spec r (F r' s)) ->
    type_spec_roles rs s ->
    type_spec_roles rs (spec_all_lift F s).
Proof.
  intros F.
  induction rs as [|r rs]; simpl.
  intuition.
  intros s F_pres [Tr Trs].
  split; [idtac | intuition].

  unfold spec_all_lift. clear IHrs.
  destruct s as [rss a]. simpl.
  clear Trs. revert Tr.
  cut (forall l s, type_spec r s -> type_spec r (fold_right (fun r s => F r s) s l)).
  intros H. intuition.

  clear rss rs a.
  induction l as [|r' rs]; simpl; intuition.
Qed.

Hint Resolve spec_all_lift_type_roles_pres.

Lemma spec_all_lift_type_pres:
  forall F s,
    (forall r s, spec_roles (F r s) = spec_roles s) -> 
    (forall r r' s, type_spec r s -> type_spec r (F r' s)) ->
    type_spec_all s ->
    type_spec_all (spec_all_lift F s).
Proof.
  intros F s F_eq F_pres.
  unfold type_spec_all.
  rewrite spec_all_lift_roles_eq.
  unfold spec_all_lift.
  destruct s as [rss a]. simpl.
  intros Trs.
  generalize (spec_all_lift_type_roles_pres
    F (role_list rss) (Spec rss a)
    F_pres Trs). unfold spec_all_lift.
  intuition. intuition.
Qed.

Hint Resolve spec_lift_type_pres.

Lemma spec_all_lift_connected:
  forall F s,
    (forall r s, connected_spec s -> connected_spec (F r s)) ->
    connected_spec s ->
    connected_spec (spec_all_lift F s).
Proof.
  intros F s F_conn.
  intros Cs. unfold spec_all_lift.
  cut (forall rs, connected_spec (fold_right (fun r s => F r s) s rs)).
  intros H. apply H. intros rs.
  induction rs as [|r rs]; simpl; intuition.
Qed.

Hint Resolve spec_all_lift_connected.

(* / spec_lift *)

Definition gen_spec :=
  spec_lift (fun r bs es a => gen_action r a).
Hint Unfold gen_spec.

Lemma gen_spec_roles_eq:
  forall r s,
    spec_roles (gen_spec r s) = spec_roles s.
Proof.
  unfold gen_spec. intuition.
Qed.

Hint Resolve gen_spec_roles_eq.

Lemma gen_spec_type_pres:
  forall r s,
    type_spec r s ->
    type_spec r (gen_spec r s).
Proof.
  unfold gen_spec. intuition.
Qed.

Hint Resolve gen_spec_type_pres.

Lemma gen_spec_type_pres':
  forall r r' s,
    type_spec r s ->
    type_spec r (gen_spec r' s).
Proof.
  unfold gen_spec. intuition.
Qed.
Hint Resolve gen_spec_type_pres'.

Lemma gen_spec_connected:
  forall r s,
    connected_spec s ->
    connected_spec (gen_spec r s).
Proof.
  unfold gen_spec. intuition.
Qed.

Hint Resolve gen_spec_connected.

Definition gen_spec_all :=
  spec_all_lift gen_spec.

Lemma gen_spec_all_roles_eq:
  forall s,
    spec_roles (gen_spec_all s) = spec_roles s.
Proof.
  unfold gen_spec_all. intuition.
Qed.

Hint Resolve gen_spec_all_roles_eq.

Lemma gen_spec_all_type_pres:
  forall s,
    type_spec_all s ->
    type_spec_all (gen_spec_all s).
Proof.
  unfold gen_spec_all.
  intros s Ts.
  apply spec_all_lift_type_pres; intuition.
Qed.

Hint Resolve gen_spec_all_type_pres.

Lemma gen_spec_all_connected:
  forall s,
    connected_spec s ->
    connected_spec (gen_spec_all s).
Proof.
  unfold gen_spec_all. intuition.
Qed.

Hint Resolve gen_spec_all_connected.

(* rerecv *)
Fixpoint rerecv_action a :=
  match a with
  | Return => 
    Return
  | Comm fr tr m a =>
    let ni := new_id tt in
    let nv := Var ni Msg in
      Bind fr nv m
       (Comm fr tr (MVar nv)
         (Match tr nv m
           (rerecv_action a)))
  | New ar v nt a =>
    New ar v nt (rerecv_action a)
  | Bind ar v m a =>
    Bind ar v m (rerecv_action a)
  | Match ar v m a =>
    Match ar v m (rerecv_action a)
  | Derive ar f a =>
    Derive ar f (rerecv_action a)
  end.

Fixpoint type_recv_action a :=
  match a with
  | Return => 
    True
  | Comm fr tr m a =>
    (type_msg_recv empty m)
    /\ (type_recv_action a)
  | New ar v nt a =>
    (type_recv_action a)
  | Bind ar v m a =>
    (type_recv_action a)
  | Match ar v m a =>
    (type_recv_action a)
  | Derive ar f a =>
    (type_recv_action a)
  end.

Lemma rerecv_type_recv_action:
 forall a, type_recv_action (rerecv_action a).
Proof.
 induction a; simpl; intuition.
Qed.
Hint Resolve rerecv_type_recv_action.

Lemma rerecv_action_type_pres:
  forall r a ns vs es,
    type_action ns vs r es a ->
    type_action ns vs r es (rerecv_action a).
Proof.
  intros r.

  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros ns vs es;
      [ idtac
        | case (ME.eq_dec r fr); [intro EQ; subst fr | intro NEQ;
          case (ME.eq_dec r tr); [intros EQ'; subst tr | intros NEQ']]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]];
      simpl; intros Ta;
        try (intuition; fail).

  case (type_eq_dec Msg Msg); 
    intros EQ; [ clear EQ | absurd (Msg = Msg); intuition ].
  unfold exist_var. simpl. set_iff.
  case (ME.eq_dec r tr); intros EQ; [ subst tr | idtac ].
  set_iff.
  split. trivial.
  split. intuition.
  split. intuition.
  split. intuition.
  split. trivial.
  split. intuition.
  split. apply type_msg_send_recv. 
  apply type_msg_send_subset with vs; intuition.
  apply IHa. apply type_action_subset with vs.
  intuition. intros e. set_iff. intuition.

  set_iff.
  split. trivial.
  split. intuition.
  split. intuition.
  split. intuition.
  apply IHa. apply type_action_subset with vs; intuition.

  case (type_eq_dec Msg Msg); 
    intros EQ; [ clear EQ | absurd (Msg = Msg); intuition ].
  unfold exist_var. simpl. set_iff. intuition.
  apply IHa. 
  apply type_action_subset with (union vs (message_vars_bound m)).
  assumption. intros e. set_iff. intuition.
Qed.
Hint Resolve rerecv_action_type_pres.

Lemma rerecv_action_connected:
  forall a mr,
    connected_action mr a ->
    connected_action mr (rerecv_action a).
Proof.
  induction a; simpl; intuition.
Qed.
Hint Resolve rerecv_action_connected.

Definition rerecv_spec s := 
  match s with
    | Spec rss a =>
      Spec rss (rerecv_action a)
  end.

Lemma rerecv_spec_type_pres:
  forall r s,
    type_spec r s ->
    type_spec r (rerecv_spec s).
Proof.
  intros r [rss a]. simpl.
  destruct (rspec_find r rss) as [[r' bs es]|]; simpl; intuition.
Qed.
Hint Resolve rerecv_spec_type_pres.

Lemma rerecv_spec_connected:
  forall s,
    connected_spec s ->
    connected_spec (rerecv_spec s).
Proof.
  intros [rss a]. simpl. intuition.
Qed.
Hint Resolve rerecv_spec_connected.

(* message_replace *)
Fixpoint message_replace i rm m { struct m } :=
  if msg_eq_dec rm m then
    MVar (Var i Msg)
    else
      match m with 
        | MNil => MNil
        | MVar v => MVar v
        | MConst c => MConst c
        | Join m1 m2 =>
          Join (message_replace i rm m1) (message_replace i rm m2)
        | Hash m => 
          Hash (message_replace i rm m)
        | PubEnc v m => 
          PubEnc v (message_replace i rm m)
        | PubSign v m =>
          PubSign v (message_replace i rm m)
        | SymEnc v m => 
          SymEnc v (message_replace i rm m)
        | SymSign v m =>
          SymSign v (message_replace i rm m)
        | Common.Bind v m =>
          Common.Bind v (message_replace i rm m)
      end.

Lemma message_replace_type_send:
  forall (i:id) (rm:msg) (m:msg) (vs:IdSet.t),
    type_msg_send vs m ->
    type_msg_send (add i vs) (message_replace i rm m).
Proof.
  intros i rm m vs.

  revert m.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl;
      [ case (msg_eq_dec rm MNil)
        | case (msg_eq_dec rm (MVar v))
        | case (msg_eq_dec rm (MConst c))
        | case (msg_eq_dec rm (Join m1 m2))
        | case (msg_eq_dec rm (Hash m1))
        | case (msg_eq_dec rm (SymEnc v m1))
        | case (msg_eq_dec rm (SymSign v m1))
        | case (msg_eq_dec rm (PubEnc v m1))
        | case (msg_eq_dec rm (PubSign v m1))
        | case (msg_eq_dec rm (Common.Bind v m1)) ];
      simpl;
        unfold exist_var;
          simpl; set_iff;
            intros EQ Tm;
              intuition.
Qed.
Hint Resolve message_replace_type_send.

Lemma message_replace_type_recv:
  forall (i:id) (rm:msg) (m:msg) (vs:IdSet.t),
    type_msg_recv vs m ->
    type_msg_recv (add i vs) (message_replace i rm m).
Proof.
  intros i rm m vs.

  revert m.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl;
      [ case (msg_eq_dec rm MNil)
        | case (msg_eq_dec rm (MVar v))
        | case (msg_eq_dec rm (MConst c))
        | case (msg_eq_dec rm (Join m1 m2))
        | case (msg_eq_dec rm (Hash m1))
        | case (msg_eq_dec rm (SymEnc v m1))
        | case (msg_eq_dec rm (SymSign v m1))
        | case (msg_eq_dec rm (PubEnc v m1))
        | case (msg_eq_dec rm (PubSign v m1))
        | case (msg_eq_dec rm (Common.Bind v m1)) ];
      simpl;
        unfold exist_var;
          simpl; set_iff;
            intros EQ Tm;
              intuition.
Qed.
Hint Resolve message_replace_type_recv.

Definition messages_replace ms m :=
  List.fold_right (fun e m => message_replace (fst e) (snd e) m) m ms.
Hint Unfold messages_replace.

Definition idset := List.fold_right add empty.
Hint Unfold idset.

Lemma in_idset_iff:
  forall is e,
    In e (idset is) <-> List.In e is.
Proof.
  induction is as [|i is]; simpl; intros e; set_iff.
  intuition. rewrite IHis. intuition.
Qed.
Hint Resolve in_idset_iff.

Lemma messages_replace_type_send:
  forall ms m vs,
    type_msg_send vs m ->
    type_msg_send (union (idset (List.map (fun e => fst e) ms)) vs)
     (messages_replace ms m).
Proof.
  unfold messages_replace.
  induction ms as [|[i rm] ms]; simpl; intros m vs Tm.
  apply type_msg_send_subset with vs; intuition.
  apply type_msg_send_subset with 
    (add i (union (idset (map (fun e => fst e) ms)) vs)).
  intuition.
  intros a. set_iff. intuition.
Qed.
Hint Resolve messages_replace_type_send.

Lemma messages_replace_type_recv:
  forall ms m vs,
    type_msg_recv vs m ->
    type_msg_recv (union (idset (List.map (fun e => fst e) ms)) vs)
     (messages_replace ms m).
Proof.
  unfold messages_replace.
  induction ms as [|[i rm] ms]; simpl; intros m vs Tm.
  apply type_msg_recv_subset with vs; intuition.
  apply type_msg_recv_subset with 
    (add i (union (idset (map (fun e => fst e) ms)) vs)).
  intuition.
  intros a. set_iff. intuition.
Qed.
Hint Resolve messages_replace_type_recv.

(* bindc *)
Fixpoint bindc_msg m 
  : (prod (list (prod id msg)) msg) :=	
  match m with 
  | MNil =>	
    (nil, MNil)
  | MVar v =>
    (nil, MVar v)
  | MConst c =>
    (nil, MConst c)
  | Join m1 m2 =>
    let (mrs1, m1') := bindc_msg m1 in
    let (mrs2, m2') := bindc_msg m2 in
     (mrs1 ++ mrs2, Join m1' m2')
  | Hash m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg m in
     ((ni, Hash m') :: mrs, MVar (Var ni Msg))
  | SymEnc v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg m in
     ((ni, SymEnc v m') :: mrs, MVar (Var ni Msg))
  | SymSign v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg m in
     ((ni, SymSign v m') :: mrs, MVar (Var ni Msg))
  | PubEnc v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg m in
     ((ni, PubEnc v m') :: mrs, MVar (Var ni Msg))
  | PubSign v m =>
    let ni := new_id tt in
    let (mrs, m') := bindc_msg m in
     ((ni, PubSign v m') :: mrs, MVar (Var ni Msg))
  | Common.Bind v m =>
    let (mrs, m') := bindc_msg m in
     (mrs, Common.Bind (Var (varid v) Msg) m')
  end.

Lemma bindc_msg_type_recv:
  forall m vs,
    type_msg_recv vs (snd (bindc_msg m)).
Proof.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    intros vs;
    simpl; intuition;
      [ destruct (bindc_msg m1) as [ms1 m1'];
        destruct (bindc_msg m2) as [ms2 m2'];
          simpl in *; intuition
      | idtac
      | idtac
      | idtac
      | idtac
      | idtac
      | idtac ];
  destruct (bindc_msg m1) as [ms1 m1'];
  simpl in *; intuition.
  case (type_eq_dec Msg Msg); 
    intros EQ; [ trivial | absurd (Msg = Msg); intuition ].
Qed.
Hint Resolve bindc_msg_type_recv.

(* open *)
Require Export ListOrder.

Definition LO_f vs (bmsg:id * msg) :=
  let (msgi, msgv) := bmsg in
    if type_msg_recv_dec vs msgv then
      mem msgi vs
    else
      false.

Lemma LO_f_equal:
  forall (s s':t) a,
    Equal s s' -> LO_f s a = LO_f s' a.
Proof.
  intros s s' [msgi msgv] EQs.
  unfold LO_f.
  case (type_msg_recv_dec s msgv); intros TMs;
  case (type_msg_recv_dec s' msgv); intros TMs';
  case_eq (mem msgi s); intros Ms;
  case_eq (mem msgi s'); intros Ms';
    try (reflexivity);
    try (rewrite <- mem_iff in *);
    try (rewrite <- not_mem_iff in *);
    rewrite EQs in *;
    contradiction.
Qed.

Definition LO_C (bmsg:id * msg) :=
  message_vars_bound (snd bmsg).

Definition open_bmsgs (bmsgs:list (id*msg)) (vs:t) :=
  list_order_auto (id * msg) LO_f LO_C vs bmsgs.

Definition bmsg_vars_bound (bmsgs:list (id*msg)) : IdSet.t :=
  LC (id*msg) LO_C bmsgs.

Lemma open_bmsgs_correct_perm:
  forall bmsgs vs,
    let oc := open_bmsgs bmsgs vs in
      Permutation bmsgs ((fst oc) ++ (snd oc)).
Proof.
  intros bmsgs vs oc.
  unfold oc.
  unfold open_bmsgs.
  apply list_order_auto_perm.
Qed.

Lemma open_bmsgs_closed:
  forall bmsgs vs o c,
    open_bmsgs bmsgs vs = (o,c) ->
    let vs' := union vs (bmsg_vars_bound o) in
    forall msgi msgv,
      List.In (msgi,msgv) c ->
      ((~ In msgi vs') \/ (~ type_msg_recv vs' msgv)).
Proof.
  intros bmsgs vs o c EQ vs' msgi msgv IN.
  generalize (list_order_auto_closed 
    (id*msg) LO_f LO_f_equal LO_C
    vs bmsgs o c EQ (msgi,msgv) IN).
  unfold LO_f.
  unfold bmsg_vars_bound in *.
  fold vs'.
  case (type_msg_recv_dec vs' msgv); intros TMRv.
  rewrite <- not_mem_iff. intuition.
  intuition.
Qed.

Lemma LO_f_subset:
  forall (s s':t) (a:id*msg),
    LO_f s a = true -> Subset s s' -> LO_f s' a = true.
Proof.
  intros s s' [msgi msgv].
  unfold LO_f.
  case (type_msg_recv_dec s msgv); intros TMs;
  case (type_msg_recv_dec s' msgv); intros TMs';
  try (rewrite <- mem_iff);
  try (rewrite <- not_mem_iff);
  try (rewrite <- mem_iff);
  try (rewrite <- not_mem_iff);
  intros INs SUB.
  apply SUB. assumption.
  assert False. apply TMs'.
  apply type_msg_recv_subset with s; intuition.
  contradiction.
  discriminate INs.
  discriminate INs.
Qed.

Theorem open_bmsgs_opened:
  forall bmsgs vs o c,
    open_bmsgs bmsgs vs = (o,c) ->
    forall msgi msgv pre post,
      o = pre ++ (msgi,msgv) :: post ->
      (In msgi (union vs (bmsg_vars_bound pre)))
      /\ (type_msg_recv (union vs (bmsg_vars_bound pre)) msgv).
Proof.
  intros bmsgs vs o c EQ msgi msgv pre post EQo.
  generalize (list_order_auto_opened
    (id*msg) LO_f LO_f_subset LO_f_equal LO_C
    vs bmsgs o c EQ (msgi,msgv) pre post EQo).
  unfold LO_f. fold (bmsg_vars_bound pre).
  case (type_msg_recv_dec (union vs (bmsg_vars_bound pre)) msgv).
  rewrite mem_iff. intuition.
  intros T F. discriminate F.
Qed.
  
Definition build_match (r:id) (opened:list (id*msg)) (a:action) :=
  fold_right
  (fun (bmsg:id*msg) (a:action) =>
    let (msgi,msgv) := bmsg in
      (Match r (Var msgi Msg) msgv a))
  a opened.

Lemma build_match_type :
  forall vs bmsgs o c,
    open_bmsgs bmsgs vs = (o,c) ->
    let vs' := (union vs (bmsg_vars_bound o)) in
      forall ns r es a,
        type_action ns vs' r es a ->
        type_action ns vs r es (build_match r o a).
Proof.
  intros vs bmsgs o c EQ vs' ns r es a Ta.
  generalize (open_bmsgs_opened bmsgs vs o c EQ).
  intros Hm.
  clear EQ. unfold vs' in *. clear vs'.
  revert vs Ta Hm.

  induction o as [|[msgi msgv] o]; simpl;
    intros vs Ta Hm.
  rewrite empty_union_2 in Ta; intuition.

  case (ME.eq_dec r r); intros EQr; [ clear EQr | 
    absurd (r = r); intuition ].
  case (type_eq_dec Msg Msg); intros TEQm; [ clear TEQm |
    absurd (Msg = Msg); intuition ].

  unfold exist_var. simpl.
  destruct (Hm msgi msgv nil o) as [IN TMR].
  simpl. reflexivity. simpl in *.
  rewrite empty_union_2 in IN; intuition.
  rewrite empty_union_2 in TMR; intuition.
  apply IHo. unfold LO_C in Ta. simpl in Ta.
  rewrite union_assoc. assumption.

  clear IN TMR Ta IHo.
  intros msgi' msgv' pre post EQ'.
  destruct (Hm msgi' msgv' ((msgi,msgv)::pre) post) as [IN TMR].
  rewrite EQ'. simpl. reflexivity.
  simpl in *. unfold LO_C in *. simpl in *.
  rewrite union_assoc. intuition.
Qed.
Hint Resolve build_match_type.

Lemma build_match_connected:
  forall r a bmsgs,
    connected_action (Some r) a ->
    connected_action (Some r) (build_match r bmsgs a).
Proof.
  induction bmsgs as [|[vi vm] bs]; simpl; intuition.
Qed.
Hint Resolve build_match_connected.

(* bindco *)
Fixpoint bindco_action r vs ms a :=
  match a with
  | Return => 
    Return
  | Comm fr tr m a =>
    if ME.eq_dec r fr then
      let m' := messages_replace ms m in
        Comm fr tr m
         (bindco_action r vs ms a)
    else if ME.eq_dec r tr then
      let vs' := union vs (message_vars_bound m) in
      let (o,c) := open_bmsgs ms vs' in
      let vs'' := union vs' (bmsg_vars_bound o) in
        Comm fr tr m
         (build_match tr o
           (bindco_action tr vs'' c a))
    else
      Comm fr tr m
       (bindco_action r vs ms a)
  | New ar v nt a =>
    if ME.eq_dec r ar then
      let vs' := add (varid v) vs in
        New ar v nt
         (bindco_action r vs' ms a)
    else
      New ar v nt
       (bindco_action r vs ms a)
  | Bind ar v m a =>
    if ME.eq_dec r ar then
      let m' := messages_replace ms m in
        Bind ar v m'
         (bindco_action r (add (varid v) vs) ms a)
    else
      Bind ar v m
       (bindco_action r vs ms a)
  | Match ar v m a =>
    if ME.eq_dec r ar then
      let (ms',m') := bindc_msg m in
      let vs' := union vs (message_vars_bound m') in
      let (o,c) := open_bmsgs (ms++ms') vs' in
      let vs'' := union vs' (bmsg_vars_bound o) in
        Match ar v m'
         (build_match ar o
           (bindco_action ar vs'' c a))
    else
      Match ar v m
       (bindco_action r vs ms a)
  | Derive ar f a =>
    if ME.eq_dec r ar then
      let vs' := union vs (formulas_vars_bound f) in
      let (o,c) := open_bmsgs ms vs' in
      let vs'' := union vs' (bmsg_vars_bound o) in
        Derive ar f
         (build_match ar o
           (bindco_action ar vs'' c a))
    else
      Derive ar f
       (bindco_action r vs ms a)
  end.

Lemma bindco_action_connected:
  forall r a mr vs ms,
    connected_action mr a ->
    connected_action mr (bindco_action r vs ms a).
Proof.
  intros r.
  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros mr vs ms;
      [ idtac
        | case (ME.eq_dec r fr); [intro EQ; subst fr | intro NEQ;
          case (ME.eq_dec r tr); [intros EQ'; subst tr | intros NEQ']]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]];
      simpl; intros Ca; intuition.

  destruct (open_bmsgs ms (union vs (message_vars_bound m))).
  simpl. intuition.

  destruct (bindc_msg m) as [ms' m'].
  destruct (open_bmsgs (ms++ms') (union vs (message_vars_bound m'))).
  simpl. intuition.

  destruct (open_bmsgs ms (union vs (formulas_vars_bound phi))).
  simpl. intuition.
Qed.
Hint Resolve bindco_action_connected.

Lemma bindc_vars_bound:
  forall m ms' m',
    bindc_msg m = (ms', m') ->
    Subset (message_vars_bound m) 
     (union (message_vars_bound m') (bmsg_vars_bound ms')).
Proof.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl; intros ms' m'.

  intros EQbm. inversion_clear EQbm. simpl. intuition.
  intros EQbm. inversion_clear EQbm. simpl. intuition.
  intros EQbm. inversion_clear EQbm. simpl. intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  case_eq (bindc_msg m2). intros ms2 m2' EQm2.
  intros EQbm. inversion_clear EQbm. simpl.
  unfold bmsg_vars_bound in *. rewrite <- LC_comm.
  generalize (IHm1 ms1 m1' EQm1). intros SUB1.
  generalize (IHm2 ms2 m2' EQm2). intros SUB2.
  clear IHm1 IHm2 ms' m' EQm1 EQm2.
  intros a. set_iff. intuition.
  generalize (SUB1 a H0). set_iff. intuition.
  generalize (SUB2 a H0). set_iff. intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  intros EQbm. inversion_clear EQbm. simpl.
  intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  intros EQbm. inversion_clear EQbm. simpl.
  generalize (IHm1 ms1 m1' EQm1). intros SUB1.
  intros a. set_iff. intuition.
  generalize (SUB1 a H). set_iff. intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  intros EQbm. inversion_clear EQbm. simpl.
  generalize (IHm1 ms1 m1' EQm1). intros SUB1.
  intros a. set_iff. intuition.
  generalize (SUB1 a H). set_iff. intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  intros EQbm. inversion_clear EQbm. simpl.
  generalize (IHm1 ms1 m1' EQm1). intros SUB1.
  intros a. set_iff. intuition.
  generalize (SUB1 a H). set_iff. intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  intros EQbm. inversion_clear EQbm. simpl.
  generalize (IHm1 ms1 m1' EQm1). intros SUB1.
  intros a. set_iff. intuition.
  generalize (SUB1 a H). set_iff. intuition.

  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  intros EQbm. inversion EQbm. subst ms'. subst m'.
  clear EQbm. simpl.
  generalize (IHm1 ms1 m1' EQm1). intros SUB1.
  intros a. set_iff. intuition.
  generalize (SUB1 a H0). set_iff. intuition.
Qed.
Hint Resolve bindc_vars_bound.

(*
Lemma open_bmsgs_app:
  forall ms1 ms2 vs o o',
    open_bmsgs ms1 vs = (o, nil) ->
    open_bmsgs ms2 (union vs (bmsg_vars_bound o)) = (o', nil) ->
    open_bmsgs (ms1 ++ ms2) vs = (o ++ o', nil).
Proof.
Admitted. (* XXX *)


Lemma type_bindc_open_eq:
  forall m vs vs' ms' m',
    type_msg_recv vs m ->
    bindc_msg m = (ms', m') ->
    Subset (union vs (message_vars_bound m')) vs' ->
    open_bmsgs ms' vs' = (ms', nil).
Proof.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl; 
    intros vs vs' ms' m' Tmr EQbm SUB.

  inversion_clear EQbm. simpl. compute. reflexivity.
  inversion_clear EQbm. simpl. compute. reflexivity.
  inversion_clear EQbm. simpl. compute. reflexivity.
  
  revert EQbm SUB.
  case_eq (bindc_msg m1). intros ms1 m1' EQm1.
  case_eq (bindc_msg m2). intros ms2 m2' EQm2.
  intros EQbm. inversion_clear EQbm. simpl.
  intros SUB.
  destruct Tmr as [Tmr1 Tmr2].
  cut (Subset (union vs (message_vars_bound m1')) vs'). intros SUB1.
  generalize (IHm1 vs vs' ms1 m1' Tmr1 EQm1 SUB1).
  intros EQob1.
  rewrite (open_bmsgs_app ms1 ms2 vs' ms1 nil ms2 nil).
  reflexivity. assumption.
  simpl. apply IHm2 with vs m2'.
  assumption. assumption.
  intro e. set_iff. intuition.
  left. apply SUB. set_iff. intuition.
  left. apply SUB. set_iff. intuition.
  intro e. set_iff. intuition.
  apply SUB. set_iff. intuition.
  apply SUB. set_iff. intuition.

XXX


Lemma type_bindc_open_perm:
  forall m vs ms ms' m' o c,
    type_msg_recv vs m ->
    bindc_msg m = (ms', m') ->
    let vs' := union vs (message_vars_bound m') in
      open_bmsgs (ms ++ ms') vs' = (o, c) ->
      sublist (id*msg) ms' o.
Proof.
  intros m vs ms ms' m' o c.
  intros Tm EQbm vs' EQob.
  intros [msgi msgv] IN.
  cut (~ List.In (msgi, msgv) c). intros NIN.
  cut (List.In (msgi,msgv) (o++c)). intros INm.
  rewrite <- in_app_iff in INm. intuition.
  apply Permutation_in with (ms++ms').
  replace o with (fst (open_bmsgs (ms++ms') vs')).
  replace c with (snd (open_bmsgs (ms++ms') vs')).
  apply open_bmsgs_correct_perm.
  rewrite EQob. simpl. reflexivity.
  rewrite EQob. simpl. reflexivity.
  rewrite <- in_app_iff. intuition.
  intros InC.
  generalize (open_bmsgs_closed (ms++ms') vs' o c EQob msgi msgv InC).
  
  

XXX

Admitted.


Lemma bindco_action_type_pres:
  forall r a ns vs es (ms:list (id*msg)),
    type_action ns vs r es a ->
    Subset (idset (map (fun e => fst e) ms)) vs ->
    type_action ns vs r es (bindco_action r vs ms a).
Proof.
  intros r.

  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr m a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar phi a IHa ];
    intros ns vs es ms;
      [ idtac
        | case (ME.eq_dec r fr); [intro EQ; subst fr | intro NEQ;
          case (ME.eq_dec r tr); [intros EQ'; subst tr | intros NEQ']]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]
        | case (ME.eq_dec r ar); [intro EQ; subst ar | intro NEQ]];
      simpl; intros Ta SUB.

  trivial.

  case (ME.eq_dec r r); intros EQ; [clear EQ | absurd (r = r); intuition].
  intuition.

  case_eq (open_bmsgs ms (union vs (message_vars_bound m))).
  intros o c EQoc. simpl.
  case (ME.eq_dec r fr); intros EQ; [intuition | clear EQ].
  case (ME.eq_dec r r); intros EQ; [clear EQ | absurd (r = r); intuition].
  split. intuition.
  apply build_match_type with ms c. assumption.
  apply IHa.
  apply type_action_subset with (union vs (message_vars_bound m)).
  intuition. intuition.
  apply subset_trans with (union vs (message_vars_bound m));
    [idtac | intuition].
  apply subset_trans with vs; [idtac | intuition].
  intros e Inc. apply SUB.
  rewrite in_idset_iff in *.
  rewrite in_map_iff in *.
  destruct Inc as [[bv bm] [EQe Inc]].
  subst e. exists (bv,bm). simpl. split.
  trivial.
  apply Permutation_in with (o++c).
  apply Permutation_sym.
  generalize (open_bmsgs_correct_perm 
    ms (union vs (message_vars_bound m))).
  rewrite EQoc. simpl. intuition.
  rewrite <- in_app_iff. intuition.

  case (ME.eq_dec r fr); intros EQ; [intuition | clear EQ].
  case (ME.eq_dec r tr); intros EQ; [intuition | clear EQ].
  intuition.

  case (ME.eq_dec r r); intros EQ; [clear EQ | absurd (r = r); intuition].
  intuition.

  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ].
  intuition.

  case (ME.eq_dec r r); intros EQ; [clear EQ | absurd (r = r); intuition].
  intuition. 
  apply type_msg_send_subset with 
    (union (idset (map (fun e => fst e) ms)) vs); intuition.

  case (ME.eq_dec r ar); intros EQ; [intuition | clear EQ].
  intuition.

  case_eq (bindc_msg m). intros ms' m' EQm.
  set (vs' := union vs (message_vars_bound m')).
  case_eq (open_bmsgs (ms ++ ms') vs').
  intros o c EQoc. simpl.
  case (ME.eq_dec r r); intros EQ; [clear EQ | absurd (r = r); intuition].
  split. intuition.
  split. intuition.
  split. replace m' with (snd (bindc_msg m)).
  apply bindc_msg_type_recv. rewrite EQm. simpl. reflexivity.
  fold vs'.
  apply build_match_type with (ms++ms') c. trivial.
  apply IHa.
  apply type_action_subset with (union vs (message_vars_bound m)).
  intuition. 

XXX

Qed.
Hint Resolve bindco_action_type_pres.

XXXX
*)

Definition bindco_spec_all s :=
  spec_all_lift 
   (spec_lift (fun r bs es a => bindco_action r (varlist_set bs) nil a))
   (rerecv_spec s).