Require Export WPPL.
Set Implicit Arguments.

Definition env a := id -> a.

Definition init a (d:a) : env a := fun _ => d.

Definition ext a i f (s:env a) : env a :=
  fun i' =>
    if ME.eq_dec i i' then
      f (s i')
      else
        s i'.

Theorem ext_eq_iff:
  forall A i f s,
    (@ext A i f s) i = f (s i).
Proof.
  intros A i f s. unfold ext.
  case (ME.eq_dec i i). trivial.
  intros H. absurd (i = i); auto.
Qed. 

Theorem ext_subset_1:
  forall i f s,
    (forall s', Subset s' (f s')) ->
    Subset (s i) (ext i f s i).
Proof.
  intros i f s S e N.
  rewrite ext_eq_iff. apply S. trivial.
Qed.
Hint Resolve ext_subset_1.

Theorem ext_subset_2:
  forall i i' f s,
    i' <> i ->
    Subset (s i) (ext i' f s i).
Proof.
  intros i i' f s NQ e.
  unfold ext. case (ME.eq_dec i' i); intros Q; [contradiction|clear Q].
  trivial.
Qed.
Hint Resolve ext_subset_2.

Theorem ext_subset:
  forall i i' f s,
    (forall s', Subset s' (f s')) ->
    Subset (s i) (ext i' f s i).
Proof.
  intros i i' f s S.
  case (ME.eq_dec i' i); intros Q.
  subst i'. apply ext_subset_1. trivial.
  apply ext_subset_2. trivial.
Qed.
Hint Resolve ext_subset.

Fixpoint type_action' rs ns vs es a :=
 match a with
 | Return => 
   forall r, List.In r rs ->
     exist_vars (vs r) (varlist_set (es r))
 | Comm f t ffs m tfs a =>
   (In t (ns f) \/ In t (vs f)) 
   /\ f <> t
   /\ exist_vars (vs f) (formulas_vars_prove ffs)
   /\ let vs' := ext f (union (formulas_vars_bound ffs)) vs in
      (type_msg_send (vs' f) m) 
   /\ (type_msg_recv (vs' t) m)
   /\ let vs'' := ext t (union (message_vars_bound m)) vs' in
      exist_vars (vs'' t) (formulas_vars_prove tfs)
   /\ (type_action' rs
       (ext t (add f) (ext f (add t) ns))
       (ext t (union (formulas_vars_bound tfs)) vs'')
       es a)
   | New ar v nt a =>
     (type_newtype v nt) 
     /\ (type_action' rs ns (ext ar (add (varid v)) vs) es a)
 | Bind ar v m a =>
    (type_varT v Msg) 
    /\ (type_msg_send (vs ar) m)
    /\ (type_action' rs ns (ext ar (add (varid v)) vs) es a)
 | Match ar v m a =>
   type_varT v Msg
   /\ exist_var (vs ar) v
   /\ type_msg_recv (vs ar) m
   /\ type_action' rs ns (ext ar (union (message_vars_bound m)) vs) es a
 | Derive ar f a =>
   exist_vars (vs ar) (formulas_vars_prove f)
   /\ type_action' rs ns (ext ar (union (formulas_vars_bound f)) vs) es a
 end.

Theorem type_action'_subset:
  forall a rs ns vs vs' es,
    type_action' rs ns vs es a ->
    (forall r, Subset (vs r) (vs' r)) ->
    type_action' rs ns vs' es a.
Proof.
  simple induction a; clear a; simpl;
    unfold exist_vars, exist_var;
    [ idtac
      | intros fr tr ffs m tfs a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar f a IHa ];
    intros rs ns vs vs' es Ta S.

  intros r Nr e Nv. apply S. apply Ta; trivial.

  split. intuition. right. apply S. trivial.
  split. intuition. split. intros b IN. apply S. intuition.
  split. apply type_msg_send_subset with 
   (ext fr (union (formulas_vars_bound ffs)) vs fr). intuition.
  intros b. rewrite ext_eq_iff. rewrite ext_eq_iff. set_iff.
  generalize (S fr b). intuition.
  destruct Ta as [IN [EQ [SS [TMS [TMR [SS' TA]]]]]].
  split. unfold ext in TMR. unfold ext.
  destruct (ME.eq_dec fr tr) as [Q'|Q']; [contradiction | clear Q'].
  apply type_msg_recv_subset with (vs tr); intuition.
  split. rewrite ext_eq_iff. unfold ext.
  rewrite ext_eq_iff in SS'. unfold ext in SS'.
  destruct (ME.eq_dec fr tr) as [Q'|Q']; [contradiction | clear Q'].
  intros b IN'. apply SS' in IN'. revert IN'. set_iff.
  generalize (S tr b). intuition.
  apply IHa with (ext tr (union (formulas_vars_bound tfs))
    (ext tr (union (message_vars_bound m))
     (ext fr (union (formulas_vars_bound ffs)) vs))). trivial.
  intros r b. unfold ext. generalize (S r b).
  destruct (ME.eq_dec tr r) as [Qtr | Qtr];
  destruct (ME.eq_dec fr r) as [Qfr | Qfr].
  subst r. contradiction.
  set_iff. intuition. set_iff. intuition. trivial.

  intuition. apply IHa with (ext ar (add (varid v)) vs). trivial.
  intros r e. unfold ext.
  case (ME.eq_dec ar r); intros Q; set_iff.
  intros [P|P]. intuition. right. apply S. trivial. apply S.

  split. intuition.
  split. apply type_msg_send_subset with (vs ar); intuition.
  apply IHa with (ext ar (add (varid v)) vs). intuition.
  intros r e. unfold ext.
  case (ME.eq_dec ar r); intros Q; set_iff.
  intros [P|P]. intuition. right. apply S. trivial. apply S.

  split. intuition.
  split. intuition. apply S. trivial.
  split. apply type_msg_recv_subset with (vs ar); intuition.
  apply IHa with (ext ar (union (message_vars_bound m)) vs). intuition.
  intros r e. unfold ext.
  case (ME.eq_dec ar r); intros Q; set_iff.
  intros [P|P]. intuition. right. apply S. trivial. apply S.

  split. intros e N. apply S. intuition.
  apply IHa with (ext ar (union (formulas_vars_bound f)) vs). intuition.
  intros r e. unfold ext.
  case (ME.eq_dec ar r); intros Q; set_iff.
  intros [P|P]. intuition. right. apply S. trivial. apply S.
Qed.
Hint Resolve type_action'_subset.

Definition type_spec' s :=
match s with
| Spec rss a =>
  let rs := List.map (fun rs => match rs with RSpec r _ _ => r end) rss in
  let ns := init empty in
  let vs :=  List.fold_right (fun rs env =>
    match rs with RSpec r bs _ => ext r (fun _ => (varlist_set bs)) env end)
    (init empty) rss in
  let es := List.fold_right (fun rs env =>
    match rs with RSpec r _ es => ext r (fun _ => es) env end)
    (init nil) rss in
  type_action' rs ns vs es a
end.

Definition wellformed_spec' s :=
  connected_spec s 
  /\ type_spec' s.

(* Goal: Create a new nonce for each role and thread it through the
whole comm chain. *)

(*Parameter nonce_id : id -> id.*)
Definition nonceid (i:id) : id := str (nonce_id (rts i)).

(*Definition nonce_id (i:id) := new_id tt.*)
Definition nonce_var i := Var i Nonce.

Definition toset l :=
  List.fold_right add empty l.

Theorem toset_iff:
  forall l x, List.In x l <-> In x (toset l).
Proof.
  induction l as [|e l]; simpl. intros x; set_iff. intuition.
  intros x. set_iff. rewrite IHl. intuition.
Qed.

Theorem elements_toset:
  forall s,
    Equal s (toset (elements s)).
Proof.
  intros s. intros a. rewrite <- toset_iff.
  rewrite elements_iff. rewrite InA_alt.
  split. intros [y [P1 P2]]. subst a. trivial.
  intros P. exists a. intuition.
Qed.

Lemma type_msg_send_vars_msg_toset:
  forall l vs,
    type_msg_send vs (vars_msg l) <-> Subset (toset (List.map varid l)) vs.
Proof.
  unfold Subset. 
  induction l as [|e l]; intros vs; simpl; split; unfold exist_var.
  intros H a. set_iff. contradiction. trivial.
  intros [N Tms] a. set_iff. rewrite <- toset_iff.
  intros [P|P]. subst a. trivial.
  rewrite IHl in Tms. apply Tms. rewrite <- toset_iff. trivial.
  intros P. rewrite IHl. split.
  apply P. set_iff. intuition.
  intros a N. apply P. set_iff. intuition.
Qed.

Definition prep_thread rs :=
  (vars_msg (List.map nonce_var (elements rs))).

Definition prep_message rs m :=
  Join (prep_thread rs) m.

Lemma map_v_nv:
  forall l, map varid (map nonce_var l) = l.
Proof.
  induction l; simpl; intuition. rewrite IHl. trivial.
Qed.

Theorem prep_thread_type_send:
  forall rs vs,
    Subset rs vs <->
    type_msg_send vs (prep_thread rs).
Proof.
  intros rs vs. unfold prep_thread.
  rewrite type_msg_send_vars_msg_toset.
  rewrite map_v_nv. rewrite <- elements_toset. intuition.
Qed.

Theorem prep_thread_type_recv:
  forall rs vs,
    type_msg_recv vs (prep_thread rs) <-> True.
Proof.
  intros rs vs. unfold prep_thread. intuition.
Qed.

Lemma vars_msg_bound_toset:
  forall l, 
    Equal (message_vars_bound (vars_msg l)) 
    (toset (List.map varid l)).
Proof.
  intros l a. rewrite <- toset_iff.
  induction l as [|e l]. simpl. set_iff. intuition.
  simpl. set_iff. rewrite IHl. intuition.
Qed.

Theorem prep_thread_bound:
  forall rs,
    Equal (message_vars_bound (prep_thread rs)) rs.
Proof.
  unfold prep_thread. intros rs.
  rewrite vars_msg_bound_toset.
  rewrite map_v_nv.
  rewrite <- elements_toset.
  reflexivity.
Qed.

Fixpoint thread_action rs a :=
  match a with
    | Return => Return
    | Comm fr tr ffs m tfs a =>
      let i := nonceid fr in
      if mem i rs then
        let m' := prep_message rs m in
          Comm fr tr ffs m' tfs (thread_action rs a)
      else
        let rs' := add i rs in
        let m' := prep_message rs' m in
          New fr (nonce_var i) New_nonce
           (Comm fr tr ffs m' tfs (thread_action rs' a))
    | New r v nt a =>
      New r v nt (thread_action rs a)
    | Bind r v m a => 
      Bind r v m (thread_action rs a)
    | Match r v m a =>
      Match r v m (thread_action rs a)
    | Derive r fs a =>
      Derive r fs (thread_action rs a)
  end.

Fixpoint retract_action r a :=    
  match a with
    | Return => Return
    | Comm fr tr ffs m tfs a =>
      if ME.eq_dec r fr then
        Comm fr tr ffs m tfs (retract_action r a)
      else if ME.eq_dec r tr then
        Comm fr tr ffs m tfs (retract_action r a)
      else
        retract_action r a
    | New ar v nt a =>
      if ME.eq_dec r ar then
        New ar v nt (retract_action r a)
      else
        retract_action r a
    | Bind ar v m a => 
      if ME.eq_dec r ar then
        Bind ar v m (retract_action r a)
      else
        retract_action r a
    | Match ar v m a =>
      if ME.eq_dec r ar then
        Match ar v m (retract_action r a)
      else
        retract_action r a
    | Derive ar fs a =>
      if ME.eq_dec r ar then
        Derive ar fs (retract_action r a)
      else
        retract_action r a
  end.

Lemma retract_action_type_1:
  forall a ns vs r es,
    type_action ns vs r es (retract_action r a) ->
    type_action ns vs r es a.
Proof.
 simple induction a; clear a; simpl;
 [ idtac
 | intros fr tr ffs m tfs a IHa
 | intros ar v nt a IHa
 | intros ar v m a IHa
 | intros ar v m a IHa
 | intros ar f a IHa ];
 intros ns vs r es;
 [ idtac
 | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
   case (ME.eq_dec r tr); intro EQ ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
 simpl; try (intuition; fail);
 [ case (ME.eq_dec r fr); [intro EQ' | intro NEQ';
   case (ME.eq_dec r tr); intro EQ' ]
 | case (ME.eq_dec r fr); [intro EQ' | intro NEQ';
   case (ME.eq_dec r tr); intro EQ' ]
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']];
 intuition.
Qed.

Lemma retract_action_type_2:
  forall a ns vs r es,
    type_action ns vs r es a ->
    type_action ns vs r es (retract_action r a).
Proof.
 simple induction a; clear a; simpl;
 [ idtac
 | intros fr tr ffs m tfs a IHa
 | intros ar v nt a IHa
 | intros ar v m a IHa
 | intros ar v m a IHa
 | intros ar f a IHa ];
 intros ns vs r es;
 [ idtac
 | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
   case (ME.eq_dec r tr); intro EQ ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
 simpl; try (intuition; fail);
 [ case (ME.eq_dec r fr); [intro EQ' | intro NEQ';
   case (ME.eq_dec r tr); intro EQ' ]
 | case (ME.eq_dec r fr); [intro EQ' | intro NEQ';
   case (ME.eq_dec r tr); intro EQ' ]
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']
 | case (ME.eq_dec r ar); [intro EQ' | intro NEQ']];
 intuition.
Qed.

Theorem retract_action_iff:
  forall a ns vs r es,
    type_action ns vs r es a <->
    type_action ns vs r es (retract_action r a).
Proof.
  split. apply retract_action_type_2. apply retract_action_type_1.
Qed.

Theorem thread_action_type'_some:
  forall a rs ns vs es,
    type_action' rs ns vs es a ->
    forall r vrs,
      connected_action (Some r) a ->
      Subset vrs (vs r) ->
      type_action' rs ns vs es (thread_action vrs a).
Proof.
  simple induction a; clear a; simpl;
    unfold exist_vars, type_varT;
      [ idtac
        | intros fr tr ffs m tfs a IHa
        | intros ar v nt a IHa
        | intros ar v m a IHa
        | intros ar v m a IHa
        | intros ar f a IHa ];
      intros rs ns vs es Ta r vrs [Q Ca] S.

  intuition.

  destruct Ta as [IN [EQ [SS1 [TMS [TMR [SS2 TA]]]]]].
  case_eq (mem (nonceid fr) vrs); intros MEM; simpl;
  rewrite <- prep_thread_type_send; rewrite prep_thread_type_recv.

  subst r. split. intuition. split. intuition.
  split. intuition. split. split.
  apply Subset_trans with (vs fr). trivial. 
  rewrite ext_eq_iff. intros b IN'; set_iff; intuition. intuition.
  split. split. trivial. intuition.
  split.  unfold exist_vars. rewrite ext_eq_iff.
  rewrite prep_thread_bound.
  intros b IN'. apply SS2 in IN'. revert IN'. rewrite ext_eq_iff.
  set_iff. intuition.
  apply IHa with tr. apply type_action'_subset with
   (ext tr (union (formulas_vars_bound tfs))
    (ext tr (union (message_vars_bound m))
     (ext fr (union (formulas_vars_bound ffs)) vs))).
  intuition. intros r' b. unfold ext.
  destruct (ME.eq_dec fr r') as [Qfr | Qfr];
    destruct (ME.eq_dec tr r') as [Qtr | Qtr].
  subst r'. subst fr. absurd (tr = tr); auto. trivial.
  rewrite prep_thread_bound. set_iff. intuition. trivial. trivial.
  intros b IN'. rewrite ext_eq_iff. rewrite ext_eq_iff.
  unfold ext.
  destruct (ME.eq_dec fr tr) as [Qfr | Qfr]; [contradiction | clear Qfr].
  set_iff. rewrite prep_thread_bound. intuition.

  destruct (type_eq_dec Nonce Nonce) as [Qn|Qn];
    [clear Qn|absurd (Nonce=Nonce); auto]. subst r.
  split. trivial. split. rewrite ext_eq_iff. set_iff. intuition.
  split. trivial. split. unfold exist_vars. rewrite ext_eq_iff.
  intros b IN'. set_iff. intuition.
  split. split. rewrite ext_eq_iff. rewrite ext_eq_iff.
  intros b. set_iff. intros [Q1|IN1]. intuition. apply S in IN1. intuition.
  rewrite ext_eq_iff. rewrite ext_eq_iff. rewrite ext_eq_iff in TMS.
  apply type_msg_send_subset with (union (formulas_vars_bound ffs) (vs fr)).
  trivial. intros b. set_iff. intuition.
  split. split. trivial. revert TMR. unfold ext.
  destruct (ME.eq_dec fr tr) as [Qfr|Qfr]; [contradiction|clear Qfr].
  trivial. split. revert SS2. unfold exist_vars, ext.
  destruct (ME.eq_dec fr tr) as [Qfr|Qfr]; [contradiction|clear Qfr].
  destruct (ME.eq_dec tr tr) as [Qtr|Qtr]; [clear Qtr|absurd(tr=tr);auto].
  rewrite prep_thread_bound.
  intros SS2 b IN'. apply SS2 in IN'. revert IN'. set_iff. intuition.
  apply IHa with tr; trivial. apply type_action'_subset with
   (ext tr (union (formulas_vars_bound tfs))
    (ext tr (union (message_vars_bound m))
     (ext fr (union (formulas_vars_bound ffs)) vs))). trivial.
  intros r' b. unfold ext.
  destruct (ME.eq_dec fr r') as [Qfr | Qfr]; 
    destruct (ME.eq_dec tr r') as [Qtr | Qtr].
  subst r'. subst fr. absurd (tr = tr); auto. trivial.
  set_iff. intuition. rewrite prep_thread_bound.
  set_iff. intuition. trivial. unfold ext.
  destruct (ME.eq_dec fr tr) as [Qfr|Qfr]; [contradiction|clear Qfr].
  destruct (ME.eq_dec tr tr) as [Qtr|Qtr]; [clear Qtr|absurd(tr=tr);auto].
  rewrite prep_thread_bound. intros b. set_iff. intuition.

  subst r. intuition. apply IHa with ar. trivial. trivial.
  rewrite ext_eq_iff. intros e; set_iff; intuition.

  subst r. intuition. apply IHa with ar. trivial. trivial.
  rewrite ext_eq_iff. intros e; set_iff; intuition.

  subst r. intuition. apply IHa with ar. trivial. trivial.
  rewrite ext_eq_iff. intros e; set_iff; intuition.

  subst r. intuition. apply IHa with ar. trivial. trivial.
  rewrite ext_eq_iff. intros e; set_iff; intuition.
Qed.
Hint Resolve thread_action_type'_some.

Theorem thread_action_type'_none:
  forall a rs ns vs es,
    type_action' rs ns vs es a ->
    connected_action None a ->
    type_action' rs ns vs es (thread_action empty a).
Proof.
  intros a rs ns vs es Ta Ca.
  destruct a; simpl in Ca;
    try (simpl in *; trivial; fail);
  (apply thread_action_type'_some with i;
    [ trivial
    | simpl; intuition
    | intros e; set_iff; intuition ]).
Qed.
Hint Resolve thread_action_type'_none.

Theorem thread_action_connected_some:
  forall a r rs,
    connected_action (Some r) a ->
    connected_action (Some r) (thread_action rs a).
Proof.
  simple induction a; clear a; simpl;
    [ idtac
      | intros fr tr ffs m tfs a IHa
      | intros ar v nt a IHa
      | intros ar v m a IHa
      | intros ar v m a IHa
      | intros ar f a IHa ];
    intros r rs; intuition.
  case (mem (nonceid fr) rs); simpl; intuition.
Qed.
Hint Resolve thread_action_connected_some.

Theorem thread_action_connected_none:
  forall a rs,
    connected_action None a ->
    connected_action None (thread_action rs a).
Proof.
  destruct a; simpl; intros rs; intuition.
  case (mem (nonceid i) rs); simpl; intuition.
Qed.
Hint Resolve thread_action_connected_none.

Definition thread_spec s :=
match s with
| Spec rss a =>
  Spec rss (thread_action empty a)
end.

Theorem thread_spec_wellformed':
  forall s,
    wellformed_spec' s ->
    wellformed_spec' (thread_spec s).
Proof.
  destruct s as [rss a]. unfold wellformed_spec'.
  simpl. intuition.
Qed.

(* Does the transformation do the larger task? *)

Require Export CPPLlb.

Lemma vars_msg_redact:
  forall rs vs,
    msg_redact_recv vs (vars_msg rs) = vars_msg rs.
Proof.
  induction rs as [|r rs]; intros vs; simpl. trivial.
  rewrite IHrs. trivial.
Qed.

Theorem prep_message_redact:
  forall rs vs m,
    msg_redact_recv vs (prep_message rs m)
    = prep_message rs (msg_redact_recv vs m).
Proof.
  unfold prep_message. unfold prep_thread.
  intros rs vs m. simpl. rewrite vars_msg_redact. trivial.
Qed.

Definition vartype v := match v with Var vi vt => vt end.

Lemma vars_msg_matchm:
  forall v1 v2 rs1 rs2 i1 i2,
    (forall v, List.In v rs1 -> vartype v = Nonce) ->
    (forall v, List.In v rs2 -> vartype v = Nonce) ->
    List.In (Var i1 Nonce) rs1 -> In i1 v1 ->
    List.In (Var i2 Nonce) rs2 -> In i2 v2 ->
    ~ matchm (cmpv_dist v1 v2) 
      (vars_msg rs1) (vars_msg rs2).
Proof.
 intros v1 v2.

 induction rs1 as [|r1 rs1]; simpl. intuition.
 induction rs2 as [|r2 rs2]; simpl. intuition.

 intros i1 i2 AN1 AN2 N1r N1v N2r N2v [P1 P2].
 destruct N1r as [Q1 | N1r]. subst r1.
 unfold cmpv_dist in P1. destruct r2 as [vi vt].
 intuition.
 destruct N2r as [Q2 | N2r]. subst r2.
 destruct r1 as [vi vt]. cut (vt = Nonce). intros Q. subst vt.
 unfold cmpv_dist in P1. intuition.
 generalize (AN1 (Var vi vt)). simpl. intuition.
 clear IHrs2. apply (IHrs1 rs2 i1 i2); intuition.
Qed.

Theorem prep_matchm:
  forall v1 v2 rs1 rs2 m1 m2 i1 i2,
    In i1 rs1 -> In i1 v1 ->
    In i2 rs2 -> In i2 v2 ->
    ~ matchm (cmpv_dist v1 v2) (prep_message rs1 m1) (prep_message rs2 m2).
Proof.
 intros v1 v2 rs1 rs2 m1 m2 i1 i2.
 intros N1r N1v N2r N2v. unfold prep_message.
 simpl. intros [P1 P2]. clear P2 m1 m2.
 unfold prep_thread in P1.
 apply (@vars_msg_matchm v1 v2
   (map nonce_var (elements rs1)) (map nonce_var (elements rs2)) i1 i2);
 trivial.
 intros [vi vt]. rewrite in_map_iff. unfold nonce_var. simpl. intros [x [Q P]].
 inversion Q. trivial.
 intros [vi vt]. rewrite in_map_iff. unfold nonce_var. simpl. intros [x [Q P]].
 inversion Q. trivial.
 rewrite in_map_iff. exists i1. unfold nonce_var. split. trivial.
 rewrite elements_iff in N1r. rewrite InA_alt in N1r. 
 destruct N1r as [y [Q P]]. subst y. trivial.
 rewrite in_map_iff. exists i2. unfold nonce_var. split. trivial.
 rewrite elements_iff in N2r. rewrite InA_alt in N2r. 
 destruct N2r as [y [Q P]]. subst y. trivial.
Qed.

(*Theorem thread_optimization:
  forall r rc s p,
    compile_spec r rc (thread_spec s) = Some p ->
    let f1 := (fun s => 
      strand_redact empty (strand_chop_fst_recv (strand_ext_only s))) in
    let f2 := (fun s =>
      strand_redact empty (strand_ext_only s)) in
    let mk_cmpv := cmpv_dist in
      forall s1 s2 v1 v2,
        eval_proc p s1 v1 ->
        eval_proc p s2 v2 ->
        ~ strand_share_recvs (mk_cmpv v1 v2) (mk_cmpv v2 v1) (f1 s1) (f2 s2).
Proof.
  intros r rc s p Q f1 f2 mk_cmpv s1 s2 v1 v2 EP1 EP2.
  unfold strand_share_recvs. unfold list_share_recvs.
  intros [m1 [m2 [MM [N1 N2]]]].
  unfold f1, f2, mk_cmpv in *. clear f1 f2 mk_cmpv.
  revert Q. destruct s as [rss a]. simpl.
  case_eq (rspec_find r rss). intros [r' bs es].
Focus 2.
  intros P1 P2. discriminate P2.
  intros Q1 Q2. inversion Q2. subst p. clear Q2.
  inversion EP1. clear EP1. 
  unfold msg, s', msg0, msg1 in *. clear s' msg0 msg1.
  unfold msg in H0. clear msg. subst v n xs psi c s1.
  rename H5 into EC1. simpl in N1. revert N1.
  case (const_eq_dec Common.C.CALL CPPL.C.MSG). intros Q.
*)
