(* An abstract syntax tree for WPPL programs without types
   Copyright (C) 2007 Jay McCarthy
*)

open Ast_t
open Pos

type undecl = id * id_type option
let undecl_eq ((_, lds), _) ((_, rds), _) =
  lds = rds

type unfdecl =
    | UDataV of undecl
    | ULogicV of undecl

type unmessage =
    | UDecl of undecl
    | UCat of unmessage * unmessage
    | UConst of pos * string
    | UPubEncrypt of unmessage * undecl
    | UPubSign of unmessage * undecl
    | USymEncrypt of unmessage * undecl
    | USymSign of unmessage * undecl
    | UHash of unmessage
    | UMBind of undecl * unmessage

type unterm = 
    | UTerm of id * unfdecl list

type unterms = unterm list

type unrole_spec = role * undecl list * undecl list

type unaction =
    | UZero
    | UComm of role * role * unterms * unmessage * unterms * unaction
    | UNew of role * undecl * new_type * unaction
    | UBind of role * undecl * unmessage * unaction
    | UMatch of role * undecl * unmessage * unaction
    | UDerive of role * unterms * unaction

type unspec = unrole_spec list * unaction

type unletval =
    | UNew_nonce
    | UNew_symkey
    | UNew_pubkey
    | UAccept
    | UConnect of undecl
    | UMsg of unmessage

type uncont =
    | UFail
    | UReturn
    | ULet of undecl * unletval * uncont
    | UCDerive of unterms * uncont * uncont
    | USend of unterms * undecl * unmessage * uncont * uncont
    | UReceive of undecl * unmessage * unterms * uncont * uncont
    | UCall of unterms * id * undecl list * 
               undecl list * unterms * uncont * uncont
    | UCMatch of undecl * unmessage * unterms * uncont * uncont


type unproc = 
    | UProc of id * undecl list * undecl list * unterms * unterms * uncont
