(* Dealing with WPPL environments
   Copyright (C) 2007 Jay McCarthy
*)

open Ast_t
open Pos

(* Combiners *)
let cmb_equal t0 t1 =
  match (t0,t1) with
      Name _, Name _ -> Name empty_pos
    | Text _, Text _ -> Text empty_pos
    | Msg _, Msg _ -> Msg empty_pos
    | Nonce _, Nonce _ -> Nonce empty_pos
    | Pubkey _, Pubkey _ -> Pubkey empty_pos
    | Symkey _, Symkey _ -> Symkey empty_pos
    | Channel _, Channel _ -> Channel empty_pos
    | _ -> 
	failwith 
	  (Printf.sprintf "Types do not match: %s vs %s"
	      (type2str t0) (type2str t1))
  
let cmb_merge t0 t1 =
  match (t0,t1) with
      None, None ->
	None
    | Some _, None ->
	t0
    | None, Some _ ->
	t1
    | Some it0, Some it1 ->
	Some (cmb_equal it0 it1)

(* Environment helpers *)    
let decl_in_env d renv =
  let ((_, ids), _) = d in
    PMap.mem ids renv

let pmap_add_decl d v p =
  let ((_, ids), _) = d in
    PMap.add ids v p

let pmap_find_decl d p =
  let ((_, ids), _) = d in
    PMap.find ids p

let renv_to_str renv =
  PMap.foldi
    (fun ids _ str ->
      Printf.sprintf "%s %s" str ids)
    renv
    ""

let role_env_to_str role_env =
  PMap.foldi
    (fun r renv str ->
      Printf.sprintf "%s [%s: %s]" str r (renv_to_str renv))
    role_env
    ""

let role_lookup r role_env =
  let (_, rstr) = r in
    try
      PMap.find rstr role_env
    with
	Not_found ->
	  PMap.empty

let role_add r v role_env =
  let (_, rstr) = r in
    PMap.add rstr v role_env

let extend_role r renv_fn role_env =
  let renv =
    try
      role_lookup r role_env
    with
	Not_found ->
	  PMap.empty
  in
    role_add r (renv_fn renv) role_env	

let extend_decl cmb d renv =
  let ((_, ids), t) = d in
    if PMap.mem ids renv then
      PMap.add ids (cmb (PMap.find ids renv) t) renv
    else
      PMap.add ids t renv

let extend_var cmb ids t renv =
    if PMap.mem ids renv then
      PMap.add ids (cmb (PMap.find ids renv) t) renv
    else
      PMap.add ids t renv

let extend_decls cmb ds renv =
  List.fold_left
    (fun renv decl ->
      extend_decl cmb decl renv)
    renv
    ds

let add_decls ds renv =
  List.fold_left
    (fun renv decl ->
      pmap_add_decl decl () renv)
    renv
    ds

let merge_env env1 env2 =
  PMap.foldi PMap.add env1 env2
