(* A reader for WPPL programs.
   Copyright (C) 2007 Jay McCarthy
*)
open Lexing

let lexbuf_message lexbuf msg =
  let start_p = lexeme_start_p lexbuf in
  let end_p = lexeme_end_p lexbuf in
  let f = start_p.pos_fname in
  let l = start_p.pos_lnum in
  let s =  start_p.pos_cnum - start_p.pos_bol in
  let e = end_p.pos_cnum - end_p.pos_bol in
  Printf.sprintf
    "File \"%s\", line %d, at %d-%d: %s"
    f l s e msg

let report_error lexbuf msg =
  failwith (lexbuf_message lexbuf msg)

let report_parse_error lexbuf =
  report_error lexbuf "syntax error"

let open_in_or_stdin file_name =
  if file_name = "-" then
    stdin
  else
    open_in file_name

let read start file_name =
  let chan = open_in_or_stdin file_name in
  let lexbuf = Lexing.from_channel chan in
    Scanner.init lexbuf file_name ;
    try
      let prog =
	start Scanner.token lexbuf in
	close_in chan;
	prog
    with Parsing.Parse_error ->
      close_in chan;
      report_parse_error lexbuf
      | Failure s ->
	  close_in chan;
	  report_error lexbuf s
	    
let read_spec file_name =
  read Parser.spec file_name

let read_proc file_name =
  read Parser.proc file_name
